/* This file is part of the Palabos library.
 *
 * Copyright (C) 2011 FlowKit Sarl
 * Avenue de Chailly 23
 * 1012 Lausanne, Switzerland
 * E-mail contact: contact@flowkit.com
 *
 * The most recent release of Palabos can be downloaded at
 * <http://www.palabos.org/>
 *
 * The library Palabos is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * The library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "palabos2D.h"
#include "palabos2D.hh"
#include <vector>
#include <cmath>
#include <iostream>
#include <fstream>
#include <iomanip>

using namespace plb;
using namespace plb::descriptors;
using namespace std;

typedef double T;
#define DESCRIPTOR D2Q9Descriptor


void cavitySetup( MultiGridLattice2D<T,DESCRIPTOR>& lattice,
                  ConvectiveRefinementParameters<T> const& parameters,
                  OnLatticeBoundaryCondition2D<T,DESCRIPTOR>& boundaryCondition )
{
    // for each lattice inside the multigrid
    for (plint iLevel=0; iLevel<lattice.getNumLevels(); ++iLevel){
        const plint nx = parameters.getParameters(iLevel).getNx();
        const plint ny = parameters.getParameters(iLevel).getNy();

        boundaryCondition.setVelocityConditionOnBlockBoundaries(lattice.getComponent(iLevel));

        setBoundaryVelocity(lattice.getComponent(iLevel), 
                            lattice.getComponent(iLevel).getBoundingBox(),Array<T,2>(0.,0.) );
                            
        initializeAtEquilibrium(lattice.getComponent(iLevel),
                                lattice.getComponent(iLevel).getBoundingBox(),
                                1., Array<T,2>(0.,0.) );

        T u = parameters.getParameters(iLevel).getLatticeU();
        setBoundaryVelocity(lattice.getComponent(iLevel), Box2D(1, nx-2, ny-1, ny-1), Array<T,2>(u,0.) );
        initializeAtEquilibrium(lattice.getComponent(iLevel), Box2D(1, nx-2, ny-1, ny-1), 1., Array<T,2>(u,0.) );

    }

    lattice.initialize();
}

/// Produce a GIF snapshot of the velocity-norm.
template<typename T, template<typename U> class Descriptor>
void writeGifs(MultiBlockLattice2D<T,Descriptor> lattice, plint iter, string fName) {
    plint imSizeX = 300;
    plint imSizeY = 300;
    ImageWriter<T> imageWriter("leeloo");
    imageWriter.writeScaledGif(createFileName(fName, iter, 6),
                               *computeVelocityNorm(lattice),imSizeX,imSizeY);
}


template<typename T, template<typename U> class Descriptor>
void writeGifs(MultiGridLattice2D<T,Descriptor> const& lattice, plint iter) {
    for (plint iLevel=0; iLevel<lattice.getNumLevels(); ++iLevel) {
        std::stringstream fname("level_");
        fname << "level_" << iLevel << "_";
        writeGifs(lattice.getComponent(iLevel), iter, fname.str());
    }
}



int main(int argc, char* argv[]) {
    plbInit(&argc, &argv);

    global::directories().setOutputDir("./tmp/");
    global::setDefaultMultiScaleManager(new ConvectiveMultiScaleManager());
    
    plint referenceLevel = 0;
    plint behaviorLevel = referenceLevel;
    plint numLevel = 3;
    
    plint N=100;

    IncomprFlowParam<T> baseParameters(
            (T) 1e-2,  // uMax
            (T) 100.,  // Re
            N  ,       // N
            1.,        // lx
            1.         // ly
    );
    const T logT     = (T)0.1;
    const T imSave   = (T)0.2;
    const T maxT     = (T)10.1;
    
    // we choose a convective refinement dx=dt
    ConvectiveRefinementParameters<T> convectiveParameters (
                                            numLevel,referenceLevel,baseParameters );

    Dynamics<T,DESCRIPTOR>* backgroundDynamics = new BGKdynamics<T,DESCRIPTOR>(baseParameters.getOmega());
    
    // main object for the grid refinement
    MultiGridManagement2D management(baseParameters.getNx(),baseParameters.getNy(), numLevel, referenceLevel);
    
    // adding refinements to management
    management.refine(0,Box2D(0,N,5*N/6,N)); // refine the upper sixth of the domain
    management.refine(1,Box2D(0,N/9,8*N/9,N).multiply(2)); // left corner
    management.refine(1,Box2D(8*N/9,N,8*N/9,N).multiply(2)); // right corner
    
    // Parallelization. Here, the domain is parallelized along the x-direction. The position
    //   of the different slices is explicitly created, and provided in units of the
    //   finest lattice.
    std::vector<plint> parallelizeIn;
    plint procNumber = global::mpi().getSize();
    std::vector<std::pair<plint,plint> > ranges;
    plint finestNy = convectiveParameters[numLevel-1].getNy()-1;
    util::linearRepartition(0,finestNy,procNumber,ranges);
    for (plint iBlock=0; iBlock<(plint)ranges.size()-1; ++iBlock){
        parallelizeIn.push_back(ranges[iBlock].second);
    }
    
    management.parallelizeX(parallelizeIn);

    // the boundary condition
    OnLatticeBoundaryCondition2D<T,DESCRIPTOR>*
        boundaryCondition = createLocalBoundaryCondition2D<T,DESCRIPTOR>();

    MultiGridLattice2D<T,DESCRIPTOR> 
        lattice( management, 
                 defaultMultiGridPolicy2D().getBlockCommunicator<T>(management.getNumLevels()),
                 defaultMultiGridPolicy2D().getCombinedStatistics(management.getNumLevels()),
                 backgroundDynamics, behaviorLevel );
        
    T previousIterationTime = T();
    
    cavitySetup(lattice, convectiveParameters, *boundaryCondition);
    
    // Main loop over time iterations.
    for (plint iT=0; iT*baseParameters.getDeltaT()<maxT; ++iT) {
        global::timer("mainLoop").restart();

        if (iT%baseParameters.nStep(imSave)==0 && iT>0) {
            pcout << "Saving Gif ..." << endl;
            writeGifs(lattice, iT);
            pcout << endl;
        }

        if (iT%baseParameters.nStep(logT)==0) {
            pcout << "step " << iT
                  << "; t=" << iT*baseParameters.getDeltaT();
        }

        // Lattice Boltzmann iteration step.
        lattice.collideAndStream();

        if (iT%baseParameters.nStep(logT)==0) {
            pcout << "; av energy="
                  << setprecision(10) << getStoredAverageEnergy(lattice)
                  << "; av rho="
                  << getStoredAverageDensity(lattice) << endl;
            pcout << "Time spent during previous iteration: "
                  << previousIterationTime << endl;
        }

        previousIterationTime = global::timer("mainLoop").stop();
    }

    delete boundaryCondition;
}
