#!/usr/bin/python

#  This file is part of the Palabos library.
#  Copyright (C) 2009-2010 Jonas Latt
#
#  The library Palabos is free software: you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.

import sys; sys.path.append('../src')
import time
from pypal import *

lx,ly    = 300,300       # Dimensions of the system in lattice units.
dx       = 1./lx         # Discrete space step (ly is the reference length).
nx,ny    = lx+1,ly+1     # Lattice resolution.
uMax     = 0.02          # Maximum velocity in lattice units.
Re       = 100.          # Reynolds number (wrt ly).
nu       = uMax*ly / Re  # Viscosity in lattice units.
tau      = 3*nu+0.5      # Relaxation time.
maxIter  = 4000          # Maximum number of iterations.
statIter = 1000          # Intervals at which some statistics is printed.

# Allocate memory for the block-lattice, and assign the dynamics.
lattice = Block2D(nx, ny, D2Q9, BGK(omega=1./tau))
pout.display( lattice.multiBlockInfo() )

# Use a regularized boundary condition.
boundary.regularized().defineVelocityBC(lattice)
# Default all external walls to zero velocity.
lattice.setBoundaryVelocity([0.,0.])
# Top lid has constant in-plane velocity.
lattice[1:nx-2,ny-1].setBoundaryVelocity([uMax,0.])

lattice.initializeAtEquilibrium(1., [0.,0.])

pout.display('Iterating')
snapshots = []
startTime = time.time()
for i in range(0,maxIter+1):
    if (i%statIter==0 and i>0):
        # Compute a velocity snapshot of the system, and keep it safe.
        snapshots.append(lattice.velocityNorm())
        # Apply a reduction operation to compute the average kinetic energy.
        pout.display('The energy is ', lattice.averageEnergy())
        # Use Python's timer to compute the efficiency of the computations.
        elapsed = time.time()-startTime
        Msups = i*nx*ny / elapsed / 1e6
        pout.display('Efficiency: ', Msups, ' Mega site updates per second.')

    # This is the real computation: evolving through a collision-streaming cycle.
    lattice.collideAndStream()

pout.display('Plotting results')
pout.imagesc(snapshots)

# If instead you want to write to a PDF file:
# pout.imagesc(snapshots, 'snapshots.pdf')

