#!/usr/bin/python

#  This file is part of the Palabos library.
#  Copyright (C) 2009-2010 Jonas Latt
#
#  The library Palabos is free software: you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.

import sys; sys.path.append('../src')
import time
from pypal import *

lx,ly    = 800,120       # Dimensions of the system in lattice units.
dx       = 1./ly         # Discrete space step (ly is the reference length).
nx,ny    = lx+1,ly+1     # Lattice resolution.
uMax     = 0.02          # Maximum velocity in lattice units.
Re       = 200.          # Reynolds number (wrt ly).
nu       = uMax*ly / Re  # Viscosity in lattice units.
tau      = 3*nu+0.5      # Relaxation time.
maxIter  = 4000          # Maximum number of iterations.
statIter = 1000          # Intervals at which some statistics is printed.

# Allocate memory for the block-lattice, and assign the dynamics.
lattice = Block2D(nx, ny, D2Q9, BGK(omega=1./tau))
lattice.multiBlockInfo()
# Here, x and y are (parallelized) matrices. It's similar to Matlab's meshgrid.
x,y   = lattice.meshGrid()

# Compute pressure and velocity according to the analytical Poiseuille solution.
pois_pressure = 8*nu*uMax/(ly*ly)*(lx/2.-x)
pois_velocity = Block2D(nx,ny, [0., 0.])
pois_velocity.setComponent(0, 4.*uMax*dx*(y-dx*y*y))

# Use a regularized boundary condition. Inlet and outlet are Poiseuille profile.
boundary.regularized().defineVelocityBC(lattice)
lattice.setBoundaryVelocity(pois_velocity)

lattice.initializeAtEquilibrium(pois_pressure+1., pois_velocity)

# Define the shape of a cylinder and associate bounce-back nodes to its domain.
cx,cy    = nx/4, ny/2+2
r        = 10
cylinder = (x-cx)**2 + (y-cy)**2 < r**2
lattice[cylinder].defineDynamics(BounceBack(1.))

pout.display('Iterating')
snapshots = []
startTime = time.time()
for i in range(0,maxIter):
    if (i%statIter==0 and i>0):
        # Compute a velocity snapshot of the system, and keep it safe.
        snapshots.append(lattice.velocityNorm())
        # Apply a reduction operation to compute the average kinetic energy.
        pout.display('The energy is ', lattice.averageEnergy())
        # Use Python's timer to compute the efficiency of the computations.
        elapsed = time.time()-startTime
        Msups = i*nx*ny / elapsed / 1e6
        pout.display('Efficiency: ', Msups, ' Mega site updates per second.')

    # This is the real computation: evolving through a collision-streaming cycle.
    lattice.collideAndStream()

pout.display('Plotting results')
pout.imagesc(snapshots)
