#!/usr/bin/python

#  This file is part of the Palabos library.
#  Copyright (C) 2009-2010 Jonas Latt
#
#  The library Palabos is free software: you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.

import sys; sys.path.append('../src')
import time
from pypal import *

lx,ly,lz = 400,50,50       # Dimensions of the system in lattice units.
dx       = 1./lx           # Discrete space step (ly is the reference length).
nx,ny,nz = lx+1,ly+1,lz+1  # Lattice resolution.
rhoIn, rhoOut = 1.02, 1.   # Density at inlet/outlet.
tau      = 0.55            # Relaxation time.
maxIter  = 120           # Maximum number of iterations.
statIter = 20            # Intervals at which some statistics is printed.

# Allocate memory for the block-lattice, and assign the dynamics.
lattice = Block3D(nx, ny, nz, D3Q19, BGK(omega=1./tau))
pout.display(lattice.multiBlockInfo())

# Use a regularized boundary condition.
boundary.regularized().definePressureBC(lattice)
lattice[0,    1:ny-2, 1:nz-2].setBoundaryDensity(rhoIn)
lattice[nx-1, 1:ny-2, 1:nz-2].setBoundaryDensity(rhoOut)

# Here, x, y, and z are (parallelized) matrices. It's similar to Matlab's meshgrid.
x,y,z = lattice.meshGrid()
cy,cz = ny/2, nz/2
rTube = 23
tube  = (y-cy)**2 + (z-cz)**2 < rTube**2
lattice[tube.negate()].defineDynamics(BounceBack(1.))

cx    = nx/4
rCyl  = 4
cylinder = (x-cx)**2 + (y-cy-2)**2 < rCyl**2
lattice[cylinder].defineDynamics(BounceBack(1.))

lattice.initializeAtEquilibrium(1., [0.,0.,0.])

pout.display('Iterating from 0 to ', maxIter)
snapshots = []
startTime = time.time()
for i in range(0,maxIter+1):
    if (i%statIter==0 and i>0):
        # Compute a velocity snapshot of the system, and keep it safe.
        snapshots.append(lattice[:,:,cz].velocityNorm())
        # Apply a reduction operation to compute the average kinetic energy.
        pout.display('At iteration step ', i, ' the energy is ', lattice.averageEnergy())
        elapsed = time.time()-startTime
        Msups = i*nx*ny*nz / elapsed / 1e6
        pout.display('Efficiency: ', Msups, ' Mega site updates per second.')

    # This is the real computation: evolving through a collision-streaming cycle.
    lattice.collideAndStream()

pout.display('Plotting results')
velNorm = lattice.velocityNorm()
pout.isoSurface(velNorm, [velNorm.max()/10., velNorm.max()*2./3.])
pout.imagesc(snapshots)
