#!/usr/bin/python

#  This file is part of the Palabos library.
#  Copyright (C) 2009-2010 Jonas Latt
#
#  The library Palabos is free software: you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.

"""
Palabos example program: Flow through a porous media, driven by a pressure-
gradient. This simple 100-liner illustrates how to formulate relatively
advanced programs.  It includes the creation of the geometry of a porous
media, definition of boundary conditions, and evaluation of the result. And,
as usual, all steps (including the creation of the porous media) are fully
parallelized.
"""

import sys; sys.path.append('../src')
from pypal import *
import time
import random

# Initialize the random number generator.
random.seed(10)
# This convenience function generates a list with "num_val" random numbers
# comprized between 0 and max_val-1.
def rand_int(num_val, max_val):
    return [int(random.random()*max_val) for i in range(0,num_val)]

pout.display('Creating the lattice')
# This example uses a relatively small lattice, in order to get a result
# quickly on a single processor. On a parallel machine, don't hesitate to
# go for substantially larger systems.
nx, ny, nz = 200, 100, 100
# The integer-valued block "media" shall hold a description of the porous media.
media = Block3D(nx,ny,nz, 0)

x,y,z = media.meshGrid()
pout.display('Creating the porous media')
# The media consists of 30 spheres with radius 8 around a random position.
nobst = 5 
r     = 14
for cx,cy,cz in zip(rand_int(nobst,nx), rand_int(nobst,ny), rand_int(nobst,nz)):
    pout.display('  Creating a sphere centered at [', cx, cy, cz, ']')
    media += (x-cx)**2 + (y-cy)**2 + (z-cz)**2 < r**2
pout.display('Displaying the porous media'\
             '(close the window to proceed with program execution)')
# A boolean operation is used to convert the integer values in "media" to
# a zero-or-one representation.
pout.isoSurface(media>0, [1.])

# A relaxation parameter tau=0.6 is used here for illustration purposes. In
# a real simulation, this should preferrably be computed from the fluid
# viscosity, after proper unit version.
lattice = Block3D(nx,ny,nz, D3Q19, BGK(omega=1/0.6))
lattice.initializeAtEquilibrium(1., [0.,0.,0.])

# Set a pressure boundary condition at inlet and outlet, with a larger
# pressure for the inlet (remember that density is proportional to pressure).
boundary.regularized().definePressureBC(lattice[0,   1:ny-2,1:nz-2])
boundary.regularized().definePressureBC(lattice[nx-1,1:ny-2,1:nz-2])
lattice[0,   0:ny-1,0:nz-1].setBoundaryDensity(1.02)
lattice[nx-1,0:ny-1,0:nz-1].setBoundaryDensity(1.)

# Set bounce-back dynamics on all nodes on which "media" is on-zero.
lattice[media].defineDynamics(BounceBack(1.))

# Simulate 2000 steps, and print statistics to the terminal every 200th step.
nsteps, nvisual = 200, 20
pout.display('Starting simulation (', nsteps, 'time steps )')
startTime = time.time()
for i in range(1,nsteps):
    if (i%nvisual==0 and i>0):
        elapsed = time.time()-startTime
        Msups = i*nx*ny*nz / elapsed / 1e6
        pout.display( float(i)/float(nsteps)*100, '% performed. Efficiency: ',
                      Msups, ' Mega site updates per second.' )
        pout.display('The energy is ', lattice.averageEnergy())
    lattice.collideAndStream()

# Compute the flow throught the media on two selected y-z planes. The flow
# must be computed on fluid-nodes only (excluding the spherical obstacles),
# and it is based on the x-component of the velocity, normal to the plane.
flow1 = lattice[media==0][nx/3,:,:].velocityComponent(0).average()
flow2 = lattice[media==0][2*nx/3,:,:].velocityComponent(0).average()

# Note that the two values should be the same (up to the accuracy of the
# numerical scheme), because the flow is incompressible. The difference
# between the two values can indicate that the flow is still in an initial
# transient state, or is under-resolved, or has a too large Mach number,
# leading to compressibility effects.
pout.display( 'The average flow at 1/2 distance is ', flow1,
              ' and at 3/4 distance: ', flow2 )


# Display two snapshots of the (norm of the) strain-rate. On the first image,
# the strain rate is evaluated with second-order accuracy from the velocity-
# gradients, using a finite difference scheme. On the second image, it is
# evaluated from the stress tensor, using the particle populations. The second
# approach is first-order accurate, and the image obtained by this approach
# has visibly lower quality.
pout.imagesc( \
        [ lattice.velocity().strainRate().symmetricTensorNorm()[:,:,20],
          lattice.strainRateFromStress().symmetricTensorNorm()[:,:,20] ] )
