#  This file is part of the Palabos library.
#  Copyright (C) 2009-2010 Jonas Latt
#  E-mail contact: jonas@lbmethod.org
#  The most recent release of Palabos can be downloaded at 
#  <http://www.lbmethod.org/palabos/>
# 
#  The library Palabos is free software: you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
# 
#  The library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# 
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
This is the central module of the Palabos-Python binding. All Palabos container
objects (block-lattice, scalar-field, tensor-field) are instantiated by calling
the functions Block2D or Block3D.
"""

# MEMO
# To access the SWIG-wrapped Palabos stuff, one needs a _plb handler. There are
# many handlers: one for d2q9-float, one for d3q19-double, one for block-int,
# etc. A handler can be generated automatically from an interface; an interface
# is generated in the datatype module:
#   interface=datatype.block_interface(Float)
#   interface=datatype.lattice_interface(D2Q9, Float)
# From an interface, you get
# - The _plb handler: _plb = interface.library
# - The data type: dtype = interface.dtype
# - The data type for communication with SWIG: dtype = interface.swig_dtype

__all__ = ['Block2D', 'Block3D']

import plb
import numpy
import ppexcept
import datatype
import dynamics
import util
import lb

def Block2D(*args):
    """
    Generate any of the 2D Palabos container objects (block-lattice, scalar-field,
    tensor-field).

    For scalar- or tensor-fields, the first two arguments specify the dimensions,
    and the third argument a constant initial value for all elements.
    Syntax:
    >>> scalarField = Block2D(100, 100, 0.)
    >>> vectorField = Block2D(100, 100, [0., 0.])
    Note that the vector- (or tensor-) field can have an arbitrary number of
    per-cell elements, depending on the size of the array provided as third
    argument. The third argument can be any iterable object.

    For a block-lattice, the lattice structure, and the initial dynamics must be
    provided. Remember that the dynamics is initialized with a relaxation parameter
    omega.
    Syntax:
    >>> lattice = Block2D(100, 100, D2Q9, BGK(omega=1.))

    Note that all blocks can be valued to single- or double-precision floats, or to
    integer for the scalar- and tensor-fields.

    For further information, type 'help RawBlock' or 'help BlockLattice'
    """
    if (hasattr(args[0], '__iter__')):
        domain = plb.core.Box2D( args[0][0], args[0][1],
                                 args[0][2], args[0][3] )
        numSizeArg=1
    else:
        domain = plb.core.Box2D(0, args[0]-1, 0, args[1]-1)
        numSizeArg=2

    if (len(args)==numSizeArg+1):
        return RawBlock2D(domain, args[numSizeArg])
    else:
        return BlockLattice2D(domain, args[numSizeArg], args[numSizeArg+1])


def Block3D(*args):
    """
    Generate any of the 3D Palabos container objects (block-lattice, scalar-field,
    tensor-field).

    For scalar- or tensor-fields, the first two arguments specify the dimensions,
    and the third argument a constant initial value for all elements.
    Syntax:
    >>> scalarField = Block3D(100, 100, 100, 0.)
    >>> vectorField = Block3D(100, 100, 100, [0., 0.])
    Note that the vector- (or tensor-) field can have an arbitrary number of
    per-cell elements, depending on the size of the array provided as third
    argument. The third argument can be any iterable object.

    For a block-lattice, the lattice structure, and the initial dynamics must be
    provided. Remember that the dynamics is initialized with a relaxation parameter
    omega.
    Syntax:
    >>> lattice = Block3D(100, 100, 100, D3Q19, BGK(omega=1.))

    Note that all blocks can be valued to single- or double-precision floats, or to
    integer for the scalar- and tensor-fields.

    For further information, type 'help RawBlock' or 'help BlockLattice'
    """
    if (hasattr(args[0], '__iter__')):
        domain = plb.core.Box3D( args[0][0], args[0][1],
                                 args[0][2], args[0][3],
                                 args[0][4], args[0][5] )
        numSizeArg=1
    else:
        domain = plb.core.Box3D(0, args[0]-1, 0, args[1]-1, 0, args[2]-1)
        numSizeArg=3

    if (len(args)==numSizeArg+1):
        return RawBlock3D(domain, args[numSizeArg])
    else:
        return BlockLattice3D(domain, args[numSizeArg], args[numSizeArg+1])


class BlockBase2D:
    """
    All stuff about blocks which is explicitly 2D is put into this base class.
    There's not a lot, because the Python interface is pretty generic about
    2D vs. 3D.
    """
    def multiBlockInfo(self):
        """ Print some info on the internal multi-block structure of the block. """
        info = self._plb().PlbMultiBlockInfo2D(self._plb_block)
        nx = info.getNx()
        ny = info.getNy()
        smallest = info.getSmallestBlock()
        largest  = info.getLargestBlock()
        millionAllocCells = info.getNumAllocatedCells()/1.e6
        cellPercentage    = float(info.getNumAllocatedCells())/float(nx*ny)*100
        return "Internal structure of the " + str(nx) + "-by-" + str(ny) + " multi-block:" + \
               "\nNumber of blocks in multi-block:" + str(info.getNumBlocks()) + \
               "\nSmallest atomic-block: [" + str(smallest.x0) + "," + str(smallest.x1) + ", " + \
                                              str(smallest.y0) + "," + str(smallest.y1) + "]" + \
               "\nLargest atomic-block: [" + str(largest.x0) + "," + str(largest.x1) + ", " + \
                                             str(largest.y0) + "," + str(largest.y1) + "]" + \
               "\nNumber of allocated cells: " + str(millionAllocCells) + " million" + \
               "\nPercentage of allocated cells in multi-block: " + str(cellPercentage)
    def nx(self):
        return self._domain.getNx()
    def ny(self):
        return self._domain.getNy()
    def meshGrid(self):
        plb_x = self._plb().generateIntNTensorField2D(self._plb_block, self._domain, 1)
        x = RawBlock2D(plb_x, datatype.block_interface(numpy.int32), self._domain, self._mask)
        x._interface.library.plbSetToCoordinate(plb_x, self._domain, 0)
        plb_y = self._plb().generateIntNTensorField2D(self._plb_block, self._domain, 1)
        y = RawBlock2D(plb_y, datatype.block_interface(numpy.int32), self._domain, self._mask)
        y._interface.library.plbSetToCoordinate(plb_y, self._domain, 1)
        return x,y
    def generateMultiNTensorField(self, domain, ndim):
        return self._plb().generateMultiNTensorField2D(domain, ndim)
    def NTensorFieldSerializer(self):
        return self._plb().NTensorFieldSerializer2D(self._plb_block, self._domain)
    def NTensorFieldUnSerializer(self):
        return self._plb().NTensorFieldUnSerializer2D(self._plb_block, self._domain)


class BlockBase3D:
    """
    All stuff about blocks which is explicitly 3D is put into this base class.
    There's not a lot, because the Python interface is pretty generic about
    2D vs. 3D.
    """
    def multiBlockInfo(self):
        """ Print some info on the internal multi-block structure of the block. """
        info = self._plb().PlbMultiBlockInfo3D(self._plb_block)
        nx = info.getNx()
        ny = info.getNy()
        nz = info.getNz()
        smallest = info.getSmallestBlock()
        largest  = info.getLargestBlock()
        millionAllocCells = info.getNumAllocatedCells()/1.e6
        cellPercentage    = float(info.getNumAllocatedCells())/float(nx*ny*nz)*100
        return "Internal structure of the " + str(nx) + "-by-" + str(ny) + "-by-" + str(nz) + " multi-block:" + \
               "\nNumber of blocks in multi-block:" + str(info.getNumBlocks()) + \
               "\nSmallest atomic-block: [" + str(smallest.x0) + "," + str(smallest.x1) + ", " + \
                                              str(smallest.y0) + "," + str(smallest.y1) + ", " + \
                                              str(smallest.z0) + "," + str(smallest.z1) + "]" + \
               "\nLargest atomic-block: [" + str(largest.x0) + "," + str(largest.x1) + ", " + \
                                             str(largest.y0) + "," + str(largest.y1) + ", " + \
                                             str(largest.z0) + "," + str(largest.z1) + "]" + \
               "\nNumber of allocated cells: " + str(millionAllocCells) + " million" + \
               "\nPercentage of allocated cells in multi-block: " + str(cellPercentage)
    def nx(self):
        return self._domain.getNx()
    def ny(self):
        return self._domain.getNy()
    def nz(self):
        return self._domain.getNz()
    def meshGrid(self):
        plb_x = self._plb().generateIntNTensorField3D(self._plb_block, self._domain, 1)
        x = RawBlock3D(plb_x, datatype.block_interface(numpy.int32), self._domain, self._mask)
        x._interface.library.plbSetToCoordinate(plb_x, self._domain, 0)
        plb_y = self._plb().generateIntNTensorField3D(self._plb_block, self._domain, 1)
        y = RawBlock3D(plb_y, datatype.block_interface(numpy.int32), self._domain, self._mask)
        y._interface.library.plbSetToCoordinate(plb_y, self._domain, 1)
        plb_z = self._plb().generateIntNTensorField3D(self._plb_block, self._domain, 1)
        z = RawBlock3D(plb_z, datatype.block_interface(numpy.int32), self._domain, self._mask)
        z._interface.library.plbSetToCoordinate(plb_z, self._domain, 2)
        return x,y,z
    def generateMultiNTensorField(self, domain, ndim):
        return self._plb().generateMultiNTensorField3D(domain, ndim)
    def NTensorFieldSerializer(self):
        return self._plb().NTensorFieldSerializer3D(self._plb_block, self._domain)
    def NTensorFieldUnSerializer(self):
        return self._plb().NTensorFieldUnSerializer3D(self._plb_block, self._domain)


class BlockBase:
    """
    Common operations for all blocks. This is mostly about managing sub-
    domains, slices and masks.
    Constructors:
    BlockBase(plb.core.Box2D(0,100,0,100))
    BlockBase(plb_block, interface, domain, mask)
    """
    def __init__(self, *args):
        if (hasattr(args[0], 'x0')): # Construction of new block.
            self._isrestricted = False
            self._domain = args[0]
            self._mask = None
        else: # Construction of a reference-to-block with restricted domain or mask.
            self._isrestricted = True
            self._plb_block = args[0]
            self._interface = args[1]
            self._domain = args[2]
            self._mask = args[3]
    def initialize(self):
        self._plb_block.initialize()
    def domain(self):
        return boxToList(self._domain)
    def dtype(self):
        return self._interface.dtype
    def _plb(self):
        return self._interface.library
    def __getitem__(self, sliceArg):
        if (hasattr(sliceArg, '_mask')): # sliceArg is a Block.
            assert sliceArg.dtype() == numpy.int32, \
                   'A bool mask must be integer valued.'
            return self.__class__ (
                    self._plb_block, self._interface,
                    self._domain, _combineMasks(self._mask, sliceArg ) )
        else: # sliceArg is a range.
            return self.__class__ (
                    self._plb_block, self._interface,
                    util.slice_to_box(self._domain, *sliceArg),
                    self._mask )


class BlockLattice (BlockBase):
    """
    Not an end-user class. The end-user class must also inherit from a 2D-
    or 3D- class to get the appropriate interface.
    Constructors:
    BlockLattice(plb.core.Box2D(0,100,0,100), D2Q9, BGK(1.))
    BlockLattice(plb_block, interface, domain, mask)
    """
    def __init__(self, *args):
        BlockBase.__init__(self, *args)
        if (hasattr(args[0], 'x0')):
            lattice = args[1]
            dynamics = args[2]
            self._interface = dynamics.interface(lattice)
            self._plb_block = \
                self._plb().generateMultiBlockLattice ( \
                        self._domain, dynamics.plb_dynamics(lattice) )

    def collideAndStream(self):
        """
        Take all selected nodes through an LB iteration (collision-
        streaming cycle. Attention: the domain can only be restriced
        by ranges as follows:
        lattice[:,:,10].collideAndStream()
        but not by a logical mask as in the following:
        lattice[x>0].collideAndStream()
        A logical mask, if specified, is ignored.
        """
        if self._isrestricted:
            self._plb_block.collideAndStream(self._domain)
        else:
            self._plb_block.collideAndStream()

    def defineDynamics(self, dynamics):
        """
        Assign a new dynamics (i.e. a collision rule) to cells.
        Examples:
        lattice.defineDynamics(BGK(omega=1.))
        lattice.defineDynamics(BounceBack(1.))
        """
        lattice = self._interface.lattice
        if self._mask is None: # Unmasked operation.
            self._plb().plbDefineDynamics(self._plb_block, self._domain,
                                          dynamics.plb_dynamics(lattice))
        else: # Masked operation.
            self._plb().m_plbDefineDynamics(self._plb_block, self._mask._plb_block,
                                            self._domain, dynamics.plb_dynamics(lattice))

    def setBoundaryDensity(self, density):
        self._plb().plbSetBoundaryDensity(self._plb_block, self._domain,
                                          self._interface.swig_dtype(density))

    def averageEnergy(self):
        return self._plb().averageEnergy(self._plb_block, self._domain)

    def averageDensity(self):
        return self._plb().averageDensity(self._plb_block, self._domain)

    def setBoundaryVelocity(self, velocity):
        self._inplaceOperationOneArg("plbSetBoundaryVelocity",velocity)

    def initializeAtEquilibrium(self, density, velocity):
        self._inplaceOperationTwoArgs("plbInitializeAtEquilibrium", density, velocity)

    def setPopulations(self, populations):
        self._inplaceOperationOneArg("plbSetPopulations", populations)

    def density(self):
        return self._analysisOperation("density")

    def velocityNorm(self):
        return self._analysisOperation("velocityNorm")

    def kineticEnergy(self):
        return self._analysisOperation("kineticEnergy")

    def velocity(self):
        return self._analysisOperation("velocity")

    def deviatoricStress(self):
        return self._analysisOperation("deviatoricStress")

    def strainRateFromStress(self):
        return self._analysisOperation("strainRateFromStress")

    def velocityComponent(self, whichComponent):
        return self._analysisOperationWithArg("velocityComponent", whichComponent)

    def getPopulation(self, whichPopulation):
        return self._analysisOperationWithArg("population", whichPopulation)

    def getPopulations(self):
        return self._analysisOperation("populations")

    def to_array(self):
        """ Convert the content into a non-parallel Numpy array, for further
            serial processing (for example visualization). """
        serializer = self._plb().LatticeSerializer(self._plb_block, self._domain)
        fullSize = serializer.getSize()
        serialized_array = numpy.zeros(fullSize, self._interface.dtype)
        serializer.execute(serialized_array);
        util.reshape_ndarray(serialized_array, self._domain, self._interface.lattice.q)
        return serialized_array.squeeze() # Eliminate degenerate dimensions.

    def from_array(self, serial_array):
        """ Copy the content of a non-parallel Numpy array into a block,
            or into a sub-domain of the block. """
        util.assert_ndarray_equals(serial_array, self._domain, self._interface.lattice.q)
        unserializer = self._plb().LatticeUnSerializer(self._plb_block, self._domain)
        unserializer.execute(serial_array.astype(self.dtype()).flatten())

    def info(self):
        print 'Internal data type: ' + repr(self.dtype())
        print 'Accessing library: ' + repr(self._plb())

    def _arrayOrTensorFieldOrScalar(self, argument):
        if hasattr(argument, '__iter__'):  # Array
            return numpy.array(argument,self.dtype())
        else:
            try:
                return argument._plb_block # Tensor-field
            except:
                return argument            # Scalar

    def _analysisOperation(self, operation):
        if self._mask is None: # Unmasked operation.
            plb_field = getattr(self._plb(),operation)(self._plb_block, self._domain)
        else: # Masked operation.
            plb_field = getattr(self._plb(),'m_'+operation)(self._plb_block,
                                                            self._mask._plb_block, self._domain)
        return self._myRawBlock()(plb_field, datatype.block_interface(self.dtype()), self._domain, self._mask)

    def _analysisOperationWithArg(self, operation, arg):
        if self._mask is None: # Unmasked operation.
            plb_field = getattr(self._plb(),operation)(self._plb_block, self._domain, arg)
        else: # Masked operation.
            plb_field = getattr(self._plb(),'m_'+operation)(self._plb_block, self._mask._plb_block,
                                                            self._domain, arg)
        return self._myRawBlock()(plb_field, datatype.block_interface(self.dtype()), self._domain, self._mask)

    def _inplaceOperationOneArg(self, operation, arg):
        try: # Arguments are blocks.
            inters_domain = util.intersect(self._domain, arg._domain)
            common_mask = _combineMasks(self._mask, arg._mask)
        except: # Arguments are scalars.
            inters_domain = self._domain
            common_mask = self._mask
        if common_mask is None: # Unmasked operation.
            getattr(self._plb(),operation)(self._plb_block, self._arrayOrTensorFieldOrScalar(arg), inters_domain)
        else: # Masked operation.
            getattr(self._plb(),'m_'+operation)(self._plb_block, self._arrayOrTensorFieldOrScalar(arg),
                                                common_mask._plb_block, inters_domain)

    def _inplaceOperationTwoArgs(self, operation, arg1, arg2):
        try: # Arguments are blocks.
            inters_domain = util.intersect(self._domain, util.intersect(arg1._domain,arg2._domain))
            common_mask = _combineMasks(self._mask, _combineMasks(arg1._mask, arg2._mask))
        except: # Arguments are scalars.
            inters_domain = self._domain
            common_mask = self._mask
        if common_mask is None: # Unmasked operation.
            getattr(self._plb(),operation)(self._plb_block, self._arrayOrTensorFieldOrScalar(arg1),
                                           self._arrayOrTensorFieldOrScalar(arg2), inters_domain)
        else: # Masked operation.
            getattr(self._plb(),'m_'+operation)(self._plb_block, self._arrayOrTensorFieldOrScalar(arg1),
                                                self._arrayOrTensorFieldOrScalar(arg2),
                                                common_mask._plb_block, inters_domain)
    def _myRawBlock(self):
        if self._interface.lattice.d==2:
            return RawBlock2D
        elif self._interface.lattice.d==3:
            return RawBlock3D


class RawBlock (BlockBase):
    """
    The RawBlock stands for either scalar- or tensor-fields. This is not an
    end-user class, because one also needs to inherit a 2D- or 3D- interface
    for the whole thing to become useful.
    Constructors:
    RawBlock(plb.core.Box2D(0,100,0,100), [0., 0.])
    RawBlock(plb_block, interface, domain, mask)
    """
    def __init__(self, *args):
        BlockBase.__init__(self, *args)
        if (hasattr(args[0], 'x0')): # First constructor:
                                     # Construct a new raw-block
            if hasattr(args[1], '__iter__'):
                ini_val = numpy.array(args[1])
            else:
                ini_val = numpy.array([args[1]])
            ndim = ini_val.size
            if ndim<1:
                raise ppexcept.PyPalConstructorError()
            self._interface = datatype.block_interface(type(ini_val[0]))
            self._plb_block = self.generateMultiNTensorField(self._domain, ndim)
            self.assign(numpy.array(ini_val, self._interface.swig_dtype))

    def convert(self, dtype):
        """
        Return a clone of the block, where the clone has a different data type.
        Whatever data type you provide, it will be somehow interpreted as either
        numpy.float32, numpy.float64, or numpy.int32.
        Examples:
        b = a.convert(float)
        b = a.convert(numpy.float64)
        b = a.convert(type(1.))
        """
        conv_interface = datatype.block_interface(dtype)
        if conv_interface.dtype==numpy.int32:
            new_field = self._plb().plbCopyConvertInt(self._plb_block, self._domain)
        elif conv_interface.dtype==numpy.float32:
            new_field = self._plb().plbCopyConvertFloat(self._plb_block, self._domain)
        elif conv_interface.dtype==numpy.float64:
            new_field = self._plb().plbCopyConvertDouble(self._plb_block, self._domain)
        return self.__class__(new_field, conv_interface, self._domain, self._mask)

    def ndim_elements(self):
        """ Number of components of each element of the block. If your block is
            a scalar-field, this function returns 1, for example. """
        return self._plb_block.getNdim()

    def assign(self, other):
        """ Assign all data from a raw-block to another. """
        return self._inplace_operation(other, 'iassign')

    def average(self):
        """ Computes the average over spatial dimensions, and returns a scalar
            or an array, depending on whether the dimension of the field elements
            is 1 or more. """
        return self._reduction_operation("plbAverage")

    def boundedAverage(self):
        """ Computes a bounded average over spatial dimensions (the edges, corners, etc.
            are weighted in order to mimick a second-order integration over space with
            the trapezium rule), and returns a scalar or an array, depending on whether
            the dimension of the field elements is 1 or more. """
        return self._reduction_operation("plbBoundedAverage")

    def min(self):
        """ Computes the minimum over spatial dimensions, and returns a scalar
            or an array, depending on whether the dimension of the field elements
            is 1 or more. """
        return self._reduction_operation("plbMin")

    def max(self):
        """ Computes the maximum over spatial dimensions, and returns a scalar
            or an array, depending on whether the dimension of the field elements
            is 1 or more. """
        return self._reduction_operation("plbMax")

    def __add__(self, other):
        """ Evaluate a+b, where a is a raw-block, and b is either a raw-block
            or a scalar (resp. array if ndim_elements()>1) """
        return self._arithmetic_operation(other, 'add')

    def __radd__(self, other):
        """ Evaluate a+b, where a is scalar (resp. array if ndim_elements()>1),
            and b is a raw-block. """
        return self._right_arithmetic_operation(other, 'add')

    def __iadd__(self, other):
        """ In-place element-wise add values to the current raw-block. The
            argument is either a raw-block or a scalar (resp. array if
            ndim_elements()>1) """
        return self._inplace_operation(other, 'iadd')

    def __sub__(self, other):
        """ Evaluate a-b, where a is a raw-block, and b is either a raw-block
            or a scalar (resp. array if ndim_elements()>1) """
        return self._arithmetic_operation(other, 'subtract')

    def __rsub__(self, other):
        """ Evaluate a-b, where a is scalar (resp. array if ndim_elements()>1),
            and b is a raw-block. """
        return self._right_arithmetic_operation(other, 'subtract')

    def __isub__(self, other):
        """ In-place element-wise subtract values from the current raw-block. The
            argument is either a raw-block or a scalar (resp. array if
            ndim_elements()>1) """
        return self._inplace_operation(other, 'isubtract')

    def __mul__(self, other):
        """ Evaluate a*b, where a is a raw-block, and b is either a raw-block
            or a scalar (resp. array if ndim_elements()>1) """
        return self._arithmetic_operation(other, 'multiply')

    def __rmul__(self, other):
        """ Evaluate a*b, where a is scalar (resp. array if ndim_elements()>1),
            and b is a raw-block. """
        return self._right_arithmetic_operation(other, 'multiply')

    def __imul__(self, other):
        """ In-place element-wise multiply values against the current raw-block. The
            argument is either a raw-block or a scalar (resp. array if
            ndim_elements()>1) """
        return self._inplace_operation(other, 'multiply')

    def __div__(self, other):
        """ Evaluate a/b, where a is a raw-block, and b is either a raw-block
            or a scalar (resp. array if ndim_elements()>1) """
        return self._arithmetic_operation(other, 'divide')

    def __rdiv__(self, other):
        """ Evaluate a/b, where a is scalar (resp. array if ndim_elements()>1),
            and b is a raw-block. """
        return self._right_arithmetic_operation(other, 'divide')

    def __idiv__(self, other):
        """ In-place element-wise divide values from the current raw-block. The
            argument is either a raw-block or a scalar (resp. array if
            ndim_elements()>1) """
        return self._inplace_operation(other, 'divide')

    def __pow__(self, other):
        """ Evaluate a-to-the-power-b, where a is a raw-block, and b is either a raw-block
            or a scalar (resp. array if ndim_elements()>1).
            Remember that in Python this overloads the operator ** ."""
        return self._arithmetic_operation(other, 'toThePower')

    def __rpow__(self, other):
        """ Evaluate a-to-the-power-b, where a is scalar (resp. array if ndim_elements()>1),
            and b is a raw-block.
            Remember that in Python this overloads the operator ** ."""
        return self._right_arithmetic_operation(other, 'toThePower')

    def __ipow__(self, other):
        """ In-place element-wise evaluate to-the-power in current raw-block. The
            argument is either a raw-block or a scalar (resp. array if
            ndim_elements()>1)
            Remember that in Python this overloads the operator **= ."""
        return self._inplace_operation(other, 'toThePower')

    def __eq__(self, other):
        """ Evaluate boolean a==b, where a is a raw-block, and b is either a raw-block
            or a scalar (resp. array if ndim_elements()>1) """
        return self._bool_operation(other, 'equals')

    def __ne__(self, other):
        """ Evaluate boolean a!=b, where a is a raw-block, and b is either a raw-block
            or a scalar (resp. array if ndim_elements()>1) """
        return self._bool_operation(other, 'equals').negate()

    def __lt__(self, other):
        """ Evaluate boolean a<b, where a is a raw-block, and b is either a raw-block
            or a scalar (resp. array if ndim_elements()>1) """
        return self._bool_operation(other, 'lessThan')

    def __le__(self, other):
        """ Evaluate boolean a<=b, where a is a raw-block, and b is either a raw-block
            or a scalar (resp. array if ndim_elements()>1) """
        return self._bool_operation(other, 'lessEqual')

    def __gt__(self, other):
        """ Evaluate boolean a>b, where a is a raw-block, and b is either a raw-block
            or a scalar (resp. array if ndim_elements()>1) """
        return self._bool_operation(other, 'greaterThan')

    def __ge__(self, other):
        """ Evaluate boolean a>=b, where a is a raw-block, and b is either a raw-block
            or a scalar (resp. array if ndim_elements()>1) """
        return self._bool_operation(other, 'greaterEqual')

    def __and__(self, other):
        """ Evaluate boolean a && b, where a is a raw-block, and b is either a raw-block
            or a scalar (resp. array if ndim_elements()>1).
            Attention: it's the binary-and "&" which is being overloaded here, because
            the logical-and "&&" cannot be overloaded in Python. Be careful,
            because the binary-and has different precedence rules (per-use parentheses
            to be safe). """
        return self._bool_operation(other, 'logicalAnd')

    def __or__(self, other):
        """ Evaluate boolean a || b, where a is a raw-block, and b is either a raw-block
            or a scalar (resp. array if ndim_elements()>1).
            Attention: it's the binary-or "|" which is being overloaded here, because
            the logical-or "|| cannot be overloaded in Python. Be careful,
            because the binary-and has different precedence rules (per-use parentheses
            to be safe). """
        return self._bool_operation(other, 'logicalOr')

    def negate(self):
        """ Evaluate boolean !a operation. """
        if self._mask is None: # Unmasked operation.
            new_field = self._plb().negate (
                    self._plb_block, self._domain )
        else: # Masked operation.
            new_field = self._plb().m_negate (
                    self._plb_block, self._mask._plb_block, self._domain )
        interface = datatype.block_interface(int)
        return self.__class__(new_field, interface, self._domain, self._mask)

    def getComponent(self, component):
        """ Extract one component and put result into a 1-element scalar field.
            The spatial dimensions of the result are the same as the ones of
            the original field. """
        if self._mask is None: # Unmasked operation.
            new_field = self._plb().plbExtractComponent (
                    self._plb_block, self._domain, component )
        else: # Masked operation.
            new_field = self._plb().m_plbExtractComponent (
                    self._plb_block, self._mask._plb_block, self._domain, component )
        return self.__class__(new_field, self._interface, self._domain, self._mask)

    def setComponent(self, component, other):
        """ Assign values to one of the components. The argument 'other' is
            either a scalar or a 1-component raw-block. """
        return self._inplace_operation_with_arg(other, 'plbAssignComponent', component)

    def to_coordinate(self, coordinate):
        """ Assuming that self is a 1-component scalar-field, initialize all
            elements to their lattice-units coordinate value. """
        if self._mask is None: # Unmasked operation.
            self._plb().plbSetToCoordinate (
                    self._plb_block, self._domain, coordinate )
        else: # Masked operation.
            self._plb().m_plbSetToCoordinate (
                    self._plb_block, self._mask._plb_block, self._domain, coordinate )
        return self

    def norm(self):
        """
        At each element, evaluate the norm of the tensor, and return a scalar-field.
        """
        return self._analysis_operation('plbComputeNorm')

    def normSqr(self):
        """
        At each element, evaluate the square-norm of the tensor, and return a scalar-field.
        """
        return self._analysis_operation('plbComputeNormSqr')

    def symmetricTensorNorm(self):
        """
        Assuming that each element of self holds the upper-diagonal elements of a
        symmetric tensor, compute its tensor norm.
        """
        return self._analysis_operation('plbComputeSymmetricTensorNorm')

    def symmetricTensorNormSqr(self):
        """
        Assuming that each element of self holds the upper-diagonal elements of a
        symmetric tensor, compute the square of its tensor norm.
        """
        return self._analysis_operation('plbComputeSymmetricTensorNormSqr')

    def symmetricTensorTrace(self):
        """
        Assuming that each element of self holds the upper-diagonal elements of a
        symmetric tensor, compute the square of its trace.
        """
        return self._analysis_operation('plbComputeSymmetricTensorTrace')

    def vorticity(self):
        """
        Assuming that each element of self holds an n-dimensional vector (where
        n is 2 or 3, and is equal to the number of space dimensions), compute
        the vorticity (scalar in 2D, vector in 3D).
        """
        return self._analysis_operation('plbComputeVorticity')

    def strainRate(self):
        """
        Assuming that each element of self holds an n-dimensional vector (where
        n is 2 or 3, and is equal to the number of space dimensions), compute
        the symmetric strain-rate tensor.
        """
        return self._analysis_operation('plbComputeStrainRate')

    def to_array(self):
        """ Convert the content into a non-parallel Numpy array, for further
            serial processing (for example visualization). """
        serializer = self.NTensorFieldSerializer()
        fullSize = serializer.getSize()
        serialized_array = numpy.zeros(fullSize, self._interface.dtype)
        serializer.execute(serialized_array)
        util.reshape_ndarray(serialized_array, self._domain, self.ndim_elements())
        return serialized_array.squeeze() # Eliminate degenerate dimensions.

    def from_array(self, serial_array):
        """ Copy the content of a non-parallel Numpy array into a block,
            or into a sub-domain of the block. """
        util.assert_ndarray_equals(serial_array, self._domain, self.ndim_elements())
        unserializer = self.NTensorFieldUnSerializer()
        unserializer.execute(serial_array.astype(self.dtype()).flatten())

    def __setitem__(self, sliceArg, rhs):
        """
        This is a Python thingy which is not obvious to everybody.
        Usually, an expression like "a=b", means "a shall be a reference to
        the same object as b", and nothing is being copied. However, in
        presence of slices, the "=" suddenly becomes an assignment operator:
        a[:,:]=b means that all content from b is copied over to a.
        """
        self.__getitem__(sliceArg).assign(rhs)

    def _numeric_arg(self, value):
        if hasattr(value, '__iter__'):
            arg = numpy.array(value, self.dtype())
        else:
            arg = numpy.array([value], self.dtype())
        return arg

    def _analysis_operation(self, operation):
        if self._mask is None: # Unmasked operation.
            new_field = getattr(self._plb(),operation) (
                    self._plb_block, self._domain )
        else: # Masked operation.
            new_field = getattr(self._plb(),'m_'+operation) (
                    self._plb_block, self._mask._plb_block, self._domain )
        return self.__class__(new_field, self._interface,
                              self._domain, self._mask)

    def _reduction_operation(self, operation):
        reducted_value = numpy.zeros(self.ndim_elements(), self._interface.dtype)
        if self._mask is None: # Unmasked operation.
            getattr(self._plb(),operation) (
                self._plb_block, self._domain, reducted_value )
        else: # Masked operation:
            getattr(self._plb(),'m_'+operation) (
                self._plb_block, self._mask._plb_block, self._domain, reducted_value )
        if len(reducted_value)==1: # In case of a scalar-field, return
                                   # a scalar value.
            return reducted_value[0]
        else:
            return reducted_value

    def _arithmetic_operation(self, other, operation):
        if isinstance(other, self.__class__):
            first, second = _type_adjust_two_blocks(self, other)
            common_domain = util.intersect(first._domain, second._domain)
            common_mask = _combineMasks(first._mask, second._mask)
            if common_mask is None: # Unmasked operation.
                new_field = getattr(first._plb(),operation) (
                        first._plb_block, second._plb_block, common_domain )
            else: # Masked operation.
                new_field = getattr(first._plb(),'m_'+operation) (
                        first._plb_block, second._plb_block,
                        common_mask._plb_block, common_domain )
            return self.__class__(new_field, first._interface,
                                  common_domain, common_mask)
        else:
            block = _type_adjust_block_on_scalar(self, other)
            arg = block._numeric_arg(other)
            if self._mask is None: # Unmasked operation.
                new_field = getattr(block._plb(),operation) (
                        block._plb_block, arg, block._domain )
            else: # Masked operation.
                new_field = getattr(block._plb(),'m_'+operation) (
                        block._plb_block, arg,
                        self._mask._plb_block, block._domain )
            return self.__class__(new_field, block._interface,
                                  block._domain, self._mask)

    def _right_arithmetic_operation(self, other, operation):
        if isinstance(other, self.__class__):
            first, second = _type_adjust_two_blocks(self, other)
            common_domain = util.intersect(first._domain, second._domain)
            common_mask = _combineMasks(first._mask, second._mask)
            if common_mask is None: # Unmasked operation.
                new_field = getattr(first._plb(),operation) (
                        second._plb_block, first._plb_block, common_domain )
            else: # Masked operation.
                new_field = getattr(first._plb(),'m_'+operation) (
                        second._plb_block, first._plb_block,
                        common_mask._plb_block, common_domain )
            return self.__class__(new_field, first._interface, common_domain, common_mask)
        else:
            block = _type_adjust_block_on_scalar(self, other)
            arg = block._numeric_arg(other)
            if self._mask is None: # Unmasked operation.
                new_field = getattr(block._plb(),operation) (
                        block._plb_block, arg, block._domain )
            else: # Masked operation.
                new_field = getattr(block._plb(),'m_'+operation) (
                        block._plb_block, arg,
                        self._mask._plb_block, block._domain )
            return self.__class__(new_field, block._interface, block._domain, self._mask)

    def _inplace_operation(self, other, operation):
        if isinstance(other, self.__class__):
            second = _type_adjust_block(self.dtype(), other)
            common_domain = util.intersect(self._domain, second._domain)
            common_mask = _combineMasks(self._mask, second._mask)
            if common_mask is None: # Unmasked operation.
                getattr(self._plb(),operation) (
                           self._plb_block, second._plb_block,
                           common_domain )
            else: # Masked operation.
                getattr(self._plb(),'m_'+operation) (
                           self._plb_block, second._plb_block,
                           common_mask._plb_block, common_domain )
        else:
            if self._mask is None: # Unmasked operation.
                getattr(self._plb(),operation) (
                          self._plb_block, self._numeric_arg(other), self._domain )
            else: # Masked operation.
                getattr(self._plb(),'m_'+operation) (
                          self._plb_block, self._numeric_arg(other),
                          self._mask._plb_block, self._domain )
        return self

    def _inplace_operation_with_arg(self, other, operation, arg):
        if isinstance(other, self.__class__):
            second = _type_adjust_block(self.dtype(), other)
            common_domain = util.intersect(self._domain, second._domain)
            common_mask = _combineMasks(self._mask, second._mask)
            if common_mask is None: # Unmasked operation.
                getattr(self._plb(),operation) (
                           self._plb_block, arg, second._plb_block,
                           common_domain )
            else: # Masked operation.
                getattr(self._plb(),'m_'+operation) (
                           self._plb_block, arg, second._plb_block,
                           common_mask._plb_block, common_domain )
        else:
            if self._mask is None: # Unmasked operation.
                getattr(self._plb(),operation) (
                          self._plb_block, arg, self._numeric_arg(other), self._domain )
            else: # Masked operation.
                getattr(self._plb(),'m_'+operation) (
                          self._plb_block, arg, self._numeric_arg(other),
                          self._mask._plb_block, self._domain )
        return self

    def _bool_operation(self, other, operation):
        if isinstance(other, self.__class__):
            first, second = _type_adjust_two_blocks(self, other)
            common_domain = util.intersect(first._domain, second._domain)
            common_mask = _combineMasks(first._mask, second._mask)
            if common_mask is None: # Unmasked operation.
                new_field = getattr(first._plb(),operation) (
                        first._plb_block, second._plb_block, common_domain )
            else: # Masked operation.
                new_field = getattr(first._plb(),'m_'+operation) (
                        first._plb_block, second._plb_block,
                        common_mask._plb_block, common_domain )
            interface = datatype.block_interface(int)
            return self.__class__(new_field, interface, common_domain, common_mask)
        else:
            block = _type_adjust_block_on_scalar(self, other)
            arg = block._numeric_arg(other)
            if block._mask is None: # Unmasked operation.
                new_field = getattr(block._plb(),operation) (
                        block._plb_block, arg, block._domain )
            else: # Masked operation.
                new_field = getattr(block._plb(),'m_'+operation) (
                        block._plb_block, arg,
                        block._mask._plb_block, block._domain )
            interface = datatype.block_interface(int)
            return self.__class__(new_field, interface, block._domain, block._mask)


class RawBlock2D (RawBlock,BlockBase2D):
    """
    The 2D scalar-field or tensor-field class. Note that it is usually more
    convenient to create class-instances through the generator function Block2D.

    Syntax:
    >>> scalarField = RawBlock2D(plb.core.Box2D(0,100,0,100), 1.)
    >>> vectorField = RawBlock2D(plb.core.Box2D(0,100,0,100), [0., 0.])
    """
    def __init__(self, *args):
        RawBlock.__init__(self, *args)

class RawBlock3D (RawBlock,BlockBase3D):
    """
    The 3D scalar-field or tensor-field class. Note that it is usually more
    convenient to create class-instances through the generator function Block3D.

    Syntax:
    >>> scalarField = RawBlock3D(plb.core.Box2D(0,100,0,100,0,100), 1.)
    >>> vectorField = RawBlock3D(plb.core.Box2D(0,100,0,100,0,100), [0., 0.])
    """
    def __init__(self, *args):
        RawBlock.__init__(self, *args)

class BlockLattice2D(BlockLattice, BlockBase2D):
    """
    The 2D block-lattice class. Note that it is usually more convenient
    to create class-instances through the generator function Block2D.

    Syntax:
    >>> lattice = BlockLattice(plb.core.Box2D(0,100,0,100), D2Q9, BGK(1.))
    """
    def __init__(self, *args):
        BlockLattice.__init__(self, *args)

class BlockLattice3D(BlockLattice, BlockBase3D):
    """
    The 3D block-lattice class. Note that it is usually more convenient
    to create class-instances through the generator function Block3D.

    Syntax:
    >>> lattice = BlockLattice(plb.core.Box3D(0,100,0,100,0,100), D3Q19, BGK(1.))
    """
    def __init__(self, *args):
        BlockLattice.__init__(self, *args)


def _combineMasks(mask1, mask2):
    """ Combine two masks through an 'and' operation. Takes into account that
        a mask can be 'None', which stands for 'True' everywhere. """
    if mask1 is None:
        return mask2
    elif mask2 is None:
        return mask1
    else:
        int_interf = datatype.block_interface(int)
        conv_mask1 = _type_adjust_block(int_interf.dtype, mask1)
        conv_mask2 = _type_adjust_block(int_interf.dtype, mask2)
        common_domain = util.intersect(conv_mask1._domain, conv_mask2._domain)
        result_field = int_interf.library.logicalAnd(mask1._plb_block, mask2._plb_block, common_domain)
        return mask1.__class__(result_field, int_interf, common_domain, None)

def _type_adjust_two_blocks(first, second):
    """ Prepare two blocks for an arithmetic operation. For example, when
        adding an int-block with a float-block, the int one must be converted
        to float, because float is stronger than int. """
    first_dt = first.dtype()
    second_dt = second.dtype()
    if first_dt==second_dt:
        return first, second
    else:
        dominant_dt = datatype.to_strong_dominant_dtype(first_dt, second_dt)
        if first_dt==dominant_dt:
            return first, second.convert(dominant_dt)
        else:
            return first.convert(dominant_dt), second

def _type_adjust_block_on_scalar(block, value):
    """
    Prepare a block for an arithmetic operation with a scalar. Only the
    block is adjusted here. The type of the scalar must then be adjusted
    to the type of the block manually.
    Note that if the block is float-valued and the value is double, the
    block is not being converted to double, because this is not what the
    user intuitively expects (if the Palabos structures are float, one
    expects that the whole program goes with floats, although technically
    speaking a literal like "1." is always double).
    """
    block_dt = block.dtype()
    try:              # "value" can be an array ...
        scalar_dt = datatype.pypal_dtype(type(value[0]))
    except (TypeError, IndexError): # ... or a scalar.
        scalar_dt = datatype.pypal_dtype(type(value))
    # 'weak' means: double does not dominate over float.
    dominant_dt = datatype.to_weak_dominant_dtype(block_dt, scalar_dt)
    if dominant_dt == block_dt:
        return block
    else:
        return block.convert(dominant_dt)

def _type_adjust_block(first_dt, second):
    """ Force type conversion for a block. """
    second_dt = second.dtype()
    if first_dt==second_dt:
        return second
    else:
        return second.convert(first_dt)

