import plb
import numpy
import lb
import ppexcept

"""
To access the SWIG-wrapped Palabos stuff, one needs a _plb handler. There are
many handlers: one for d2q9-float, one for d3q19-double, one for block-int,
etc. A handler can be generated automatically from an interface, which is
automatically generated for you in the present module. Example usage:
  interface=datatype.block_interface(Float)
  interface=datatype.lattice_interface(D2Q9, Float)
From an interface, you get
- The _plb handler: _plb = interface.library
- The data type: dtype = interface.dtype
- The data type for communication with SWIG: dtype = interface.swig_dtype

The module also provides various type conversion utilities.
"""

class Int32_Block:
    library    = plb.int_block
    dtype      = numpy.int32
    swig_dtype = int

class Float_Block:
    library    = plb.float_block
    dtype      = numpy.float32
    swig_dtype = float

class Double_Block:
    library    = plb.double_block
    dtype      = numpy.float64
    swig_dtype = float

class Float_D2Q9:
    library    = plb.float_d2q9
    lattice    = lb.D2Q9
    dtype      = numpy.float32
    swig_dtype = float

class Double_D2Q9:
    library    = plb.double_d2q9
    lattice    = lb.D2Q9
    dtype      = numpy.float64
    swig_dtype = float

class Float_D3Q19:
    library    = plb.float_d3q19
    lattice    = lb.D3Q19
    dtype      = numpy.float32
    swig_dtype = float

class Double_D3Q19:
    library    = plb.double_d3q19
    lattice    = lb.D3Q19
    dtype      = numpy.float64
    swig_dtype = float

def pypal_dtype(dtype):
    """ Pypal works with just three types (single-precision, double-precision,
        32-bit int. This function converts any basic data type to one of these
        three. """
    if dtype==int or dtype==numpy.int32 or dtype==numpy.int64:
        return numpy.int32
    elif dtype==numpy.float32:
        return numpy.float32
    elif dtype==float or dtype==numpy.float64:
        return numpy.float64
    else:
        raise ppexcept.PypalUnknownTypeError(dtype)

# Usual type conversion rules.
def to_strong_dominant_dtype(dtype1, dtype2):
    """ Return the dominant data type (the datatype resulting from an
        arithmetic operation between two values) between the two arguments. """
    if dtype1==dtype2:
        return dtype1
    elif dtype1==numpy.int32:
        return dtype2
    elif dtype1==numpy.float32:
        if dtype2==numpy.int32:
            return dtype1
        else:
            return dtype2
    elif dtype1==numpy.float64:
        return dtype1
    else:
        raise ppexcept.PypalUnknownTypeError(dtype)

# Double does not dominate over float.
def to_weak_dominant_dtype(dtype1, dtype2):
    """ Return the dominant data type (the datatype resulting from an
        arithmetic operation between two values) between the two arguments.
        However, if the first data type is float, it is not dominated by
        double. This yields an intuitive behavior for a user who does
        everything in float and uses Python literals (e.g. 1.) which are
        alwasys double-precision. """
    if dtype1==numpy.int32:
        return dtype2
    else:
        return dtype1

def block_interface(dtype):
    """ Create a Palabos interface for raw-blocks, for a given data type. """
    adapt_dtype = pypal_dtype(dtype)
    if adapt_dtype==numpy.int32:
        return Int32_Block
    elif adapt_dtype==numpy.float32:
        return Float_Block
    elif adapt_dtype==numpy.float64:
        return Double_Block
    else:
        raise ppexcept.PypalUnknownTypeError(dtype)


def lattice_interface(lattice, dtype):
    """ Create a Palabos interface for block-lattices, for a given data type and
        lattice structure. """
    adapt_dtype = pypal_dtype(dtype)
    if lattice==lb.D2Q9:
        if adapt_dtype==numpy.float32:
            return Float_D2Q9
        elif adapt_dtype==numpy.float64:
            return Double_D2Q9
        else:
            raise ppexcept.PypalUnknownTypeError(dtype)
    if lattice==lb.D3Q19:
        if adapt_dtype==numpy.float32:
            return Float_D3Q19
        elif adapt_dtype==numpy.float64:
            return Double_D3Q19
        else:
            raise ppexcept.PypalUnknownTypeError(dtype)

