"""
A few useful visualization routines.
"""

__all__ = ['repr_xy']

import mpi4py
import matplotlib.pyplot as plt

class DefaultMessageOut:
    def emit(self, arg):
        print(arg)

class BufferedMessageOut:
    def __init__(self):
        self._messages = [] 
    def emit(self, message):
        self._messages.append(message)
    def consumeMessages(self):
        messages = self._messages
        self._messages=[]
        return messages


messageOut = DefaultMessageOut()

def emit_message(arg):
    messageOut.emit(arg)

def repr_xy(matrix):
    if matrix.ndim==2:
        return matrix.transpose()[::-1,:]
    elif matrix.ndim==3:
        return matrix.transpose((2,1,0))[:,::-1,:]
    else:
        return matrix

def display(*args):
    try: # Try if argument is a PyPal block.
        _par_print_xy(args[0].to_array())
    except AttributeError: # Otherwise, just print it the classical way.
        _par_print(*args)

def show_plt():
    if mpi4py.MPI.COMM_WORLD.Get_rank() == 0:
        plt.show()

def save_plt(fname):
    if mpi4py.MPI.COMM_WORLD.Get_rank() == 0:
        plt.savefig(fname)

def imagesc(images, fname=None):
    plt.figure(1)
    l=len(images)
    try:
        arrangement = {
                1: 111,
                2: 211,
                3: 311,
                4: 221,
                5: 321,
                6: 321,
                7: 421,
                8: 421,
                9: 331,
               10: 521,
               11: 431,
               12: 431 }[len(images)]
    except KeyError: # In case there are more than 12 images
        arrangement = 431

    for i in range(0,min(12,len(images))):
        plt.subplot(arrangement+i)
        nextImage = images[i].to_array()
        try:
            nextImage = images[i].to_array()
        except AttributeError:
            nextImage = images[i]
        plt.imshow(repr_xy(nextImage), interpolation='nearest')
    try:
        save_plt(fname)
    except:
        show_plt()

def isoSurface(image, contourArg):
    try:
        nextImage = image.to_array()
    except AttributeError:
        nextImage = image
    try:
        from enthought.mayavi import mlab
        if mpi4py.MPI.COMM_WORLD.Get_rank() == 0:
            fig = mlab.figure()
            obj = mlab.contour3d(nextImage, figure=fig, contours=contourArg, transparent=True)
            mlab.show()
    except ImportError:
        display('Cannot display iso-surface, because MayaVi is not installed.')

def quiver(velocity):
    u = velocity.getComponent(0).to_array()
    v = velocity.getComponent(1).to_array()
    w = velocity.getComponent(2).to_array()
    try:
        from enthought.mayavi import mlab
        obj = mlab.quiver3d(u,v,w)
        if mpi4py.MPI.COMM_WORLD.Get_rank() == 0:
            mlab.show()
    except ImportError:
        display('Cannot display iso-surface, because MayaVi is not installed.')

def flow(u,v,w):
    u = u.to_array()
    v = v.to_array()
    w = w.to_array()
    try:
        from enthought.mayavi import mlab
        obj = mlab.flow(u,v,w)
        if mpi4py.MPI.COMM_WORLD.Get_rank() == 0:
            mlab.show()
    except ImportError:
        display('Cannot display iso-surface, because MayaVi is not installed.')

def _par_print(*args):
    if mpi4py.MPI.COMM_WORLD.Get_rank() == 0:
        message = ' '.join([str(i) for i in args])
        emit_message(message)

def _par_print_xy(arg):
    if mpi4py.MPI.COMM_WORLD.Get_rank() == 0:
        emit_message(repr_xy(arg))

