/* This file is part of the Palabos library.
 *
 * Copyright (C) 2011 FlowKit Sarl
 * Avenue de Chailly 23
 * 1012 Lausanne, Switzerland
 * E-mail contact: contact@flowkit.com
 *
 * The most recent release of Palabos can be downloaded at 
 * <http://www.palabos.org/>
 *
 * The library Palabos is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * The library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/** \file
 * A collection of dynamics classes (e.g. BGK) with which a Cell object
 * can be instantiated -- header file.
 */
#ifndef NLD_BOUNDARY_DYNAMICS_3D_H
#define NLD_BOUNDARY_DYNAMICS_3D_H

#include "core/globalDefs.h"
#include "core/array.h"
#include "boundaryCondition/regularizedBoundaryDynamics.h"

namespace plb {

template<typename T, template<typename U> class Descriptor>
class NonLocalBoundaryDynamics3D : public CompositeDynamics<T,Descriptor> {
public:
    NonLocalBoundaryDynamics3D(Dynamics<T,Descriptor>* baseDynamics_);
    virtual bool isBoundary() const;
    virtual bool isNonLocal() const;

    virtual void planeBoundaryCompletion (
            int direction, int orientation, plint iX, plint iY, plint iZ,
            BlockLattice3D<T,Descriptor>& lattice ) =0;
    virtual void edgeBoundaryCompletion (
            int plane, int normal1, int normal2, plint iX, plint iY, plint iZ,
            BlockLattice3D<T,Descriptor>& lattice ) =0;
    virtual void cornerBoundaryCompletion (
            int xNormal, int yNormal, int zNormal, plint iX, plint iY, plint iZ,
            BlockLattice3D<T,Descriptor>& lattice ) =0;

    virtual void planeComputeMacroscopic (
            int direction, int orientation, plint iX, plint iY, plint iZ,
            BlockLattice3D<T,Descriptor>& lattice,
            T& rhoBar, Array<T,3>& j, T& thetaBar ) =0;
    virtual void edgeComputeMacroscopic (
            int plane, int normal1, int normal2, plint iX, plint iY, plint iZ,
            BlockLattice3D<T,Descriptor>& lattice,
            T& rhoBar, Array<T,3>& j, T& thetaBar ) =0;
    virtual void cornerComputeMacroscopic (
            int xNormal, int yNormal, int zNormal, plint iX, plint iY, plint iZ,
            BlockLattice3D<T,Descriptor>& lattice,
            T& rhoBar, Array<T,3>& j, T& thetaBar ) =0;

public:
    static void staticPlaneComputeMacroscopic (
            int direction, int orientation, plint iX, plint iY, plint iZ,
            BlockLattice3D<T,Descriptor>& lattice,
            T& rhoBar, Array<T,3>& j, T& thetaBar )
    {
        NonLocalBoundaryDynamics3D<T,Descriptor>* dynamics =
            dynamic_cast<NonLocalBoundaryDynamics3D<T,Descriptor>*>(&lattice.get(iX,iY,iZ).getDynamics());
        if (dynamics) {
            dynamics->planeComputeMacroscopic(direction, orientation, iX, iY, iZ, lattice, rhoBar, j, thetaBar);
        }
        else {
            rhoBar = T();
            j.resetToZero();
            thetaBar = T();
        }
    }
    static void staticEdgeComputeMacroscopic (
            int plane, int normal1, int normal2, plint iX, plint iY, plint iZ,
            BlockLattice3D<T,Descriptor>& lattice,
            T& rhoBar, Array<T,3>& j, T& thetaBar )
    {
        NonLocalBoundaryDynamics3D<T,Descriptor>* dynamics =
            dynamic_cast<NonLocalBoundaryDynamics3D<T,Descriptor>*>(&lattice.get(iX,iY,iZ).getDynamics());
        if (dynamics) {
            dynamics->edgeComputeMacroscopic(plane, normal1, normal2, iX, iY, iZ, lattice, rhoBar, j, thetaBar);
        }
        else {
            rhoBar = T();
            j.resetToZero();
            thetaBar = T();
        }
    }
    static void staticCornerComputeMacroscopic (
            int xNormal, int yNormal, int zNormal, plint iX, plint iY, plint iZ,
            BlockLattice3D<T,Descriptor>& lattice,
            T& rhoBar, Array<T,3>& j, T& thetaBar )
    {
        NonLocalBoundaryDynamics3D<T,Descriptor>* dynamics =
            dynamic_cast<NonLocalBoundaryDynamics3D<T,Descriptor>*>(&lattice.get(iX,iY,iZ).getDynamics());
        if (dynamics) {
            dynamics->cornerComputeMacroscopic(xNormal, yNormal, zNormal, iX, iY, iZ, lattice, rhoBar, j, thetaBar);
        }
        else {
            rhoBar = T();
            j.resetToZero();
            thetaBar = T();
        }
    }
};

template<typename T, template<typename U> class Descriptor>
class NLD_VelocityBoundaryDynamics3D : public NonLocalBoundaryDynamics3D<T,Descriptor>
{
public:
    NLD_VelocityBoundaryDynamics3D(Dynamics<T,Descriptor>* baseDynamics_);
    NLD_VelocityBoundaryDynamics3D(HierarchicUnserializer& unserializer);
    virtual void defineVelocity(Cell<T,Descriptor>& cell, Array<T,Descriptor<T>::d> const& velocity_);
    virtual void computeVelocity( Cell<T,Descriptor> const& cell,
                                  Array<T,Descriptor<T>::d>& velocity_ ) const;
    virtual void serialize(HierarchicSerializer& serializer) const;
    virtual void unserialize(HierarchicUnserializer& unserializer);
    virtual NLD_VelocityBoundaryDynamics3D<T,Descriptor>* clone() const;
    virtual int getId() const;
    virtual void prepareCollision(Cell<T,Descriptor>& cell);
    Array<T,3> const& get_u() const { return u; }
public:
    virtual void planeBoundaryCompletion (
            int direction, int orientation, plint iX, plint iY, plint iZ,
            BlockLattice3D<T,Descriptor>& lattice );
    virtual void edgeBoundaryCompletion (
            int plane, int normal1, int normal2, plint iX, plint iY, plint iZ,
            BlockLattice3D<T,Descriptor>& lattice );
    virtual void cornerBoundaryCompletion (
            int xNormal, int yNormal, int zNormal, plint iX, plint iY, plint iZ,
            BlockLattice3D<T,Descriptor>& lattice );

    virtual void planeComputeMacroscopic (
            int direction, int orientation, plint iX, plint iY, plint iZ,
            BlockLattice3D<T,Descriptor>& lattice,
            T& rhoBar, Array<T,3>& j, T& thetaBar );
    virtual void edgeComputeMacroscopic (
            int plane, int normal1, int normal2, plint iX, plint iY, plint iZ,
            BlockLattice3D<T,Descriptor>& lattice,
            T& rhoBar, Array<T,3>& j, T& thetaBar );
    virtual void cornerComputeMacroscopic (
            int xNormal, int yNormal, int zNormal, plint iX, plint iY, plint iZ,
            BlockLattice3D<T,Descriptor>& lattice,
            T& rhoBar, Array<T,3>& j, T& thetaBar );
private:
    Array<T,3> u;
    static int id;
};

template<typename T, template<typename U> class Descriptor>
class NLD_VelocityNeumannBoundaryDynamics3D : public NonLocalBoundaryDynamics3D<T,Descriptor>
{
public:
    NLD_VelocityNeumannBoundaryDynamics3D(Dynamics<T,Descriptor>* baseDynamics_);
    NLD_VelocityNeumannBoundaryDynamics3D(HierarchicUnserializer& unserializer);
    virtual NLD_VelocityNeumannBoundaryDynamics3D<T,Descriptor>* clone() const;
    virtual int getId() const;
    virtual void prepareCollision(Cell<T,Descriptor>& cell);
public:
    virtual void planeBoundaryCompletion (
            int direction, int orientation, plint iX, plint iY, plint iZ,
            BlockLattice3D<T,Descriptor>& lattice );
    virtual void edgeBoundaryCompletion (
            int plane, int normal1, int normal2, plint iX, plint iY, plint iZ,
            BlockLattice3D<T,Descriptor>& lattice );
    virtual void cornerBoundaryCompletion (
            int xNormal, int yNormal, int zNormal, plint iX, plint iY, plint iZ,
            BlockLattice3D<T,Descriptor>& lattice );

    virtual void planeComputeMacroscopic (
            int direction, int orientation, plint iX, plint iY, plint iZ,
            BlockLattice3D<T,Descriptor>& lattice,
            T& rhoBar, Array<T,3>& j, T& thetaBar );
    virtual void edgeComputeMacroscopic (
            int plane, int normal1, int normal2, plint iX, plint iY, plint iZ,
            BlockLattice3D<T,Descriptor>& lattice,
            T& rhoBar, Array<T,3>& j, T& thetaBar );
    virtual void cornerComputeMacroscopic (
            int xNormal, int yNormal, int zNormal, plint iX, plint iY, plint iZ,
            BlockLattice3D<T,Descriptor>& lattice,
            T& rhoBar, Array<T,3>& j, T& thetaBar );
private:
    static int id;
};

}  // namespace plb

#endif  // NLD_BOUNDARY_DYNAMICS_3D_H
