/* This file is part of the Palabos library.
 *
 * Copyright (C) 2011 FlowKit Sarl
 * Avenue de Chailly 23
 * 1012 Lausanne, Switzerland
 * E-mail contact: contact@flowkit.com
 *
 * The most recent release of Palabos can be downloaded at 
 * <http://www.palabos.org/>
 *
 * The library Palabos is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * The library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/* Orestis Malaspinas designed some of the classes and concepts contained
 * in this file. */

#ifndef SMAGORINSKY_DYNAMICS_HH
#define SMAGORINSKY_DYNAMICS_HH

#include "complexDynamics/smagorinskyDynamics.h"
#include "core/util.h"
#include "core/latticeStatistics.h"
#include "core/dynamicsIdentifiers.h"
#include "latticeBoltzmann/momentTemplates.h"
#include "latticeBoltzmann/dynamicsTemplates.h"
#include "latticeBoltzmann/geometricOperationTemplates.h"
#include "latticeBoltzmann/externalForceTemplates.h"
#include <cmath>

namespace plb {

template<typename T, template<typename U> class Descriptor>
struct SmagoOperations {
    static T computePrefactor(T omega0, T cSmago) {
        return (T)0.5 * util::sqr(cSmago*omega0*Descriptor<T>::invCs2);
    }
    static T computeOmega(T omega0, T preFactor, T rhoBar, Array<T,SymmetricTensor<T,Descriptor>::n> const& PiNeq)
    {
        T PiNeqNormSqr = SymmetricTensor<T,Descriptor>::tensorNormSqr(PiNeq);
        T PiNeqNorm    = sqrt(PiNeqNormSqr);
        T alpha        = preFactor * Descriptor<T>::invRho(rhoBar);
        T linearTerm   = alpha*PiNeqNorm;
        T squareTerm   = (T)2*alpha*alpha*PiNeqNormSqr;
        // In the following formula, the square-root appearing in the explicit form of
        //   omega is developed to second-order.
        return omega0*(1-linearTerm+squareTerm);
    }
};

template<typename T, template<typename U> class Descriptor>
int SmagorinskyDynamics<T,Descriptor>::id =
    meta::registerGeneralDynamics<T,Descriptor,SmagorinskyDynamics<T,Descriptor> >("Smagorinsky_Generic");

template<typename T, template<typename U> class Descriptor>
SmagorinskyDynamics<T,Descriptor>::SmagorinskyDynamics(Dynamics<T,Descriptor>* baseDynamics_,
                                                       T omega0_, T cSmago_, bool automaticPrepareCollision)
    : OmegaFromPiDynamics<T,Descriptor>(baseDynamics_, automaticPrepareCollision),
      omega0(omega0_),
      cSmago(cSmago_),
      preFactor(SmagoOperations<T,Descriptor>::computePrefactor(omega0, cSmago))
{ }

template<typename T, template<typename U> class Descriptor>
SmagorinskyDynamics<T,Descriptor>::SmagorinskyDynamics(HierarchicUnserializer& unserializer)
    : OmegaFromPiDynamics<T,Descriptor>(0, false),
      omega0(T()),
      cSmago(T()),
      preFactor(T())
{
    unserialize(unserializer);
}


template<typename T, template<typename U> class Descriptor>
T SmagorinskyDynamics<T,Descriptor>::getOmegaFromPiAndRhoBar(Array<T,SymmetricTensor<T,Descriptor>::n> const& PiNeq, T rhoBar) const
{
    return SmagoOperations<T,Descriptor>::computeOmega(omega0, preFactor, rhoBar, PiNeq);
}

template<typename T, template<typename U> class Descriptor>
SmagorinskyDynamics<T,Descriptor>* SmagorinskyDynamics<T,Descriptor>::clone() const {
    return new SmagorinskyDynamics<T,Descriptor>(*this);
}
 
template<typename T, template<typename U> class Descriptor>
int SmagorinskyDynamics<T,Descriptor>::getId() const {
    return id;
}

template<typename T, template<typename U> class Descriptor>
void SmagorinskyDynamics<T,Descriptor>::serialize(HierarchicSerializer& serializer) const
{
    serializer.addValue(omega0);
    serializer.addValue(cSmago);
    OmegaFromPiDynamics<T,Descriptor>::serialize(serializer);
}

template<typename T, template<typename U> class Descriptor>
void SmagorinskyDynamics<T,Descriptor>::unserialize(HierarchicUnserializer& unserializer)
{
    unserializer.readValue(omega0);
    unserializer.readValue(cSmago);
    preFactor = SmagoOperations<T,Descriptor>::computePrefactor(omega0, cSmago);
    OmegaFromPiDynamics<T,Descriptor>::unserialize(unserializer);
}

template<typename T, template<typename U> class Descriptor>
void SmagorinskyDynamics<T,Descriptor>::setOmega(T omega0_)
{
    // Just to be sure to avoid an undefined state, also reset the value of
    // omega in the baseDynamics, although the actual value of omega in the
    // baseDynamics is omega=omega0+deltaOmega and is recomputed before each
    // collision.
    omega0 = omega0_;
    preFactor = SmagoOperations<T,Descriptor>::computePrefactor(omega0, cSmago);
    this->getBaseDynamics().setOmega(omega0);
}

template<typename T, template<typename U> class Descriptor>
T SmagorinskyDynamics<T,Descriptor>::getOmega() const {
    return omega0;
}


/* *************** Class SmagorinskyBGKdynamics ************************************** */

template<typename T, template<typename U> class Descriptor>
int SmagorinskyBGKdynamics<T,Descriptor>::id =
    meta::registerGeneralDynamics<T,Descriptor,SmagorinskyBGKdynamics<T,Descriptor> >("BGK_Smagorinsky");

/** \param omega_ relaxation parameter, related to the dynamic viscosity
 */
template<typename T, template<typename U> class Descriptor>
SmagorinskyBGKdynamics<T,Descriptor>::SmagorinskyBGKdynamics (
        T omega0_, T cSmago_ )
    : IsoThermalBulkDynamics<T,Descriptor>(omega0_),
      omega0(omega0_),
      cSmago(cSmago_),
      preFactor(SmagoOperations<T,Descriptor>::computePrefactor(omega0,cSmago))
{ }

template<typename T, template<typename U> class Descriptor>
SmagorinskyBGKdynamics<T,Descriptor>::SmagorinskyBGKdynamics (
        HierarchicUnserializer& unserializer )
    : IsoThermalBulkDynamics<T,Descriptor>(T()),
      omega0(T()),
      cSmago(T()),
      preFactor(T())
{
    unserialize(unserializer);
}

template<typename T, template<typename U> class Descriptor>
void SmagorinskyBGKdynamics<T,Descriptor>::setOmega(T omega0_)
{
    omega0 = omega0_;
    preFactor = SmagoOperations<T,Descriptor>::computePrefactor(omega0,cSmago);
}

template<typename T, template<typename U> class Descriptor>
T SmagorinskyBGKdynamics<T,Descriptor>::getOmega() const {
    return omega0;
}

template<typename T, template<typename U> class Descriptor>
SmagorinskyBGKdynamics<T,Descriptor>* SmagorinskyBGKdynamics<T,Descriptor>::clone() const {
    return new SmagorinskyBGKdynamics<T,Descriptor>(*this);
}
 
template<typename T, template<typename U> class Descriptor>
int SmagorinskyBGKdynamics<T,Descriptor>::getId() const {
    return id;
}

template<typename T, template<typename U> class Descriptor>
void SmagorinskyBGKdynamics<T,Descriptor>::serialize(HierarchicSerializer& serializer) const
{
    serializer.addValue(omega0);
    serializer.addValue(cSmago);
    IsoThermalBulkDynamics<T,Descriptor>::serialize(serializer);
}

template<typename T, template<typename U> class Descriptor>
void SmagorinskyBGKdynamics<T,Descriptor>::unserialize(HierarchicUnserializer& unserializer)
{
    unserializer.readValue(omega0);
    unserializer.readValue(cSmago);
    preFactor = SmagoOperations<T,Descriptor>::computePrefactor(omega0,cSmago);
    IsoThermalBulkDynamics<T,Descriptor>::unserialize(unserializer);
}

template<typename T, template<typename U> class Descriptor>
void SmagorinskyBGKdynamics<T,Descriptor>::collide (
        Cell<T,Descriptor>& cell,
        BlockStatistics& statistics )
{
    T rhoBar;
    Array<T,Descriptor<T>::d> j;
    Array<T,SymmetricTensor<T,Descriptor>::n> PiNeq;
    momentTemplates<T,Descriptor>::compute_rhoBar_j_PiNeq(cell, rhoBar, j, PiNeq);
    T omega = SmagoOperations<T,Descriptor>::computeOmega (
            omega0, preFactor, rhoBar, PiNeq );
    T uSqr = dynamicsTemplates<T,Descriptor>::bgk_ma2_collision(cell, rhoBar, j, omega);
    if (cell.takesStatistics()) {
        gatherStatistics(statistics, rhoBar, uSqr);
    }
}

template<typename T, template<typename U> class Descriptor>
T SmagorinskyBGKdynamics<T,Descriptor>::computeEquilibrium (
        plint iPop, T rhoBar, Array<T,Descriptor<T>::d> const& j,
        T jSqr, T thetaBar ) const
{
    T invRho = Descriptor<T>::invRho(rhoBar);
    return dynamicsTemplates<T,Descriptor>::bgk_ma2_equilibrium(iPop, rhoBar, invRho, j, jSqr);
}


/* *************** Class GuoExternalForceSmagorinskyBGKdynamics ************************************** */


template<typename T, template<typename U> class Descriptor>
int GuoExternalForceSmagorinskyBGKdynamics<T,Descriptor>::id =
    meta::registerGeneralDynamics<T,Descriptor,GuoExternalForceSmagorinskyBGKdynamics<T,Descriptor> >("BGK_Smagorinsky_Guo");

/** \param omega_ relaxation parameter, related to the dynamic viscosity
 */
template<typename T, template<typename U> class Descriptor>
GuoExternalForceSmagorinskyBGKdynamics<T,Descriptor>::GuoExternalForceSmagorinskyBGKdynamics (
        T omega0_, T cSmago_ )
    : IsoThermalBulkDynamics<T,Descriptor>(omega0_),
      omega0(omega0_),
      cSmago(cSmago_)
{ }

template<typename T, template<typename U> class Descriptor>
GuoExternalForceSmagorinskyBGKdynamics<T,Descriptor>::GuoExternalForceSmagorinskyBGKdynamics(HierarchicUnserializer& unserializer)
    : IsoThermalBulkDynamics<T,Descriptor>(T()),
      omega0(T()),
      cSmago(T())
{
    unserialize(unserializer);
}

template<typename T, template<typename U> class Descriptor>
GuoExternalForceSmagorinskyBGKdynamics<T,Descriptor>* GuoExternalForceSmagorinskyBGKdynamics<T,Descriptor>::clone() const {
    return new GuoExternalForceSmagorinskyBGKdynamics<T,Descriptor>(*this);
}

template<typename T, template<typename U> class Descriptor>
int GuoExternalForceSmagorinskyBGKdynamics<T,Descriptor>::getId() const {
    return id;
}

template<typename T, template<typename U> class Descriptor>
void GuoExternalForceSmagorinskyBGKdynamics<T,Descriptor>::serialize(HierarchicSerializer& serializer) const
{
    serializer.addValue(omega0);
    serializer.addValue(cSmago);
    IsoThermalBulkDynamics<T,Descriptor>::serialize(serializer);
}

template<typename T, template<typename U> class Descriptor>
void GuoExternalForceSmagorinskyBGKdynamics<T,Descriptor>::unserialize(HierarchicUnserializer& unserializer)
{
    unserializer.readValue(omega0);
    unserializer.readValue(cSmago);
    IsoThermalBulkDynamics<T,Descriptor>::unserialize(unserializer);
}

template<typename T, template<typename U> class Descriptor>
void GuoExternalForceSmagorinskyBGKdynamics<T,Descriptor>::setOmega(T omega0_)
{
    omega0 = omega0_;
}

template<typename T, template<typename U> class Descriptor>
T GuoExternalForceSmagorinskyBGKdynamics<T,Descriptor>::getOmega() const {
    return IsoThermalBulkDynamics<T,Descriptor>::getOmega();
}

template<typename T, template<typename U> class Descriptor>
void GuoExternalForceSmagorinskyBGKdynamics<T,Descriptor>::computeVelocity (
                                        Cell<T,Descriptor> const& cell,
                                        Array<T,Descriptor<T>::d>& u ) const
{
    T rhoBar;
    Array<T,Descriptor<T>::d> force, j;
    momentTemplates<T,Descriptor>::get_rhoBar_j(cell,rhoBar, j);
    force.from_cArray(cell.getExternal(Descriptor<T>::ExternalField::forceBeginsAt));

    T invRho = Descriptor<T>::invRho(rhoBar);
    for (plint iD = 0; iD < Descriptor<T>::d; ++iD)
    {
        u[iD] = j[iD]*invRho + force[iD]/(T)2;
    }

}


template<typename T, template<typename U> class Descriptor>
void GuoExternalForceSmagorinskyBGKdynamics<T,Descriptor>::collide (
        Cell<T,Descriptor>& cell,
        BlockStatistics& statistics )
{
    T rhoBar = momentTemplates<T,Descriptor>::get_rhoBar(cell);
    Array<T,Descriptor<T>::d> u,j;
    computeVelocity(cell, u);
    T rho = Descriptor<T>::fullRho(rhoBar);
    j = rho * u;
    Array<T,SymmetricTensor<T,Descriptor>::n> PiNeq;
    momentTemplates<T,Descriptor>::compute_PiNeq(cell, rhoBar, j, PiNeq);
    
    T tau = (T)1/omega0;
    
    Array<T,Descriptor<T>::d> force; force.from_cArray(cell.getExternal(Descriptor<T>::ExternalField::forceBeginsAt));
    Array<T,SymmetricTensor<T,Descriptor>::n> forceCorr; forceCorr.resetToZero();
    plint iPi = 0;
    for (plint iA = 0; iA < Descriptor<T>::d; ++iA) {
        for (plint iB = iA; iB < Descriptor<T>::d; ++iB) {
            forceCorr[iPi] = (u[iA]*force[iB]+u[iB]*force[iA]);
            ++iPi;
        }
    }
    PiNeq += forceCorr * (T)0.5;
    
    T normPiNeq = sqrt((T)2*SymmetricTensor<T,Descriptor>::tensorNormSqr(PiNeq));
    normPiNeq *= (T)2*rho*util::sqr(Descriptor<T>::cs2*cSmago*cSmago);

    T tauSGS = T();
    if (normPiNeq != T()) { // test to avoid division per 0
        Array<T,SymmetricTensor<T,Descriptor>::n> S =
                (-rho*tau*Descriptor<T>::cs2+sqrt(util::sqr(rho*tau*Descriptor<T>::cs2)+normPiNeq)) / normPiNeq * PiNeq;

        T sNorm = sqrt((T)2*SymmetricTensor<T,Descriptor>::tensorNormSqr(S));
    
        tauSGS = cSmago*cSmago*sNorm*Descriptor<T>::invCs2;
    }
    T omega = (T)1/(tau+tauSGS);

    T uSqr = dynamicsTemplates<T,Descriptor>::bgk_ma2_collision(cell, rhoBar, j, omega);
    externalForceTemplates<T,Descriptor>::addGuoForce(cell, u, omega, (T)1);
    
    if (cell.takesStatistics()) {
        gatherStatistics(statistics, rhoBar, uSqr);
    }
}

template<typename T, template<typename U> class Descriptor>
T GuoExternalForceSmagorinskyBGKdynamics<T,Descriptor>::computeEquilibrium (
        plint iPop, T rhoBar, Array<T,Descriptor<T>::d> const& j,
        T jSqr, T thetaBar ) const
{
    T invRho = Descriptor<T>::invRho(rhoBar);
    return dynamicsTemplates<T,Descriptor>::bgk_ma2_equilibrium(iPop, rhoBar, invRho, j, jSqr);
}


/* *************** Class IncGuoExternalForceSmagorinskyBGKdynamics ************************************** */


template<typename T, template<typename U> class Descriptor>
int IncGuoExternalForceSmagorinskyBGKdynamics<T,Descriptor>::id =
    meta::registerGeneralDynamics<T,Descriptor,IncGuoExternalForceSmagorinskyBGKdynamics<T,Descriptor> >("IncBGK_Smagorinsky_Guo");

/** \param omega_ relaxation parameter, related to the dynamic viscosity
 */
template<typename T, template<typename U> class Descriptor>
IncGuoExternalForceSmagorinskyBGKdynamics<T,Descriptor>::IncGuoExternalForceSmagorinskyBGKdynamics (
        T omega0_, T cSmago_ )
    : IsoThermalBulkDynamics<T,Descriptor>(omega0_),
      omega0(omega0_),
      cSmago(cSmago_)
{ }

template<typename T, template<typename U> class Descriptor>
IncGuoExternalForceSmagorinskyBGKdynamics<T,Descriptor>::IncGuoExternalForceSmagorinskyBGKdynamics(HierarchicUnserializer& unserializer)
    : IsoThermalBulkDynamics<T,Descriptor>(T()),
      omega0(T()),
      cSmago(T())
{
    unserialize(unserializer);
}

template<typename T, template<typename U> class Descriptor>
IncGuoExternalForceSmagorinskyBGKdynamics<T,Descriptor>* IncGuoExternalForceSmagorinskyBGKdynamics<T,Descriptor>::clone() const {
    return new IncGuoExternalForceSmagorinskyBGKdynamics<T,Descriptor>(*this);
}

template<typename T, template<typename U> class Descriptor>
int IncGuoExternalForceSmagorinskyBGKdynamics<T,Descriptor>::getId() const {
    return id;
}

template<typename T, template<typename U> class Descriptor>
void IncGuoExternalForceSmagorinskyBGKdynamics<T,Descriptor>::serialize(HierarchicSerializer& serializer) const
{
    serializer.addValue(omega0);
    serializer.addValue(cSmago);
    IsoThermalBulkDynamics<T,Descriptor>::serialize(serializer);
}

template<typename T, template<typename U> class Descriptor>
void IncGuoExternalForceSmagorinskyBGKdynamics<T,Descriptor>::unserialize(HierarchicUnserializer& unserializer)
{
    unserializer.readValue(omega0);
    unserializer.readValue(cSmago);
    IsoThermalBulkDynamics<T,Descriptor>::unserialize(unserializer);
}

template<typename T, template<typename U> class Descriptor>
void IncGuoExternalForceSmagorinskyBGKdynamics<T,Descriptor>::setOmega(T omega0_)
{
    omega0 = omega0_;
}

template<typename T, template<typename U> class Descriptor>
T IncGuoExternalForceSmagorinskyBGKdynamics<T,Descriptor>::getOmega() const {
    return IsoThermalBulkDynamics<T,Descriptor>::getOmega();
}

template<typename T, template<typename U> class Descriptor>
void IncGuoExternalForceSmagorinskyBGKdynamics<T,Descriptor>::computeVelocity (
                                        Cell<T,Descriptor> const& cell,
                                        Array<T,Descriptor<T>::d>& u ) const
{
    T rhoBar;
    Array<T,Descriptor<T>::d> force, j;
    momentTemplates<T,Descriptor>::get_rhoBar_j(cell,rhoBar, j);
    force.from_cArray(cell.getExternal(Descriptor<T>::ExternalField::forceBeginsAt));
    
    T invRho = Descriptor<T>::invRho(rhoBar);
    for (plint iD = 0; iD < Descriptor<T>::d; ++iD)
    {
        u[iD] = j[iD]*invRho + force[iD]/(T)2;
    }
}


template<typename T, template<typename U> class Descriptor>
void IncGuoExternalForceSmagorinskyBGKdynamics<T,Descriptor>::collide (
        Cell<T,Descriptor>& cell,
        BlockStatistics& statistics )
{
    T rhoBar = momentTemplates<T,Descriptor>::get_rhoBar(cell);
    Array<T,Descriptor<T>::d> u,j;
    computeVelocity(cell, u);
    T rho = Descriptor<T>::fullRho(rhoBar);
    j = rho * u;
    Array<T,SymmetricTensor<T,Descriptor>::n> PiNeq;
    momentTemplates<T,Descriptor>::compute_PiNeq(cell, rhoBar, j, PiNeq);

    T tau = (T)1/omega0;

    Array<T,Descriptor<T>::d> force; force.from_cArray(cell.getExternal(Descriptor<T>::ExternalField::forceBeginsAt));
    Array<T,SymmetricTensor<T,Descriptor>::n> forceCorr; forceCorr.resetToZero();
    plint iPi = 0;
    for (plint iA = 0; iA < Descriptor<T>::d; ++iA) {
        for (plint iB = iA; iB < Descriptor<T>::d; ++iB) {
            forceCorr[iPi] = (u[iA]*force[iB]+u[iB]*force[iA]);
            ++iPi;
        }
    }
    PiNeq += forceCorr * (T)0.5;

    T normPiNeq = sqrt((T)2*SymmetricTensor<T,Descriptor>::tensorNormSqr(PiNeq));
    normPiNeq *= (T)2*rho*util::sqr(Descriptor<T>::cs2*cSmago*cSmago);

    T tauSGS = T();
    if (normPiNeq != T()) { // test to avoid division per 0
        Array<T,SymmetricTensor<T,Descriptor>::n> S =
                (-rho*tau*Descriptor<T>::cs2+sqrt(util::sqr(rho*tau*Descriptor<T>::cs2)+normPiNeq)) / normPiNeq * PiNeq;

        T sNorm = sqrt((T)2*SymmetricTensor<T,Descriptor>::tensorNormSqr(S));

        tauSGS = cSmago*cSmago*sNorm*Descriptor<T>::invCs2;
    }
    
    T omega = (T)1/(tau+tauSGS);

    T uSqr = dynamicsTemplates<T,Descriptor>::bgk_inc_collision(cell, rhoBar, j, omega);
    
    externalForceTemplates<T,Descriptor>::addGuoForce(cell, u, omega, (T)1);
    if (cell.takesStatistics()) {
        gatherStatistics(statistics, rhoBar, uSqr);
    }
}

template<typename T, template<typename U> class Descriptor>
T IncGuoExternalForceSmagorinskyBGKdynamics<T,Descriptor>::computeEquilibrium (
        plint iPop, T rhoBar, Array<T,Descriptor<T>::d> const& j,
        T jSqr, T thetaBar ) const
{
    T invRho = (T)1;
    return dynamicsTemplates<T,Descriptor>::bgk_ma2_equilibrium(iPop, rhoBar, invRho, j, jSqr);
}

template<typename T, template<typename U> class Descriptor>
bool IncGuoExternalForceSmagorinskyBGKdynamics<T,Descriptor>::velIsJ() const {
    return true;
}


/* *************** Class SmagorinskyRegularizedDynamics ************************************ */

template<typename T, template<typename U> class Descriptor>
int SmagorinskyRegularizedDynamics<T,Descriptor>::id =
    meta::registerTwoParamDynamics<T,Descriptor,SmagorinskyRegularizedDynamics<T,Descriptor> >("Smagorinsky_Regularized");

/** \param omega_ relaxation parameter, related to the dynamic viscosity
 */
template<typename T, template<typename U> class Descriptor>
SmagorinskyRegularizedDynamics<T,Descriptor>::SmagorinskyRegularizedDynamics (
        T omega0_, T cSmago_ )
    : IsoThermalBulkDynamics<T,Descriptor>(omega0_),
      omega0(omega0_),
      cSmago(cSmago_),
      preFactor(SmagoOperations<T,Descriptor>::computePrefactor(omega0,cSmago))
{ }

template<typename T, template<typename U> class Descriptor>
void SmagorinskyRegularizedDynamics<T,Descriptor>::serialize(HierarchicSerializer& serializer) const
{
    serializer.addValue(omega0);
    serializer.addValue(cSmago);
    IsoThermalBulkDynamics<T,Descriptor>::serialize(serializer);
}


template<typename T, template<typename U> class Descriptor>
void SmagorinskyRegularizedDynamics<T,Descriptor>::unserialize(HierarchicUnserializer& unserializer)
{
    unserializer.readValue(omega0);
    unserializer.readValue(cSmago);
    preFactor = SmagoOperations<T,Descriptor>::computePrefactor(omega0,cSmago);
    IsoThermalBulkDynamics<T,Descriptor>::unserialize(unserializer);
}

template<typename T, template<typename U> class Descriptor>
void SmagorinskyRegularizedDynamics<T,Descriptor>::setOmega(T omega0_)
{
    omega0 = omega0_;
    preFactor = SmagoOperations<T,Descriptor>::computePrefactor(omega0,cSmago);
}

template<typename T, template<typename U> class Descriptor>
T SmagorinskyRegularizedDynamics<T,Descriptor>::getOmega() const {
    return omega0;
}

template<typename T, template<typename U> class Descriptor>
SmagorinskyRegularizedDynamics<T,Descriptor>* SmagorinskyRegularizedDynamics<T,Descriptor>::clone() const
{
    return new SmagorinskyRegularizedDynamics<T,Descriptor>(*this);
}
 
template<typename T, template<typename U> class Descriptor>
int SmagorinskyRegularizedDynamics<T,Descriptor>::getId() const {
    return id;
}

template<typename T, template<typename U> class Descriptor>
void SmagorinskyRegularizedDynamics<T,Descriptor>::collide (
        Cell<T,Descriptor>& cell,
        BlockStatistics& statistics )
{
    T rhoBar;
    Array<T,Descriptor<T>::d> j;
    Array<T,SymmetricTensor<T,Descriptor>::n> PiNeq;
    momentTemplates<T,Descriptor>::compute_rhoBar_j_PiNeq(cell, rhoBar, j, PiNeq);
    T omega = SmagoOperations<T,Descriptor>::computeOmega (
            omega0, preFactor, rhoBar, PiNeq );
    IsoThermalBulkDynamics<T,Descriptor>::setOmega(omega);
    T invRho = Descriptor<T>::invRho(rhoBar);
    T uSqr = dynamicsTemplates<T,Descriptor>::rlb_collision (
                 cell, rhoBar, invRho, j, PiNeq, omega );
    if (cell.takesStatistics()) {
        gatherStatistics(statistics, rhoBar, uSqr);
    }
}

template<typename T, template<typename U> class Descriptor>
T SmagorinskyRegularizedDynamics<T,Descriptor>::computeEquilibrium (
        plint iPop, T rhoBar, Array<T,Descriptor<T>::d> const& j, T jSqr, T thetaBar ) const
{
    T invRho = Descriptor<T>::invRho(rhoBar);
    return dynamicsTemplates<T,Descriptor>::bgk_ma2_equilibrium(iPop, rhoBar, invRho, j, jSqr);
}

} // namespace plb

#endif  // SMAGORINSKY_DYNAMICS_HH
