/* This file is part of the Palabos library.
 * Copyright (C) 2009, 2010 Jonas Latt
 * E-mail contact: jonas@lbmethod.org
 * The most recent release of Palabos can be downloaded at 
 * <http://www.lbmethod.org/palabos/>
 *
 * The library Palabos is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * The library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/* Main author: Daniel Lagrava
 */

/** \file
 * Dynamics and data processors used to implement 3D grid refinement -- header file.
 */

#ifndef FINE_GRID_PROCESSORS_3D_H
#define FINE_GRID_PROCESSORS_3D_H

#include "core/globalDefs.h"
#include "atomicBlock/blockLattice3D.h"
#include "atomicBlock/dataProcessorWrapper3D.h"
#include "multiBlock/multiBlockLattice3D.h"
#include "multiGrid/gridRefinement.h"
#include "multiGrid/gridRefinementDynamics.h"
#include "finiteDifference/fdStencils1D.h"
#include <vector>

namespace plb {


////////////////////////////////////////////////////////////////////////////////////////////////
//////////  New approach, extract everything inside a MultiTensorField3D and interpolate ///////

/// Copy data from a coarse to a fine TensorField3D, using cubic interpolation. 
/// This is for grid refinement only!
template<typename T>
class NTensorInterpolateCoarseToFine3D : public BoxProcessingFunctional3D_NN<T,T>
{
public:
    virtual void process( Box3D coarseDomain,
                          NTensorField3D<T>& coarseField,
                          NTensorField3D<T>& fineField );

    virtual NTensorInterpolateCoarseToFine3D<T>* clone() const;
    
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const {
        modified[0] = modif::nothing;
        modified[1] = modif::staticVariables;
    }
};

template<typename T, template<typename U> class Descriptor>
class CopyDecomposedValuesFromCoarse3D : public BoxProcessingFunctional3D_LN<T,Descriptor,T>
{
public:
    CopyDecomposedValuesFromCoarse3D(RescaleEngine<T,Descriptor>* rescaleEngine_);
    virtual ~CopyDecomposedValuesFromCoarse3D();
    
    virtual void process(Box3D domain, BlockLattice3D<T,Descriptor>& coarseLattice,
                                       NTensorField3D<T>& decomposedValues);
    virtual CopyDecomposedValuesFromCoarse3D<T,Descriptor>* clone() const;
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const;
    
private:
    RescaleEngine<T,Descriptor>* rescaleEngine;
};


template<typename T, template<typename U> class Descriptor>
class CopyDecomposedValuesToFine3D : public BoxProcessingFunctional3D_LN<T,Descriptor,T>
{
public:
    virtual void process(Box3D domain, BlockLattice3D<T,Descriptor>& fineLattice,
                                       NTensorField3D<T>& decomposedValues);
    virtual CopyDecomposedValuesToFine3D<T,Descriptor>* clone() const;
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const;
};

    
/// Processor to be added to coarse lattice: copies t1 values to t0 after each numTimeSteps iteration
template<typename T, template<typename U> class Descriptor>
class Copy_t1_to_t0_3D: public BoxProcessingFunctional3D_L<T,Descriptor>
{
public:
    Copy_t1_to_t0_3D(plint numTimeSteps_, plint executionTime_);
    virtual void process(Box3D domain, BlockLattice3D<T,Descriptor>& fineLattice);
    virtual Copy_t1_to_t0_3D<T,Descriptor>* clone() const;
    
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const {
        modified[0] = modif::allVariables;
    }
private:
    plint numTimeSteps;
    plint executionTime;
};

/// A helper class for grid refinement processors, used to convert location between grids
template<typename T, template<typename U> class Descriptor>
class CoarseToFineConverter3D {
public:
    CoarseToFineConverter3D( BlockLattice3D<T,Descriptor> const& coarseLattice,
                             BlockLattice3D<T,Descriptor>& fineLattice_ )
        : fineLattice(fineLattice_),
          coarseLocation(coarseLattice.getLocation()),
          fineLocation(fineLattice.getLocation())
    { }
    /// Convert the x-component of a location from coarse lattice to fine lattice.
    plint fineX(plint coarseX) const {
        return (coarseX+coarseLocation.x)*2 - fineLocation.x;
    }
    /// Convert the y-component of a location from coarse lattice to fine lattice.
    plint fineY(plint coarseY) const {
        return (coarseY+coarseLocation.y)*2 - fineLocation.y;
    }
    /// Convert the z-component of a location from coarse lattice to fine lattice.
    plint fineZ(plint coarseZ) const {
        return (coarseZ+coarseLocation.z)*2 - fineLocation.z;
    }
    
    /// Convert location from coarse to fine, add the value delta, and return the dynamic
    FineGridBoundaryDynamics<T,Descriptor>& fineDynamics (
            Dot3D coarsePos, Dot3D delta )
    {
        pcout << "--> " << fineX(coarsePos.x)+delta.x << " " << 
                        fineY(coarsePos.y)+delta.y << " " << fineZ(coarsePos.z)+delta.z << std::endl;
        return dynamic_cast<FineGridBoundaryDynamics<T,Descriptor>&> (
                   fineLattice.get( fineX(coarsePos.x) + delta.x,
                                    fineY(coarsePos.y) + delta.y,
                                    fineZ(coarsePos.z) + delta.z ).getDynamics() );
    }
    
    FineGridBoundaryDynamics<T,Descriptor>& fineDynamics (
            plint x, plint y, plint z, plint deltaX, plint deltaY, plint deltaZ ){
        return fineDynamics(Dot3D(x,y,z),Dot3D(deltaX,deltaY,deltaZ));
    }
    
    FineGridBoundaryDynamics<T,Descriptor>& fineDynamicsFromFineCoordinates (Dot3D pos){
        pcout << "\t---> " << pos.x << " " << pos.y <<  " " << pos.z << std::endl;
        return dynamic_cast<FineGridBoundaryDynamics<T,Descriptor>&> (
                   fineLattice.get( pos.x,pos.y,pos.z).getDynamics() 
                   );
    }
            
private:
    BlockLattice3D<T,Descriptor> &fineLattice;
    Dot3D coarseLocation, fineLocation;
};







}  // namespace plb

#endif  // FINE_GRID_PROCESSORS_3D_H
