/* This file is part of the Palabos library.
 * Copyright (C) 2009, 2010 Jonas Latt
 * E-mail contact: jonas@lbmethod.org
 * The most recent release of Palabos can be downloaded at 
 * <http://www.lbmethod.org/palabos/>
 *
 * The library Palabos is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * The library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/* Main author: Daniel Lagrava
 */

/** \file
 * Dynamics and data processors used to implement 3D grid refinement -- implementation.
 */

#ifndef FINE_GRID_PROCESSORS_3D_HH
#define FINE_GRID_PROCESSORS_3D_HH

#include "multiGrid/fineGridProcessors3D.h"

namespace plb {


/* ************ Class CopyDecomposedValuesFromCoarse3D *********** */

template<typename T, template<typename U> class Descriptor>
CopyDecomposedValuesFromCoarse3D<T,Descriptor>::CopyDecomposedValuesFromCoarse3D(
                                                RescaleEngine<T,Descriptor>* rescaleEngine_)
    : rescaleEngine(rescaleEngine_)
{}

template<typename T, template<typename U> class Descriptor>
CopyDecomposedValuesFromCoarse3D<T,Descriptor>::~CopyDecomposedValuesFromCoarse3D(){
    delete rescaleEngine;
}

template<typename T, template<typename U> class Descriptor>
void CopyDecomposedValuesFromCoarse3D<T,Descriptor>::process(Box3D coarseDomain, 
                                                             BlockLattice3D<T,Descriptor>& coarseLattice,               
                                                             NTensorField3D<T>& decomposedValues)
{
    plint x0 = coarseDomain.x0;
    plint y0 = coarseDomain.y0;
    plint z0 = coarseDomain.z0;
    plint x1 = coarseDomain.x1;
    plint y1 = coarseDomain.y1;
    plint z1 = coarseDomain.z1;
    
    std::vector<T> pop;
    Dot3D offset = computeRelativeDisplacement(coarseLattice, decomposedValues);
    
    for (plint iX=x0; iX<=x1; ++iX){
        for (plint iY=y0; iY<=y1; ++iY){
            for (plint iZ=z0; iZ<=z1; ++iZ){
                
                rescaleEngine->scaleCoarseFine( coarseLattice.get(iX,iY,iZ), pop );
                for (pluint iPop=0; iPop<pop.size(); ++iPop){
                    decomposedValues.get(iX+offset.x,iY+offset.y,iZ+offset.z)[iPop] = pop[iPop];
                }
            }
        }
    }
}
                                    
template<typename T, template<typename U> class Descriptor>                                    
CopyDecomposedValuesFromCoarse3D<T,Descriptor>* CopyDecomposedValuesFromCoarse3D<T,Descriptor>::clone() const
{
    return new CopyDecomposedValuesFromCoarse3D<T,Descriptor>(this->rescaleEngine->clone());
}

template<typename T, template<typename U> class Descriptor>
void CopyDecomposedValuesFromCoarse3D<T,Descriptor>::getTypeOfModification(std::vector<modif::ModifT>& modified) const
{
    modified[0] = modif::nothing;
    modified[1] = modif::allVariables;
}


/* ************** Class CopyDecomposedValuesToFine3D **************** */


template<typename T, template<typename U> class Descriptor>
void CopyDecomposedValuesToFine3D<T,Descriptor>::process(Box3D domain, BlockLattice3D<T,Descriptor>& fineLattice,
                                    NTensorField3D<T>& decomposedValues)
{
    plint x0 = domain.x0;
    plint y0 = domain.y0;
    plint z0 = domain.z0;
    plint x1 = domain.x1;
    plint y1 = domain.y1;
    plint z1 = domain.z1;
    
    std::vector<T> pop;
    plint whichTime = 1;
    
    Dot3D offset = computeRelativeDisplacement(fineLattice, decomposedValues);
    for (plint iX=x0; iX<=x1; ++iX){
        for (plint iY=y0; iY<=y1; ++iY){
            for (plint iZ=z0; iZ<=z1; ++iZ){
                std::vector<T> pop;
                for (pluint iPop=0; iPop<pop.size(); ++iPop){
                    pop.push_back(decomposedValues.get(iX+offset.x,iY+offset.y,iZ+offset.z)[iPop]);
                }
                
                dynamic_cast<FineGridBoundaryDynamics<T,Descriptor>&>(fineLattice.get(iX,iY,iZ).getDynamics()
                    ).getDecomposedValues(whichTime).assign(pop.begin(),pop.end());
            }
        }
    }
}
                                    
                                    
template<typename T, template<typename U> class Descriptor>
CopyDecomposedValuesToFine3D<T,Descriptor>* CopyDecomposedValuesToFine3D<T,Descriptor>::clone() const
{
    return new CopyDecomposedValuesToFine3D<T,Descriptor>();
}

template<typename T, template<typename U> class Descriptor>
void CopyDecomposedValuesToFine3D<T,Descriptor>::getTypeOfModification(std::vector<modif::ModifT>& modified) const
{
    modified[0] = modif::allVariables;
    modified[1] = modif::nothing;
}

/* *************** Class NTensorInterpolateCoarseToFine3D ***************** */
template<typename T>
void NTensorInterpolateCoarseToFine3D<T>::process( Box3D coarseDomain,
                          NTensorField3D<T>& coarseField,
                          NTensorField3D<T>& fineField )
{
    PLB_ASSERT( coarseField.getNdim()==fineField.getNdim() );
    
    
}
                          
template<typename T>
NTensorInterpolateCoarseToFine3D<T>* NTensorInterpolateCoarseToFine3D<T>::clone() const
{
    return new NTensorInterpolateCoarseToFine3D<T>(*this);
}
    




    
/* *************** Class Copy_t1_to_t0_3D *********************************** */

template<typename T, template<typename U> class Descriptor>
Copy_t1_to_t0_3D<T,Descriptor>::Copy_t1_to_t0_3D(plint numTimeSteps_, plint executionTime_)
    : numTimeSteps(numTimeSteps_),
      executionTime(executionTime_)
{ }

template<typename T, template<typename U> class Descriptor>
void Copy_t1_to_t0_3D<T,Descriptor>::process(Box3D domain, BlockLattice3D<T,Descriptor>& fineLattice)
{
    // Determine current relative value of time steps.
    size_t relativeTime = fineLattice.getTimeCounter().getTime()%numTimeSteps;
    // Execute data processor only if one is at the end of a cycle.
    if ((plint)relativeTime==executionTime) {
        for (plint iX=domain.x0; iX<=domain.x1; ++iX) {
            for (plint iY=domain.y0; iY<=domain.y1; ++iY) {
                for (plint iZ=domain.z0; iZ<=domain.z1; ++iZ) {
                    FineGridBoundaryDynamics<T,Descriptor>& fineDynamics =
                        dynamic_cast<FineGridBoundaryDynamics<T,Descriptor>&>
                            ( fineLattice.get(iX,iY,iZ).getDynamics() );
                    std::vector<T>& t0 = fineDynamics.getDecomposedValues(0);
                    std::vector<T>& t1 = fineDynamics.getDecomposedValues(1);
                    t0.assign(t1.begin(), t1.end());
                }
            }
        }
    }
}

template<typename T, template<typename U> class Descriptor>
Copy_t1_to_t0_3D<T,Descriptor>* Copy_t1_to_t0_3D<T,Descriptor>::clone() const {
    return new Copy_t1_to_t0_3D(*this);
}




}  // namespace plb

#endif  // FINE_GRID_PROCESSORS_3D_HH
