/* This file is part of the Palabos library.
 *
 * Copyright (C) 2011 FlowKit Sarl
 * Avenue de Chailly 23
 * 1012 Lausanne, Switzerland
 * E-mail contact: contact@flowkit.com
 *
 * The most recent release of Palabos can be downloaded at 
 * <http://www.palabos.org/>
 *
 * The library Palabos is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * The library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/* Main author: Daniel Lagrava
 **/

/** \file
 * Various factories that use a multiGridManagement3D -- Header file
 */

#ifndef MULTI_GRID_GENERATOR_3D_HH
#define MULTI_GRID_GENERATOR_3D_HH

#include "multiGrid/multiGridGenerator3D.h"
#include "multiGrid/domainDivision3D.h"

namespace plb {
    
enum {cornerLL=0,cornerUL,cornerLR,cornerUR}; // to access the corners generated with reduce()
enum {borderUp=0,borderLeft,borderDown,borderRight}; // to access the borders generated with reduce()

/// Constants for the interpolation in the different planes
// Dot3D const deltaBulk[3][2] = {
//     {Dot3D(0,1,0), Dot3D(0,0,1)}, // YZ
//     {Dot3D(1,0,0), Dot3D(0,0,1)}, // XZ
//     {Dot3D(1,0,0), Dot3D(0,1,0)}  // XY
// };
// 
// Dot3D const deltaCorner[4][3][2] = {
//     {{Dot3D(0,1,0), Dot3D(0,0,1)}, {Dot3D(1,0,0),Dot3D(0,0,1)},{Dot3D(1,0,0),Dot3D(0,1,0)}}, // cornerLL
//     {{Dot3D(0,1,0), Dot3D(0,0,-1)}, {Dot3D(1,0,0),Dot3D(0,0,-1)},{Dot3D(1,0,0),Dot3D(0,-1,0)}}, // cornerUL
//     {{Dot3D(0,-1,0), Dot3D(0,0,1)}, {Dot3D(-1,0,0),Dot3D(0,0,1)},{Dot3D(-1,0,0),Dot3D(0,1,0)}}, // cornerLR
//     {{Dot3D(0,-1,0), Dot3D(0,0,-1)}, {Dot3D(-1,0,0),Dot3D(0,0,-1)},{Dot3D(-1,0,0),Dot3D(0,-1,0)}}, // cornerUR
// };
// 
// Dot3D const deltaBoundary[4][3][2] = {
//     {{Dot3D(0,1,0), Dot3D(0,0,-1)}, {Dot3D(1,0,0),Dot3D(0,0,-1)},{Dot3D(1,0,0),Dot3D(0,-1,0)}}, // borderUp
//     {{Dot3D(0,0,1), Dot3D(0,1,0)}, {Dot3D(0,0,1),Dot3D(1,0,0)},{Dot3D(0,1,0),Dot3D(1,0,0)}}, // borderLeft
//     {{Dot3D(0,1,0), Dot3D(0,0,1)}, {Dot3D(1,0,0),Dot3D(0,0,1)},{Dot3D(1,0,0),Dot3D(0,1,0)}}, // borderDown
//     {{Dot3D(0,0,1), Dot3D(0,-1,0)}, {Dot3D(0,0,1),Dot3D(-1,0,0)},{Dot3D(0,1,0),Dot3D(-1,0,0)}}, // borderRight
// };



Box3D reduce( Box3D const& fineGridInterface, std::vector<Box3D>& excess,
                    std::vector<Box3D>& corner, plint const& direction ){
    corner.push_back(Box3D(fineGridInterface.x0,fineGridInterface.x0,
                   fineGridInterface.y0,fineGridInterface.y0,
                   fineGridInterface.z0,fineGridInterface.z0)); // corner lower left
    Box3D result;
    if (direction==0){ // plane YZ
        PLB_ASSERT( fineGridInterface.x0 == fineGridInterface.x1 );
        result = Box3D(fineGridInterface.x0,fineGridInterface.x1,
                     fineGridInterface.y0+1,fineGridInterface.y1-1,
                     fineGridInterface.z0+1,fineGridInterface.z1-1);
        excess.push_back(Box3D(fineGridInterface.x0,fineGridInterface.x1,
                                fineGridInterface.y0+1,fineGridInterface.y1-1,
                                fineGridInterface.z1,fineGridInterface.z1)); // border up
        excess.push_back(Box3D(fineGridInterface.x0,fineGridInterface.x1,
                                fineGridInterface.y0,fineGridInterface.y0,
                                fineGridInterface.z0+1,fineGridInterface.z1-1)); // border left
        excess.push_back(Box3D(fineGridInterface.x0,fineGridInterface.x1,
                                fineGridInterface.y0+1,fineGridInterface.y1-1,
                                fineGridInterface.z0,fineGridInterface.z0)); // border down
        excess.push_back(Box3D(fineGridInterface.x0,fineGridInterface.x1,
                                fineGridInterface.y1,fineGridInterface.y1,
                                fineGridInterface.z0+1,fineGridInterface.z1-1)); // border right
        corner.push_back(Box3D(fineGridInterface.x1,fineGridInterface.x1,
                   fineGridInterface.y0,fineGridInterface.y0,
                   fineGridInterface.z1,fineGridInterface.z1)); // corner upper left
        corner.push_back(Box3D(fineGridInterface.x1,fineGridInterface.x1,
                   fineGridInterface.y1,fineGridInterface.y1,
                   fineGridInterface.z0,fineGridInterface.z0)); // corner lower right 
    }
    if (direction==1){ // plane XZ
        PLB_ASSERT( fineGridInterface.y0 == fineGridInterface.y1 );
        result = Box3D(fineGridInterface.x0+1,fineGridInterface.x1-1,
                     fineGridInterface.y0,fineGridInterface.y1,
                     fineGridInterface.z0+1,fineGridInterface.z1-1);
        excess.push_back(Box3D(fineGridInterface.x0+1,fineGridInterface.x1-1,
                                fineGridInterface.y0,fineGridInterface.y1,
                                fineGridInterface.z1,fineGridInterface.z1)); // border up
        excess.push_back(Box3D(fineGridInterface.x0,fineGridInterface.x0,
                                fineGridInterface.y0,fineGridInterface.y1,
                                fineGridInterface.z0+1,fineGridInterface.z1-1)); // border left
        excess.push_back(Box3D(fineGridInterface.x0+1,fineGridInterface.x1-1,
                                fineGridInterface.y0,fineGridInterface.y1,
                                fineGridInterface.z0,fineGridInterface.z0)); // border down
        excess.push_back(Box3D(fineGridInterface.x1,fineGridInterface.x1,
                                fineGridInterface.y0,fineGridInterface.y1,
                                fineGridInterface.z0+1,fineGridInterface.z1-1)); // border right
        corner.push_back(Box3D(fineGridInterface.x0,fineGridInterface.x0,
                   fineGridInterface.y0,fineGridInterface.y1,
                   fineGridInterface.z1,fineGridInterface.z1)); // corner upper left
        corner.push_back(Box3D(fineGridInterface.x1,fineGridInterface.x1,
                   fineGridInterface.y1,fineGridInterface.y1,
                   fineGridInterface.z0,fineGridInterface.z0)); // corner lower right
    }
    if (direction==2){ // plane XY
        PLB_ASSERT( fineGridInterface.z0 == fineGridInterface.z1 );
        result = Box3D(fineGridInterface.x0+1,fineGridInterface.x1-1,
                     fineGridInterface.y0+1,fineGridInterface.y1-1,
                     fineGridInterface.z0,fineGridInterface.z1);
        excess.push_back(Box3D(fineGridInterface.x0+1,fineGridInterface.x1-1,
                                fineGridInterface.y1,fineGridInterface.y1,
                                fineGridInterface.z0,fineGridInterface.z1)); // border up
        excess.push_back(Box3D(fineGridInterface.x0,fineGridInterface.x0,
                                fineGridInterface.y0+1,fineGridInterface.y1-1,
                                fineGridInterface.z0,fineGridInterface.z1)); // border left
        excess.push_back(Box3D(fineGridInterface.x0+1,fineGridInterface.x1-1,
                                fineGridInterface.y0,fineGridInterface.y0,
                                fineGridInterface.z0,fineGridInterface.z1)); // border down
        excess.push_back(Box3D(fineGridInterface.x1,fineGridInterface.x1,
                                fineGridInterface.y0+1,fineGridInterface.y1-1,
                                fineGridInterface.z0,fineGridInterface.z1)); // border right
        
        corner.push_back(Box3D(fineGridInterface.x0,fineGridInterface.x0,
                   fineGridInterface.y1,fineGridInterface.y1,
                   fineGridInterface.z1,fineGridInterface.z1)); // corner upper left
        corner.push_back(Box3D(fineGridInterface.x1,fineGridInterface.x1,
                   fineGridInterface.y0,fineGridInterface.y0,
                   fineGridInterface.z1,fineGridInterface.z1)); // corner lower right
    }

    corner.push_back(Box3D(fineGridInterface.x1,fineGridInterface.x1,
                   fineGridInterface.y1,fineGridInterface.y1,
                   fineGridInterface.z1,fineGridInterface.z1)); // corner upper right

    return result;
}

/// Use the MultiGridManagement3D to generate a verctor of lattices that represent the multi grid.
template<typename T, template<typename U> class Descriptor>
std::vector<MultiBlockLattice3D<T,Descriptor>*> generateLattices(
                MultiGridManagement3D management,
                std::vector<Dynamics<T,Descriptor>*> backgroundDynamics,
                std::vector<BlockCommunicator3D*> communicators,
                std::vector<CombinedStatistics*> combinedStatistics )
{
    // get the MPI id's (if available)
    std::vector<std::vector<plint> > mpiProcesses = management.getMpiProcesses();
    std::vector<std::vector<Box3D> > bulks = management.getBulks();
    std::vector<std::vector<Box3D> > uniqueBulks = management.getUniqueBulks();
    PLB_PRECONDITION (bulks.size() == uniqueBulks.size());
    PLB_PRECONDITION (mpiProcesses.size()==0 || mpiProcesses.size()==bulks.size());
    // for cubic interpolation 
    plint envelopeWidth = 2; 
    // allocation of the result vector
    std::vector<MultiBlockLattice3D<T,Descriptor>*> multiBlocks(bulks.size());
    for (pluint iLevel=0; iLevel<bulks.size(); ++iLevel) {
        // create a thread attribution 
        ExplicitThreadAttribution* threadAttribution = new ExplicitThreadAttribution;
        PLB_ASSERT(mpiProcesses.size()==0 || mpiProcesses[iLevel].size()==bulks[iLevel].size());
        SparseBlockStructure3D geometry(management.getBoundingBox(iLevel));
        for (pluint iBlock=0; iBlock<bulks[iLevel].size(); ++iBlock) {
            plint mpiProcess = mpiProcesses.size() > 0 ? mpiProcesses[iLevel][iBlock] : 0;
            plint blockId = geometry.nextIncrementalId();
            Box3D bulk = bulks[iLevel][iBlock];
            Box3D uniqueBulk = bulk;
            geometry.addBlock(bulk, uniqueBulk, blockId); //TODO: add uniqueBulk
            threadAttribution->addBlock(blockId, mpiProcess);
        }
        MultiBlockManagement3D
            blockManagement( geometry, threadAttribution,
                             envelopeWidth, iLevel );
        multiBlocks[iLevel] = new MultiBlockLattice3D<T,Descriptor> (
            blockManagement,
            communicators[iLevel],
            combinedStatistics[iLevel],
            defaultMultiBlockPolicy3D().getMultiCellAccess<T,Descriptor>(),
            backgroundDynamics[iLevel] );
        multiBlocks[iLevel] -> periodicity().toggleAll(false);
    }
    return multiBlocks;
}

template<typename T, template<typename U> class Descriptor>
std::vector<MultiBlockLattice3D<T,Descriptor>*> generateLattices(
                MultiGridManagement3D management,
                std::vector<Dynamics<T,Descriptor>*> backgroundDynamics)
{
    return generateLattices 
        (management, backgroundDynamics,
         defaultMultiGridPolicy3D().getBlockCommunicator(management.getNumLevels()),
         defaultMultiGridPolicy3D().getCombinedStatistics(management.getNumLevels()));
}


template<typename T, template<typename U> class Descriptor>
void createInterfaces(std::vector<MultiBlockLattice3D<T,Descriptor>*>& multiBlocks,
                      MultiGridManagement3D management)
{
    std::vector<std::vector<Box3D> > fineGridInterfaces = management.getFineInterface();
    std::vector<std::vector<Box3D> > coarseGridInterfaces = management.getCoarseInterface();
    
    // Create the part of the coupling in which a lattice acts as a coarse lattice.
    for (pluint iCoarse=0; iCoarse<multiBlocks.size()-1; ++iCoarse) {
        for (pluint iInterface=0; iInterface<coarseGridInterfaces[iCoarse].size(); ++iInterface) {
            createCoarseGridInterface( iCoarse, coarseGridInterfaces[iCoarse][iInterface], multiBlocks );
        }
    }
    // Create the part of the coupling in which a lattice acts as a fine lattice.
    for (pluint iFine=1; iFine<multiBlocks.size(); ++iFine) {
        for (pluint iInterface=0; iInterface<fineGridInterfaces[iFine].size(); ++iInterface) {
            createFineGridInterface( iFine-1, fineGridInterfaces[iFine][iInterface], multiBlocks );
        }
    }
}


template<typename T, template<typename U> class Descriptor>
void createCoarseGridInterface (
        plint coarseLevel, Box3D coarseGridInterface, std::vector<MultiBlockLattice3D<T,Descriptor>*>& multiBlocks )
{
    PLB_PRECONDITION( coarseGridInterface.getNx()==1 ||
                      coarseGridInterface.getNy()==1 ||
                      coarseGridInterface.getNz()==1 );
                      
    pluint fineLevel = coarseLevel + 1;
    PLB_PRECONDITION( fineLevel < multiBlocks.size() );

    MultiBlockLattice3D<T,Descriptor>& coarseLattice = *multiBlocks[coarseLevel];
    MultiBlockLattice3D<T,Descriptor>& fineLattice   = *multiBlocks[fineLevel];

    // Number of time steps executed by the fine grid during a coarse iteration.
    plint numTimeSteps = 2; 
    plint executionTime = numTimeSteps-1;
    // Add a data processor which copies and rescales from the fine to the coarse lattice.
    plint scalingOrder=1;
    integrateProcessingFunctional (
            new CopyFineToCoarse3D<T, Descriptor,Descriptor> (
                new ConvectiveRescaleEngine<T,Descriptor>(scalingOrder), numTimeSteps, executionTime),
            coarseGridInterface.multiply(2),
            fineLattice, coarseLattice );
}


template<typename T, template<typename U> class Descriptor>
void createFineGridInterface (
        plint coarseLevel, Box3D fineGridInterface, std::vector<MultiBlockLattice3D<T,Descriptor>*>& multiBlocks )
{
    plint scalingOrder=1;
    RescaleEngine<T,Descriptor>* rescaleEngine = new ConvectiveRescaleEngine<T,Descriptor>(scalingOrder);
            
    PLB_PRECONDITION( fineGridInterface.getNx()==1 || fineGridInterface.getNy()==1 || fineGridInterface.getNz()==1);
    pluint fineLevel = coarseLevel + 1;
    PLB_PRECONDITION( fineLevel < multiBlocks.size() );

    MultiBlockLattice3D<T,Descriptor>& coarseLattice = *multiBlocks[coarseLevel];
    MultiBlockLattice3D<T,Descriptor>& fineLattice   = *multiBlocks[fineLevel];

    plint direction;
    if (fineGridInterface.getNx()==1) direction = 0;
    if (fineGridInterface.getNy()==1) direction = 1;
    if (fineGridInterface.getNz()==1) direction = 2;

    // Number of time steps executed by the fine grid during a coarse iteration.
    plint numTimeSteps = 2; 
    plint executionTime = numTimeSteps-1;

    // Instantiate specific time-interpolation dynamics on the interface of the fine grid.
    setCompositeDynamics (
            fineLattice,
            fineGridInterface.multiply(2),
            new FineGridBoundaryDynamics<T,Descriptor> (
                new NoDynamics<T,Descriptor>, fineLattice.getTimeCounter(), numTimeSteps,
                rescaleEngine->getDecompositionOrder() ) );

    
    // The coarse processors are at processing level 1, because they need to access information
    //   from within the coarse envelope.
    plint processorLevelCoarse=1;
   
    //TODO add interpolation processors
    
    // Add a data processor which imposes a time-cyclic behavior on the fine grid
    //   boundary-dynamics.
    integrateProcessingFunctional (
            new Copy_t1_to_t0_3D<T, Descriptor>(numTimeSteps, executionTime),
            fineGridInterface.multiply(2),
            fineLattice );

    delete rescaleEngine;
}



template<typename T>
std::vector<MultiScalarField3D<T>*> generateScalarFields(
                        MultiGridManagement3D const& management,
                        std::vector<BlockCommunicator3D*> communicators,
                        std::vector<CombinedStatistics*> combinedStatistics )
{
    std::vector<std::vector<Box3D> > bulks = management.getBulks(); 
    std::vector<std::vector<Box3D> > uniqueBulks = management.getUniqueBulks(); 
    std::vector<std::vector<plint> > mpiProcesses = management.getMpiProcesses();
    PLB_PRECONDITION (bulks.size() == uniqueBulks.size());
    PLB_PRECONDITION (mpiProcesses.size()==0 || mpiProcesses.size()==bulks.size());
    plint envelopeWidth=1;
    std::vector<MultiScalarField3D<T>*> scalars(bulks.size());
    for (plint iLevel=0; iLevel<(plint)bulks.size(); ++iLevel){
        PLB_ASSERT(mpiProcesses.size()==0 || mpiProcesses[iLevel].size()==bulks[iLevel].size());
        ExplicitThreadAttribution* threadAttribution = new ExplicitThreadAttribution;
        SparseBlockStructure3D geometry(management.getBoundingBox(iLevel));
        for (pluint iBlock=0; iBlock<bulks[iLevel].size(); ++iBlock) {
            plint mpiProcess = mpiProcesses.size() > 0 ? mpiProcesses[iLevel][iBlock] : 0;
            plint blockId = geometry.nextIncrementalId();
            geometry.addBlock(bulks[iLevel][iBlock], bulks[iLevel][iBlock], blockId);
            threadAttribution->addBlock(blockId, mpiProcess);
        }
        MultiBlockManagement3D blockManagement (
                geometry, threadAttribution,
                envelopeWidth, iLevel );
        scalars[iLevel] = new MultiScalarField3D<T>(
                                blockManagement,communicators[iLevel],
                                combinedStatistics[iLevel], 
                                defaultMultiBlockPolicy3D().getMultiScalarAccess<T>());  
        
    }
    
    return scalars;
}


template<typename T, int nDim>
std::vector<MultiTensorField3D<T,nDim> *> generateTensorFields (
        MultiGridManagement3D const& management,
        std::vector<BlockCommunicator3D*> communicators,
        std::vector<CombinedStatistics*> combinedStatistics )
{
    std::vector<MultiTensorField3D<T,nDim>* > fields(management.getNumLevels());
    std::vector<std::vector<Box3D> > bulks = management.getBulks(); 
    std::vector<std::vector<plint> > ids = management.getMpiProcesses();
    PLB_PRECONDITION (ids.size()==0 || ids.size()==bulks.size());
    plint envelopeWidth=1;
    fields.resize(bulks.size());
    for (plint iLevel=0; iLevel<(plint)bulks.size(); ++iLevel){
        ExplicitThreadAttribution* threadAttribution = new ExplicitThreadAttribution;
        PLB_ASSERT(ids.size()==0 || ids[iLevel].size()==bulks[iLevel].size());
        SparseBlockStructure3D geometry(management.getBoundingBox(iLevel));
        for (pluint iBlock=0; iBlock<bulks[iLevel].size(); ++iBlock) {
            plint mpiProcess = ids.size() > 0 ? ids[iLevel][iBlock] : 0;
            plint blockId = geometry.nextIncrementalId();
            geometry.addBlock(bulks[iLevel][iBlock], blockId);
            threadAttribution->addBlock(blockId, mpiProcess);
        }
       
        MultiBlockManagement3D blockManagement (
                geometry, threadAttribution,
                envelopeWidth, iLevel );
        fields[iLevel] = new MultiTensorField3D<T,nDim>(
                                    blockManagement,   
                                    communicators[iLevel],
                                    combinedStatistics[iLevel], 
                                    defaultMultiBlockPolicy3D().getMultiTensorAccess<T,nDim>());  
        
    }
    
    return fields;
}



} // namespace plb

#endif  // MULTI_GRID_GENERATOR_3D_HH

