/* This file is part of the Palabos library.
 *
 * Copyright (C) 2011 FlowKit Sarl
 * Avenue de Chailly 23
 * 1012 Lausanne, Switzerland
 * E-mail contact: contact@flowkit.com
 *
 * The most recent release of Palabos can be downloaded at 
 * <http://www.palabos.org/>
 *
 * The library Palabos is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * The library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/* Main author: Daniel Lagrava
 **/

/** \file
 * Management of multi grid domains -- Header file
 */

#ifndef MULTI_GRID_MANAGEMENT_2D_H
#define MULTI_GRID_MANAGEMENT_2D_H

#include "core/globalDefs.h"
#include "core/geometry2D.h"
#include "multiGrid/multiScale.h"

namespace plb {

class MultiGridManagement2D {
public:
    MultiGridManagement2D(plint coarseNx, plint coarseNy, 
                          plint numLevels, plint overlapWidth_=1, plint referenceLevel=0);
    MultiGridManagement2D(Box2D coarseBoundingBox, plint numLevels,  plint overlapWidth_=1, plint referenceLevel=0);
    
    MultiGridManagement2D(MultiGridManagement2D const& rhs);
    ~MultiGridManagement2D();
    
    void swap(MultiGridManagement2D& rhs);
    
    /// Retrieve the blocks for each level
    std::vector<Box2D> const& getBulks(plint iLevel) const;
    std::vector<std::vector<Box2D> > const& getBulks() const;
    
    /// Retrieve the unique blocks for each level
    std::vector<Box2D> const& getUniqueBulks(plint iLevel) const;
    std::vector<std::vector<Box2D> > const& getUniqueBulks() const;
    
    /// Retrieve the bounding box at any given level
    Box2D getBoundingBox(plint level) const;
    
    /// Add a refinement in level at a given domain
    void refine(plint level, Box2D domain);
    /// Coarsen a given domain in certain level
    void coarsen(plint level, Box2D domain);

    /// Add a refinement patch at a certain level. Leave level unchanged
    void refineMultiGrid(plint level, Box2D domain);
    
    /// Store the overlap coordinates for further usage
    /// Get the Ids of the parallelized regions
    std::vector<std::vector<plint> > const& getMpiProcesses() const;
    
    std::vector<std::vector<Box2D> > const& getCoarseInterface() const;
    std::vector<std::vector<Box2D> > const& getFineInterface() const;
    std::vector<std::vector<Array<plint,2> > > const& getCoarseOrientation() const;
    std::vector<std::vector<Array<plint,2> > > const& getFineOrientation() const;
    
    void trimDomain (plint whichLevel, Box2D& domain,bool& touchLeft, 
                     bool& touchRight, bool& touchBottom, bool& touchTop ) const;

    plint getNumLevels() const;
    plint getReferenceLevel() const;
    plint getOverlapWidth() const;
    
    /// Parallelize along x-axis.
    /** The domain is cut into slices with x-coordinates ranging from finePositions[i] to
     *  finePositions[i+1]-1. These coordinates are expressed in the coordinates of the
     *  finest lattice.
     */
    void parallelizeX(std::vector<plint> const& finePositions);
    /// Parallelize along y-axis.
    /** The domain is cut into slices with y-coordinates ranging from finePositions[i] to
     *  finePositions[i+1]-1. These coordinates are expressed in the coordinates of the
     *  finest lattice.
     */
    void parallelizeY(std::vector<plint> const& finePositions);
    
    /// Parallelize according to the indicated domains, measured in coordinates of the finest lattice.
    void parallelize(std::vector<Box2D> parallelRegions, std::vector<plint> const& regionIDs);
    
    /// Eliminate the redundant fine interfaces
    void eliminateUnnecessaryFineInterfaces();

    friend MultiGridManagement2D extractManagement(MultiGridManagement2D management, Box2D coarsestDomain, bool crop);
    
private:
    void initialize(Box2D const& level0_box);
    void parallelizeLevel(plint whichLevel, std::vector<Box2D> const& parallelRegions, 
                          std::vector<plint> const& regionIDs);

private:
    plint                                overlapWidth;
    plint                                referenceLevel;
    
    /// All the bounding boxes
    std::vector<Box2D>                   boundingBoxes;
    
    /// Cells which form an interface with a grid of next-finer level.
    /** "Interface on which I am a coarse grid." */
    std::vector<std::vector<Box2D> >     coarseGridInterfaces;
    /// Cells which form an interface with a grid of next-coarser level.
    /** "Interface on which I am a fine grid." */
    std::vector<std::vector<Box2D> >     fineGridInterfaces;

    /// Geometry specification: list all atomic-blocks in each multi-block.
    std::vector<std::vector<Box2D> >     bulks;
    
    /// directions and orientations of the regions
    std::vector<std::vector<Array<plint,2> >  > coarseInterfaceOrientations;
    std::vector<std::vector<Array<plint,2> >  > fineInterfaceOrientations;
    
    /// IDs of the blocks, in case the code is parallelized.
    std::vector<std::vector<plint> >     mpiProcess;
    
    MultiScaleManager *scaleManager; 
};

MultiGridManagement2D extractManagement(MultiGridManagement2D management, Box2D coarsestDomain, bool crop=true);


} // namespace plb

#endif  // MULTI_GRID_MANAGEMENT_2D_H

