/* This file is part of the Palabos library.
 *
 * Copyright (C) 2011 FlowKit Sarl
 * Avenue de Chailly 23
 * 1012 Lausanne, Switzerland
 * E-mail contact: contact@flowkit.com
 *
 * The most recent release of Palabos can be downloaded at 
 * <http://www.palabos.org/>
 *
 * The library Palabos is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * The library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/* Main author: Daniel Lagrava
 **/

/** \file
 * Management of multi grid domains -- Implementation
 */

#include "multiGrid/multiGridManagement3D.h"
#include "io/parallelIO.h"

namespace plb {

enum {left=0,right,bottom,top,front,back};

MultiGridManagement3D::MultiGridManagement3D(plint nx, plint ny, plint nz, plint numLevels, plint referenceLevel_)
                                  : referenceLevel(referenceLevel_),
                                    boundingBoxes(numLevels),
                                    coarseGridInterfaces(numLevels),
                                    fineGridInterfaces(numLevels),
                                    bulks(numLevels),
                                    uniqueBulks(numLevels)
{
    PLB_ASSERT( numLevels > 0 );
    PLB_ASSERT( referenceLevel < numLevels );
    initialize(Box3D(0, nx-1, 0, ny-1, 0, nz-1));
}

MultiGridManagement3D::MultiGridManagement3D(Box3D coarseBoundingBox, plint numLevels, plint referenceLevel_)
                                  : referenceLevel(referenceLevel_),
                                    boundingBoxes(numLevels),
                                    coarseGridInterfaces(numLevels),
                                    fineGridInterfaces(numLevels),
                                    bulks(numLevels),
                                    uniqueBulks(numLevels)
{
    PLB_ASSERT( numLevels > 0 );
    PLB_ASSERT( referenceLevel < numLevels );
    initialize(coarseBoundingBox);
}

MultiGridManagement3D::~MultiGridManagement3D() {
    delete scaleManager;
}

void MultiGridManagement3D::initialize(Box3D const& level0_box)
{
    scaleManager = global::getDefaultMultiScaleManager().clone();

    // The parameters nx and ny refer to the coarsest lattice (lattice 0).
    boundingBoxes[0] = level0_box;
    // Multiply the values maxX and maxY by two to get, iteratively,
    //   the boundingBoxes of the next finer lattice respectively.
    for (pluint iDim=1; iDim<boundingBoxes.size(); ++iDim) {
        boundingBoxes[iDim] = scaleManager->scaleBox(boundingBoxes[iDim-1], 1);
    }
    // Add a block for the full domain of the first instantiated multi-block.
    bulks[referenceLevel].push_back(boundingBoxes[referenceLevel]);
}

MultiGridManagement3D::MultiGridManagement3D(MultiGridManagement3D const& rhs)
  : referenceLevel(rhs.referenceLevel),
    boundingBoxes(rhs.boundingBoxes),
    mpiProcess(rhs.mpiProcess),
    coarseGridInterfaces(rhs.coarseGridInterfaces),
    fineGridInterfaces(rhs.fineGridInterfaces),
    bulks(rhs.bulks),
    uniqueBulks(rhs.uniqueBulks),
    scaleManager(rhs.scaleManager->clone())
{}

void MultiGridManagement3D::swap(MultiGridManagement3D& rhs)
{
    std::swap(referenceLevel, rhs.referenceLevel);
    boundingBoxes.swap(rhs.boundingBoxes);
    mpiProcess.swap(rhs.mpiProcess);
    coarseGridInterfaces.swap(rhs.coarseGridInterfaces);
    fineGridInterfaces.swap(rhs.fineGridInterfaces);
    bulks.swap(rhs.bulks);
    uniqueBulks.swap(rhs.uniqueBulks);
    std::swap(scaleManager, rhs.scaleManager);
}

Box3D MultiGridManagement3D::getBoundingBox(plint level) const {
    PLB_PRECONDITION(level>=0 && level<(plint)bulks.size());
    return boundingBoxes[level];
}

std::vector<Box3D> const& MultiGridManagement3D::getBulks(plint iLevel) const{
    PLB_PRECONDITION( iLevel>=0 && iLevel<(plint)bulks.size() );
    return bulks[iLevel];
}

std::vector<std::vector<Box3D> > const& MultiGridManagement3D::getBulks() const{
  return bulks;
}

std::vector<Box3D> const& MultiGridManagement3D::getUniqueBulks(plint iLevel) const{
    PLB_PRECONDITION( iLevel>=0 && iLevel<(plint)uniqueBulks.size() );
    return uniqueBulks[iLevel];
}

std::vector<std::vector<Box3D> > const& MultiGridManagement3D::getUniqueBulks() const{
  return uniqueBulks;
}

void MultiGridManagement3D::refine(plint coarseLevel, Box3D coarseDomain){
    
    // The finest multi-block, at level blocks.size()-1, cannot be further refined.
    PLB_PRECONDITION( coarseLevel>=0 && coarseLevel<(plint)bulks.size()-1 );
    plint fineLevel = coarseLevel+1;

    // container for booleans that indicate if coarseDomain touches the six walls of the
    // domain. The convention is the following: 0=left,1=right,2=bottom,3=top,4=back,5=front
    bool touches[6] = {false,false,false,false,false,false};
    trimDomain(coarseLevel, coarseDomain, touches);

    // The reduced coarse domain is the one which is going to be excluded from
    //   the original coarse lattice.
    Box3D reducedCoarseDomain(coarseDomain.enlarge(-1));
    // The extended coarse domain it the one which is going to be added
    //   to the original fine lattice.
    Box3D extendedCoarseDomain(coarseDomain.enlarge(1));

    // Same as coarseDomain, but in units of the fine lattice.
    Box3D fineDomain(coarseDomain.multiply(2));
    // Extend the fineDomain by one node in fine lattice units, in
    //   view of defining the fine lattice's unique bulk.
    Box3D intermediateFineDomain(fineDomain.enlarge(-1));

    // If the domain in question touches a boundary of the multi-block,
    //   both the reduced and the extended coarse domain are
    //   identified with the boundary location.
    if (touches[left]) { // left
        reducedCoarseDomain.x0    -= 1;
        extendedCoarseDomain.x0   += 1;
        intermediateFineDomain.x0 -= 1;
    }
    if (touches[right]) { // right
        reducedCoarseDomain.x1    += 1;
        extendedCoarseDomain.x1   -= 1;
        intermediateFineDomain.x1 += 1;
    }
    if (touches[bottom]) { // bottom
        reducedCoarseDomain.y0    -= 1;
        extendedCoarseDomain.y0   += 1;
        intermediateFineDomain.y0 -= 1;
    }
    if (touches[top]) { // top
        reducedCoarseDomain.y1    += 1;
        extendedCoarseDomain.y1   -= 1;
        intermediateFineDomain.y1 += 1;
    }
    if (touches[front]) { // front
        reducedCoarseDomain.z0    -= 1;
        extendedCoarseDomain.z0   += 1;
        intermediateFineDomain.z0 -= 1;
    }
    if (touches[back]) { // back
        reducedCoarseDomain.z1    += 1;
        extendedCoarseDomain.z1   -= 1;
        intermediateFineDomain.z1 += 1;
    }

    // Extract reduced coarse domain from the original coarse multi-block, for the bulks.
    std::vector<Box3D> exceptedBulks;
    for (pluint iBlock=0; iBlock<bulks[coarseLevel].size(); ++iBlock) {
        except(bulks[coarseLevel][iBlock], reducedCoarseDomain, exceptedBulks);
    }
    exceptedBulks.swap(bulks[coarseLevel]);
    
    // Extract reduced coarse domain from the original coarse multi-block, for the unique bulks.
    //  this implies again as in 2D that uniqueBulks == bulks
    std::vector<Box3D> exceptedUniqueBulks;
    for (pluint iBlock=0; iBlock<uniqueBulks[coarseLevel].size(); ++iBlock) {
        except(uniqueBulks[coarseLevel][iBlock], reducedCoarseDomain, exceptedUniqueBulks);
    }
    exceptedUniqueBulks.swap(uniqueBulks[coarseLevel]);
    
    // Convert the extended coarse domain to fine units,
    //   and add to the original fine multi-block.
    Box3D extendedFineDomain(extendedCoarseDomain.multiply(2));
    bulks[fineLevel].push_back(extendedFineDomain);
    // the uniqueBulk of the fine domain is the so called intermediateFineDomain
    //  the overlappin region is removed from the fine grid always.
    uniqueBulks[fineLevel].push_back(intermediateFineDomain);

    // Define coupling interfaces for all six sides of the coarsened domain, unless they
    //   touch a boundary of the multi-block.
    if (!touches[left]) { ///LEFT
        coarseGridInterfaces[coarseLevel].push_back( Box3D( coarseDomain.x0, coarseDomain.x0,
                                                            coarseDomain.y0, coarseDomain.y1,
                                                            coarseDomain.z0, coarseDomain.z1) );
        fineGridInterfaces[fineLevel].push_back( Box3D( extendedCoarseDomain.x0, extendedCoarseDomain.x0, 
                                                        extendedCoarseDomain.y0, extendedCoarseDomain.y1,
                                                        extendedCoarseDomain.z0, extendedCoarseDomain.z1) );
    }
    if (!touches[right]) { ///RIGHT
        coarseGridInterfaces[coarseLevel].push_back( Box3D( coarseDomain.x1, coarseDomain.x1,
                                                            coarseDomain.y0, coarseDomain.y1,
                                                            coarseDomain.z0, coarseDomain.z1) );
        fineGridInterfaces[fineLevel].push_back( Box3D( extendedCoarseDomain.x1, extendedCoarseDomain.x1, 
                                                        extendedCoarseDomain.y0, extendedCoarseDomain.y1,
                                                        extendedCoarseDomain.z0, extendedCoarseDomain.z1) );
    }
    if (!touches[bottom]) { ///BOTTOM
        coarseGridInterfaces[coarseLevel].push_back( Box3D( coarseDomain.x0, coarseDomain.x1,
                                                            coarseDomain.y0, coarseDomain.y0,
                                                            coarseDomain.z0, coarseDomain.z1) );
        fineGridInterfaces[fineLevel].push_back( Box3D( extendedCoarseDomain.x0, extendedCoarseDomain.x1, 
                                                        extendedCoarseDomain.y0, extendedCoarseDomain.y0,
                                                        extendedCoarseDomain.z0, extendedCoarseDomain.z1) );
    }
    if (!touches[top]) { ///TOP
        coarseGridInterfaces[coarseLevel].push_back( Box3D( coarseDomain.x0, coarseDomain.x1,
                                                            coarseDomain.y1, coarseDomain.y1,
                                                            coarseDomain.z0, coarseDomain.z1) );
        fineGridInterfaces[fineLevel].push_back( Box3D( extendedCoarseDomain.x0, extendedCoarseDomain.x1, 
                                                        extendedCoarseDomain.y1, extendedCoarseDomain.y1,
                                                        extendedCoarseDomain.z0, extendedCoarseDomain.z1) );
    }
    if (!touches[front]) { ///FRONT
        coarseGridInterfaces[coarseLevel].push_back( Box3D( coarseDomain.x0, coarseDomain.x1,
                                                            coarseDomain.y0, coarseDomain.y1,
                                                            coarseDomain.z0, coarseDomain.z0) );
        fineGridInterfaces[fineLevel].push_back( Box3D( extendedCoarseDomain.x0, extendedCoarseDomain.x1, 
                                                        extendedCoarseDomain.y0, extendedCoarseDomain.y1,
                                                        extendedCoarseDomain.z0, extendedCoarseDomain.z0) );
    }
    if (!touches[back]) { ///BACK
        coarseGridInterfaces[coarseLevel].push_back( Box3D( coarseDomain.x0, coarseDomain.x1,
                                                            coarseDomain.y0, coarseDomain.y1,
                                                            coarseDomain.z1, coarseDomain.z1) );
        fineGridInterfaces[fineLevel].push_back( Box3D( extendedCoarseDomain.x0, extendedCoarseDomain.x1, 
                                                        extendedCoarseDomain.y0, extendedCoarseDomain.y1,
                                                        extendedCoarseDomain.z1, extendedCoarseDomain.z1) );
    }
}

void MultiGridManagement3D::coarsen(plint fineLevel, Box3D coarseDomain){
    PLB_PRECONDITION( fineLevel>=1 && fineLevel<(plint)bulks.size() );
    plint coarseLevel = fineLevel-1;

    // First, trim the domain coarseDomain in case it exceeds the extent of the
    //   multi-block, and determine whether coarseDomain touches one of the boundaries
    //   of the multi-block. This information is needed, because the coarse domain
    //   fully replaces the fine domain on boundaries of the multi-block, and there
    //   is therefore no need to create a coarse-fine coupling.
    bool touches[6] = {false,false,false,false,false,false};
    trimDomain(coarseLevel, coarseDomain, touches);

    // Convert the coarse domain to fine units.
    Box3D fineDomain(coarseDomain.multiply(2));
    // The reduced fine domain is the one which is going to be excluded from
    //   the original fine lattice.
    Box3D intermediateFineDomain(fineDomain.enlarge(-1));
    // The extended coarse domain it the one which is going to be added
    //   to the original coarse lattice.
    Box3D extendedCoarseDomain(coarseDomain.enlarge(1));

    // If the domain in question touches a boundary of the multi-block,
    //   both the reduced fine domain and the extended coarse domain are
    //   identified with the boundary location.
    // If the domain in question touches a boundary of the multi-block,
    //   both the reduced and the extended coarse domain are
    //   identified with the boundary location.
    if (touches[left]) { // left
        extendedCoarseDomain.x0 += 1;
        intermediateFineDomain.x0 -= 1;
    }
    if (touches[right]) { // right
        extendedCoarseDomain.x1 -= 1;
        intermediateFineDomain.x1 += 1;
    }
    if (touches[bottom]) { // bottom
        extendedCoarseDomain.y0 += 1;
        intermediateFineDomain.y0 -= 1;
    }
    if (touches[top]) { // top
        extendedCoarseDomain.y1 -= 1;
        intermediateFineDomain.y1 += 1;
    }
    if (touches[front]) { // front
        extendedCoarseDomain.z0 += 1;
        intermediateFineDomain.z0 -= 1;
    }
    if (touches[back]) { // back
        extendedCoarseDomain.z1 -= 1;
        intermediateFineDomain.z1 += 1;
    }

    // Extract reduced fine domain from the original fine multi-block.
    std::vector<Box3D> exceptedBlocks;
    for (pluint iBlock=0; iBlock<bulks[fineLevel].size(); ++iBlock) {
        except(bulks[fineLevel][iBlock], intermediateFineDomain, exceptedBlocks);
    }
    exceptedBlocks.swap(bulks[fineLevel]);

    // Extract the same domain to get the unique bulks
    std::vector<Box3D> exceptedUniqueBlocks;
    for (pluint iBlock=0; iBlock<uniqueBulks[fineLevel].size(); ++iBlock) {
        except(uniqueBulks[fineLevel][iBlock], intermediateFineDomain, exceptedUniqueBlocks); 
    }
    exceptedUniqueBlocks.swap(uniqueBulks[fineLevel]);
    
    // Add extended coarse domain to the original coarse multi-block.
    bulks[coarseLevel].push_back(extendedCoarseDomain);
    // Corresponding operation for unique coarse bulk.
    uniqueBulks[coarseLevel].push_back(coarseDomain.enlarge(-1));

    // Define coupling interfaces for all four sides of the refined domain, unless they
    //   touch a boundary of the multi-block.
    if (!touches[left]) { ///LEFT
        fineGridInterfaces[fineLevel].push_back( Box3D( coarseDomain.x0, coarseDomain.x0,
                                                            coarseDomain.y0, coarseDomain.y1,
                                                            coarseDomain.z0, coarseDomain.z1) );
        coarseGridInterfaces[coarseLevel].push_back( Box3D( extendedCoarseDomain.x0, extendedCoarseDomain.x0, 
                                                          extendedCoarseDomain.y0, extendedCoarseDomain.y1,
                                                          extendedCoarseDomain.z0, extendedCoarseDomain.z1) );
    }
    if (!touches[right]) { ///RIGHT
        fineGridInterfaces[fineLevel].push_back( Box3D( coarseDomain.x1, coarseDomain.x1,
                                                          coarseDomain.y0, coarseDomain.y1,
                                                          coarseDomain.z0, coarseDomain.z1) );
        coarseGridInterfaces[coarseLevel].push_back( Box3D( extendedCoarseDomain.x1, extendedCoarseDomain.x1, 
                                                        extendedCoarseDomain.y0, extendedCoarseDomain.y1,
                                                        extendedCoarseDomain.z0, extendedCoarseDomain.z1) );
    }
    if (!touches[bottom]) { ///BOTTOM
        fineGridInterfaces[fineLevel].push_back( Box3D( coarseDomain.x0, coarseDomain.x1,
                                                          coarseDomain.y0, coarseDomain.y0,
                                                          coarseDomain.z0, coarseDomain.z1) );
        coarseGridInterfaces[coarseLevel].push_back( Box3D( extendedCoarseDomain.x0, extendedCoarseDomain.x1, 
                                                        extendedCoarseDomain.y0, extendedCoarseDomain.y0,
                                                        extendedCoarseDomain.z0, extendedCoarseDomain.z1) );
    }
    if (!touches[top]) { ///TOP
        fineGridInterfaces[fineLevel].push_back( Box3D( coarseDomain.x0, coarseDomain.x1,
                                                          coarseDomain.y1, coarseDomain.y1,
                                                          coarseDomain.z0, coarseDomain.z1) );
        coarseGridInterfaces[coarseLevel].push_back( Box3D( extendedCoarseDomain.x0, extendedCoarseDomain.x1, 
                                                          extendedCoarseDomain.y1, extendedCoarseDomain.y1,
                                                          extendedCoarseDomain.z0, extendedCoarseDomain.z1) );
    }
    if (!touches[front]) { ///FRONT
        fineGridInterfaces[fineLevel].push_back( Box3D( coarseDomain.x0, coarseDomain.x1,
                                                          coarseDomain.y0, coarseDomain.y1,
                                                          coarseDomain.z0, coarseDomain.z0) );
        coarseGridInterfaces[coarseLevel].push_back( Box3D( extendedCoarseDomain.x0, extendedCoarseDomain.x1, 
                                                          extendedCoarseDomain.y0, extendedCoarseDomain.y1,
                                                          extendedCoarseDomain.z0, extendedCoarseDomain.z0) );
    }
    if (!touches[back]) { ///BACK
        fineGridInterfaces[fineLevel].push_back( Box3D( coarseDomain.x0, coarseDomain.x1,
                                                          coarseDomain.y0, coarseDomain.y1,
                                                          coarseDomain.z1, coarseDomain.z1) );
        coarseGridInterfaces[coarseLevel].push_back( Box3D( extendedCoarseDomain.x0, extendedCoarseDomain.x1, 
                                                            extendedCoarseDomain.y0, extendedCoarseDomain.y1,
                                                            extendedCoarseDomain.z1, extendedCoarseDomain.z1) );
    }
}

std::vector<std::vector<Box3D> > const& MultiGridManagement3D::getCoarseInterface() const
{
    return coarseGridInterfaces;
}

std::vector<std::vector<Box3D> > const& MultiGridManagement3D::getFineInterface() const
{
    return fineGridInterfaces;
}


void MultiGridManagement3D::trimDomain (plint whichLevel, Box3D& domain,
                                        bool* touches) const
{
    Box3D bbox = boundingBoxes[whichLevel];
    /// LEFT
    if (domain.x0 <= bbox.x0) {
        domain.x0 = bbox.x0; // Trim in case multi-block extent is exceeded.
        touches[left] = true;
    }
    /// BOTTOM
    if (domain.y0 <= bbox.y0) {
        domain.y0 = bbox.y0; // Trim in case multi-block extent is exceeded.
        touches[bottom] = true;
    }
    /// FRONT
    if (domain.z0 <= bbox.z0) {
        domain.z0 = bbox.z0; // Trim in case multi-block extent is exceeded.
        touches[front] = true;
    }
    /// RIGHT
    if (domain.x1 >= bbox.x1) {
        domain.x1 = bbox.x1; // Trim in case multi-block extent is exceeded.
        touches[right] = true;
    }
    /// TOP
    if (domain.y1 >= bbox.y1) {
        domain.y1 = bbox.y1; // Trim in case multi-block extent is exceeded.
        touches[top] = true;
    }
    /// BACK
    if (domain.z1 >= bbox.z1) {
        domain.z1 = bbox.z1; // Trim in case multi-block extent is exceeded.
        touches[back] = true;
    }
}

plint MultiGridManagement3D::getNumLevels() const {
    return (plint) bulks.size();
}

plint MultiGridManagement3D::getReferenceLevel() const {
    return referenceLevel;
}


void MultiGridManagement3D::parallelizeX(std::vector<plint> const& finePositions)
{
    // Retrieve dimensions of the finest lattice.
    plint finestLevel=bulks.size()-1;
    plint nx = boundingBoxes[finestLevel].getNx();
    plint ny = boundingBoxes[finestLevel].getNy();
    plint nz = boundingBoxes[finestLevel].getNz();

    // Add the left and the right boundary to the interface positions before creating the parallel regions.
    std::vector<plint> allPositions(finePositions.size()+2);
    allPositions[0] = 0;
    std::copy(finePositions.begin(), finePositions.end(), allPositions.begin()+1);
    allPositions.back() = nx;

    std::vector<Box3D> parallelRegions(allPositions.size()-1);
    std::vector<plint> regionIDs(parallelRegions.size());
    for (pluint iPos=0; iPos<allPositions.size()-1; ++iPos) {
        plint x0 = allPositions[iPos];
        plint x1 = allPositions[iPos+1]-1;
        parallelRegions[iPos] = Box3D(x0, x1, 0, ny-1, 0, nz-1);
        regionIDs[iPos] = iPos;
    }
    parallelize(parallelRegions, regionIDs);
}


void MultiGridManagement3D::parallelizeY(std::vector<plint> const& finePositions)
{
    // Retrieve dimensions of the finest lattice.
    plint finestLevel=bulks.size()-1;
    plint nx = boundingBoxes[finestLevel].getNx();
    plint ny = boundingBoxes[finestLevel].getNy();
    plint nz = boundingBoxes[finestLevel].getNz();
    
    // Add the left and the right boundary to the interface positions before creating the parallel regions.
    std::vector<plint> allPositions(finePositions.size()+2);
    allPositions[0] = 0;
    std::copy(finePositions.begin(), finePositions.end(), allPositions.begin()+1);
    allPositions.back() = ny;

    std::vector<Box3D> parallelRegions(allPositions.size()-1);
    std::vector<plint> regionIDs(parallelRegions.size());
    for (pluint iPos=0; iPos<allPositions.size()-1; ++iPos) {
        plint y0 = allPositions[iPos];
        plint y1 = allPositions[iPos+1]-1;
        parallelRegions[iPos] = Box3D(0, nx-1, y0, y1, 0, nz-1);
        regionIDs[iPos] = iPos;
    }
    parallelize(parallelRegions, regionIDs);
}

void MultiGridManagement3D::parallelizeZ(std::vector<plint> const& finePositions)
{
    // Retrieve dimensions of the finest lattice.
    plint finestLevel=bulks.size()-1;
    plint nx = boundingBoxes[finestLevel].getNx();
    plint ny = boundingBoxes[finestLevel].getNy();
    plint nz = boundingBoxes[finestLevel].getNz();
    
    // Add the left and the right boundary to the interface positions before creating the parallel regions.
    std::vector<plint> allPositions(finePositions.size()+2);
    allPositions[0] = 0;
    std::copy(finePositions.begin(), finePositions.end(), allPositions.begin()+1);
    allPositions.back() = nz;

    std::vector<Box3D> parallelRegions(allPositions.size()-1);
    std::vector<plint> regionIDs(parallelRegions.size());
    for (pluint iPos=0; iPos<allPositions.size()-1; ++iPos) {
        plint z0 = allPositions[iPos];
        plint z1 = allPositions[iPos+1]-1;
        parallelRegions[iPos] = Box3D(0, nx-1, 0, ny-1, z0, z1 );
        regionIDs[iPos] = iPos;
    }
    parallelize(parallelRegions, regionIDs);
}



void MultiGridManagement3D::parallelize (
        std::vector<Box3D> parallelRegions,
        std::vector<plint> const& regionIDs )
{
    plint finestLevel=bulks.size()-1;
    for (plint iLevel=finestLevel; iLevel>=0; --iLevel) {
        parallelizeLevel(iLevel, parallelRegions, regionIDs);
        // Adapt the regions to the next-coarser level.
        for (pluint iRegion=0; iRegion<parallelRegions.size(); ++iRegion) {
            parallelRegions[iRegion] = parallelRegions[iRegion].divideAndFitSmaller(2);
        }
    }
}

void MultiGridManagement3D::parallelizeLevel (
        plint whichLevel, std::vector<Box3D> const& parallelRegions,
        std::vector<plint> const& regionIDs )
{
    PLB_PRECONDITION( parallelRegions.size() == regionIDs.size() );
    PLB_PRECONDITION( whichLevel < (plint)bulks.size() );
    std::vector<Box3D> newBlocks;
    // IDs are going to be reattributed at the level whichLevel.
    if (mpiProcess.size() <= (pluint)whichLevel) {
        mpiProcess.resize(whichLevel+1);
    }
    mpiProcess[whichLevel].clear();
    for (pluint iRegion=0; iRegion<parallelRegions.size(); ++iRegion) {
        plint currentId = regionIDs[iRegion];
        for (pluint iBlock=0; iBlock<bulks[whichLevel].size(); ++iBlock) {
            Box3D intersection;
            if ( intersect( bulks[whichLevel][iBlock],
                            parallelRegions[iRegion], intersection ) )
            {
                newBlocks.push_back(intersection);
                mpiProcess[whichLevel].push_back(currentId);
            }
        }
    }
    bulks[whichLevel].swap(newBlocks);
}

std::vector<std::vector<plint> > const& MultiGridManagement3D::getMpiProcesses() const{
  return mpiProcess;
}

/// Extract a domain (in coarse coordinates) from a MultiGridManagement3D
/** This is achieved by taking all the geometric information of the MultiGridManagement3D
 *  and intersecting it with the so called coarsestDomain. All the intersecting parts
 *  are kept on the new MultiGridManagement3D.
 */
MultiGridManagement3D extractManagement(MultiGridManagement3D management, Box3D coarsestDomain, bool crop){
    //TODO: implement
    return management;
}


} // namespace plb

