/* This file is part of the Palabos library.
 *
 * Copyright (C) 2011 FlowKit Sarl
 * Avenue de Chailly 23
 * 1012 Lausanne, Switzerland
 * E-mail contact: contact@flowkit.com
 *
 * The most recent release of Palabos can be downloaded at 
 * <http://www.palabos.org/>
 *
 * The library Palabos is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * The library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/* Main author: Andrea */

#ifndef FREE_SURFACE_INITIALIZER_3D_H
#define FREE_SURFACE_INITIALIZER_3D_H

#include "core/globalDefs.h"
#include "atomicBlock/dataProcessingFunctional3D.h"
#include "multiPhysics/freeSurfaceUtil3D.h"

namespace plb {

template< typename T,template<typename U> class Descriptor>
class DefaultInitializeFreeSurface3D : public BoxProcessingFunctional3D {
public:
    DefaultInitializeFreeSurface3D(Dynamics<T,Descriptor>* dynamicsTemplate_, Array<T,3> g_)
        : dynamicsTemplate(dynamicsTemplate_), g(g_)
    { }
    DefaultInitializeFreeSurface3D(DefaultInitializeFreeSurface3D<T,Descriptor> const& rhs)
        : dynamicsTemplate(rhs.dynamicsTemplate->clone()),
          g(rhs.g)
    { }
    DefaultInitializeFreeSurface3D<T,Descriptor>* operator=(DefaultInitializeFreeSurface3D<T,Descriptor> const& rhs)
    { 
        DefaultInitializeFreeSurface3D<T,Descriptor>(rhs).swap(*this);
        return *this;
    }
    void swap(DefaultInitializeFreeSurface3D<T,Descriptor>& rhs) {
        std::swap(dynamicsTemplate, rhs.dynamicsTemplate);
        std::swap(g, rhs.g);
    }
    ~DefaultInitializeFreeSurface3D() {
        delete dynamicsTemplate;
    }
    virtual void processGenericBlocks(Box3D domain, std::vector<AtomicBlock3D*> atomicBlocks);
    virtual DefaultInitializeFreeSurface3D<T,Descriptor>* clone() const {
        return new DefaultInitializeFreeSurface3D(*this);
    }
    // Arguments: (0) fluid-lattice, (1) mass, (2) volume-fraction, (3) flag-status, (4) interfaceLists
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const {
        // Default-assign potential other parameters present in a multi-fluid system.
        std::fill(modified.begin(), modified.end(), modif::nothing);
        modified[0] = modif::dataStructure;    // Fluid
        modified[1] = modif::staticVariables;  // rhoBar.
        modified[2] = modif::staticVariables;  // j.
        modified[3] = modif::staticVariables;  // Mass
        modified[4] = modif::staticVariables;  // Volume-fraction
        modified[5] = modif::staticVariables;  // Flag-status
        modified[6] = modif::nothing;          // Interface lists
    }
private:
    Dynamics<T,Descriptor>* dynamicsTemplate;
    Array<T,3> g;
};

// Same as DefaultInitializeFreeSurface, but without initializing the Volume-fraction and the Flag-Status
template< typename T,template<typename U> class Descriptor>
class PartiallyDefaultInitializeFreeSurface3D : public BoxProcessingFunctional3D {
public:
    PartiallyDefaultInitializeFreeSurface3D(Dynamics<T,Descriptor>* dynamicsTemplate_, Array<T,3> g_)
        : dynamicsTemplate(dynamicsTemplate_), g(g_)
    { }
    PartiallyDefaultInitializeFreeSurface3D(PartiallyDefaultInitializeFreeSurface3D<T,Descriptor> const& rhs)
        : dynamicsTemplate(rhs.dynamicsTemplate->clone()),
          g(rhs.g)
    { }
    PartiallyDefaultInitializeFreeSurface3D<T,Descriptor>* operator=(PartiallyDefaultInitializeFreeSurface3D<T,Descriptor> const& rhs)
    { 
        PartiallyDefaultInitializeFreeSurface3D<T,Descriptor>(rhs).swap(*this);
        return *this;
    }
    void swap(PartiallyDefaultInitializeFreeSurface3D<T,Descriptor>& rhs) {
        std::swap(dynamicsTemplate, rhs.dynamicsTemplate);
        std::swap(g, rhs.g);
    }
    ~PartiallyDefaultInitializeFreeSurface3D() {
        delete dynamicsTemplate;
    }
    virtual void processGenericBlocks(Box3D domain, std::vector<AtomicBlock3D*> atomicBlocks);
    virtual PartiallyDefaultInitializeFreeSurface3D<T,Descriptor>* clone() const {
        return new PartiallyDefaultInitializeFreeSurface3D(*this);
    }
    // Arguments: (0) fluid-lattice, (1) mass, (2) volume-fraction, (3) flag-status, (4) interfaceLists
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const {
        // Default-assign potential other parameters present in a multi-fluid system.
        std::fill(modified.begin(), modified.end(), modif::nothing);
        modified[0] = modif::dataStructure;    // Fluid
        modified[1] = modif::staticVariables;  // rhoBar.
        modified[2] = modif::staticVariables;  // j.
        modified[3] = modif::staticVariables;  // Mass
        modified[4] = modif::nothing;          // Volume-fraction
        modified[5] = modif::nothing;          // Flag-status
        modified[6] = modif::nothing;          // Interface lists
    }
private:
    Dynamics<T,Descriptor>* dynamicsTemplate;
    Array<T,3> g;
};

template<typename T,class InsideFunction>
class AnalyticalIniVolumeFraction3D : public BoxProcessingFunctional3D {
public:
    AnalyticalIniVolumeFraction3D(InsideFunction const& insideFunction_, plint subDivision_=5)
        : insideFunction(insideFunction_),
          subDivision(subDivision_)
    { PLB_ASSERT( subDivision > 1 ); }
    virtual void processGenericBlocks(Box3D domain, std::vector<AtomicBlock3D*> atomicBlocks);
    virtual AnalyticalIniVolumeFraction3D<T,InsideFunction>* clone() const {
        return new AnalyticalIniVolumeFraction3D<T,InsideFunction>(*this);
    }
    // Arguments: (0) volume-fraction, (1) flag-status
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const {
        modified[0] = modif::staticVariables;  // Volume-fraction
        modified[1] = modif::staticVariables;  // Flag-status
    }
private:
    void subDomainVolumeFraction (
            plint iX, plint iY, plint iZ, int& flag, T& volumeFraction );
private:
    InsideFunction const& insideFunction;
    plint subDivision;
};

template<typename T, class InsideFunction>
void analyticalIniVolumeFraction (
        MultiScalarField3D<T>& volumeFraction, MultiScalarField3D<int>& flagStatus,
        InsideFunction const& insideFunction, plint subDivision = 5 );


template<typename T, template<typename U> class Descriptor>
class ConstantIniVelocityFreeSurface3D : public BoxProcessingFunctional3D {
public:
    ConstantIniVelocityFreeSurface3D(Array<T,3> velocity_)
        : velocity(velocity_)
    { }
    virtual void processGenericBlocks(Box3D domain, std::vector<AtomicBlock3D*> atomicBlocks);
    virtual ConstantIniVelocityFreeSurface3D<T,Descriptor>* clone() const {
        return new ConstantIniVelocityFreeSurface3D<T,Descriptor>(*this);
    }
    virtual void getTypeOfModification (std::vector<modif::ModifT>& modified) const {
        // Default-assign potential other parameters present in a multi-fluid system.
        std::fill(modified.begin(), modified.end(), modif::nothing);
        modified[0] = modif::staticVariables; // Fluid. Should be: staticVariables.
        modified[1] = modif::nothing;         // rhoBar.
        modified[2] = modif::staticVariables; // j.
        modified[3] = modif::nothing;         // Mass.
        modified[4] = modif::nothing;         // Volume-fraction.
        modified[5] = modif::nothing;         // Flag-status.
        modified[6] = modif::nothing;         // Interface-lists.
    }
private:
    Array<T,3> velocity;
};


template< typename T,template<typename U> class Descriptor>
class DefaultInitializeSecondFluid3D : public BoxProcessingFunctional3D {
public:
    DefaultInitializeSecondFluid3D(Dynamics<T,Descriptor>* dynamicsTemplate_, Array<T,3> g_)
        : dynamicsTemplate(dynamicsTemplate_), g(g_)
    { }
    DefaultInitializeSecondFluid3D(DefaultInitializeSecondFluid3D<T,Descriptor> const& rhs)
        : dynamicsTemplate(rhs.dynamicsTemplate->clone()),
          g(rhs.g)
    { }
    DefaultInitializeSecondFluid3D<T,Descriptor>* operator= (
            DefaultInitializeSecondFluid3D<T,Descriptor> const& rhs )
    { 
        DefaultInitializeSecondFluid3D<T,Descriptor>(rhs).swap(*this);
        return *this;
    }
    void swap(DefaultInitializeSecondFluid3D<T,Descriptor>& rhs) {
        std::swap(dynamicsTemplate, rhs.dynamicsTemplate);
        std::swap(g, rhs.g);
    }
    ~DefaultInitializeSecondFluid3D() {
        delete dynamicsTemplate;
    }
    virtual void processGenericBlocks(Box3D domain, std::vector<AtomicBlock3D*> atomicBlocks);
    virtual DefaultInitializeSecondFluid3D<T,Descriptor>* clone() const {
        return new DefaultInitializeSecondFluid3D(*this);
    }
    // Arguments: (0) fluid-lattice, (1) mass, (2) volume-fraction, (3) flag-status, (4) interfaceLists
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const {
        // Default-assign potential other parameters present in a multi-fluid system.
        std::fill(modified.begin(), modified.end(), modif::nothing);
        modified[0] = modif::nothing;          // Fluid
        modified[1] = modif::nothing;          // rhoBar.
        modified[2] = modif::nothing;          // j.
        modified[3] = modif::nothing;          // Mass
        modified[4] = modif::nothing;          // Volume-fraction
        modified[5] = modif::nothing;          // Flag-status
        modified[6] = modif::nothing;          // Interface lists
        modified[7] = modif::dataStructure;    // Fluid 2.
        modified[8] = modif::staticVariables;  // rhoBar2.
        modified[9] = modif::staticVariables;  // j2.
    }
private:
    Dynamics<T,Descriptor>* dynamicsTemplate;
    Array<T,3> g;
};


}  // namespace plb

#endif  // FREE_SURFACE_INITIALIZER_3D_H

