/* This file is part of the Palabos library.
 *
 * Copyright (C) 2011 FlowKit Sarl
 * Avenue de Chailly 23
 * 1012 Lausanne, Switzerland
 * E-mail contact: contact@flowkit.com
 *
 * The most recent release of Palabos can be downloaded at 
 * <http://www.palabos.org/>
 *
 * The library Palabos is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * The library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/* Main author: Andrea */

/**
 * CollideAndStream3D
 * --> populations
 * MassChange3D <-- Compute mass balance and store in mass matrix.
 * FreeSurfaceCompletion3D <-- Completion scheme on post-collide populations.
 * FreeSurfaceMacroscopic3D <-- Compute and store macroscopic variables and mass-fraction.
 * --> mass, mass-fraction, flag, density, momentum
 * ComputeInterfaceLists3D <-- Compute all lists describing ongoing changes on the interface.
 * IniInterfaceToAnyNodes3D <-- Execute necessary operations for cells switching from interface to fluid or to empty.
 * IniEmptyToInterfaceNodes3D <-- Execute necessary operations for cells switching from empty to interface.
 * --> flag, mass, density
 * RemoveFalseInterfaceCells3D <-- Remove isolated interface cells.
 * --> flag, mass, density
 * EqualMassExcessReDistribution3D <-- Enforce exact mass balance for interface cells that become fluid or empty.
 * --> Everything except density, flag
 * 
 **/

/* Main author: Andrea */

#ifndef FREE_SURFACE_MODEL_3D_H
#define FREE_SURFACE_MODEL_3D_H

#include <algorithm>
#include "core/globalDefs.h"
#include "atomicBlock/dataProcessingFunctional3D.h"
#include "multiBlock/defaultMultiBlockPolicy3D.h"
#include "multiPhysics/freeSurfaceUtil3D.h"
#include "multiPhysics/freeSurfaceInitializer3D.h"
#include "dataProcessors/dataInitializerWrapper3D.h"

namespace plb {

/// Compute the mass balance on every node in the domain, and store in mass matrix.
/** Input:
  *   - Flag-status:   needed in bulk+1
  *   - Mass:          needed in bulk
  *   - Volume fraction: needed in bulk
  *   - Populations:   needed in bulk+1
  * Output:
  *   - mass.
  **/
template< typename T,template<typename U> class Descriptor>
class MassChange3D : public BoxProcessingFunctional3D {
public:
    virtual void processGenericBlocks(Box3D domain, std::vector<AtomicBlock3D*> atomicBlocks);
    virtual MassChange3D<T,Descriptor>* clone() const {
        return new MassChange3D<T,Descriptor>(*this);
    }
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const {
        // Default-assign potential other parameters present in a multi-fluid system.
        std::fill(modified.begin(), modified.end(), modif::nothing);
        modified[0] = modif::nothing;         // Fluid.
        modified[1] = modif::nothing;         // rhoBar.
        modified[2] = modif::nothing;         // j.
        modified[3] = modif::staticVariables; // Mass.
        modified[4] = modif::nothing;         // Volume fraction.
        modified[5] = modif::nothing;         // Flag-status.
        modified[6] = modif::nothing;         // Interface-lists.
   }
};

/// Generic implementation of the completion scheme; can also be used
/// by the second fluid in a binary system.
template< typename T,template<typename U> class Descriptor, class Param>
void freeSurfaceCompletionImplementation3D(Param& param, Box3D domain);

/// Completion scheme on the post-collide populations on interface cells.
/** Input:
  *   - Flag-status:   needed in bulk+1
  *   - Volume fraction: needed in bulk+1
  *   - Populations:   needed in bulk+1
  *   - Momentum:      needed in bulk+1
  *   - Density:       needed in bulk+1
  * Output:
  *   - Populations.
  **/
// ASK: This data processor loops over the whole volume. Is this really
//      necessary, or could one of the lists be used instead?
template<typename T, template<typename U> class Descriptor>
class FreeSurfaceCompletion3D : public BoxProcessingFunctional3D {
public:
    FreeSurfaceCompletion3D(T rhoEmpty_)
        : rhoEmpty(rhoEmpty_)
    { }
    virtual void processGenericBlocks(Box3D domain, std::vector<AtomicBlock3D*> atomicBlocks);
    virtual FreeSurfaceCompletion3D<T,Descriptor>* clone() const {
        return new FreeSurfaceCompletion3D<T,Descriptor>(*this);
    }
    virtual void getTypeOfModification (std::vector<modif::ModifT>& modified) const {
        // Default-assign potential other parameters present in a multi-fluid system.
        std::fill(modified.begin(), modified.end(), modif::nothing);
        modified[0] = modif::nothing;         // Fluid. Should be: staticVariables.
        modified[1] = modif::nothing;         // rhoBar.
        modified[2] = modif::nothing;         // j.
        modified[3] = modif::nothing;         // Mass.
        modified[4] = modif::nothing;         // Volume fraction.
        modified[5] = modif::nothing;         // Flag-status.
        modified[6] = modif::nothing;         // Interface-lists.
    }
private:
    T rhoEmpty; 
};

/// Compute and store mass-fraction and macroscopic variables.
/** Input:
  *   - Flag-status:   needed in bulk
  *   - Mass:          needed in bulk
  *   - Populations:   needed in bulk
  * Output:
  *   - mass-fraction, density, momentum, flag (because setting bounce-back).
  **/
template<typename T, template<typename U> class Descriptor>
class FreeSurfaceMacroscopic3D : public BoxProcessingFunctional3D {
public:
    FreeSurfaceMacroscopic3D(T rhoEmpty_)
        : rhoEmpty(rhoEmpty_)
    { }
    virtual void processGenericBlocks(Box3D domain, std::vector<AtomicBlock3D*> atomicBlocks);
    virtual FreeSurfaceMacroscopic3D<T,Descriptor>* clone() const {
        return new FreeSurfaceMacroscopic3D<T,Descriptor>(*this);
    }
    virtual void getTypeOfModification (std::vector<modif::ModifT>& modified) const {
        // Default-assign potential other parameters present in a multi-fluid system.
        std::fill(modified.begin(), modified.end(), modif::nothing);
        modified[0] = modif::nothing;         // Fluid. Should be: staticVariables.
        modified[1] = modif::staticVariables; // rhoBar.
        modified[2] = modif::staticVariables; // j.
        modified[3] = modif::nothing;         // Mass. Should be: staticVariables.
        modified[4] = modif::staticVariables; // Volume fraction.
        modified[5] = modif::staticVariables; // Flag-status.
        modified[6] = modif::nothing;         // Interface-lists.
    }
private:
    T rhoEmpty;
};

template<typename T, template<typename U> class Descriptor>
class FreeSurfaceWeightedForce3D : public BoxProcessingFunctional3D {
public:
    FreeSurfaceWeightedForce3D(Array<T,3> force_)
        : force(force_)
    { }
    virtual void processGenericBlocks(Box3D domain, std::vector<AtomicBlock3D*> atomicBlocks);
    virtual FreeSurfaceWeightedForce3D<T,Descriptor>* clone() const {
        return new FreeSurfaceWeightedForce3D<T,Descriptor>(*this);
    }
    virtual void getTypeOfModification (std::vector<modif::ModifT>& modified) const {
        // Default-assign potential other parameters present in a multi-fluid system.
        std::fill(modified.begin(), modified.end(), modif::nothing);
        modified[0] = modif::nothing;         // Fluid. Should be: staticVariables.
        modified[1] = modif::nothing;         // rhoBar.
        modified[2] = modif::nothing;         // j.
        modified[3] = modif::nothing;         // Mass. Should be: staticVariables.
        modified[4] = modif::nothing;         // Volume fraction.
        modified[5] = modif::nothing;         // Flag-status.
        modified[6] = modif::nothing;         // Interface-lists.
    }
private:
    Array<T,3> force;
};


/// Based on the current flag status, decide, upon the value of mass fraction, which nodes shall
///   switch state.
/** Input:
  *   - Volume fraction: needed in bulk+2
  *   - Flag-status:   needed in bulk+2
  * Output:
  *   - interface-to-fluid list: defined in bulk+2
  *   - interface-to-empty list: defined in bulk+1
  *   - empty-to-interface list: defined in bulk+1
  **/
template<typename T,template<typename U> class Descriptor>
class ComputeInterfaceLists3D : public BoxProcessingFunctional3D
{
public:
    virtual void processGenericBlocks(Box3D domain, std::vector<AtomicBlock3D*> atomicBlocks);
    virtual ComputeInterfaceLists3D<T,Descriptor>* clone() const {
        return new ComputeInterfaceLists3D<T,Descriptor>(*this);
    }
    virtual void getTypeOfModification (std::vector<modif::ModifT>& modified) const {
        // Default-assign potential other parameters present in a multi-fluid system.
        std::fill(modified.begin(), modified.end(), modif::nothing);
        modified[0] = modif::nothing;         // Fluid (not used in this processor).
        modified[1] = modif::nothing;         // rhoBar.
        modified[2] = modif::nothing;         // j.
        modified[3] = modif::nothing;         // Mass (not used in this processor).
        modified[4] = modif::nothing;         // Volume fraction, read-only.
        modified[5] = modif::nothing;         // Flag-status, read-only.
        modified[6] = modif::staticVariables; // Interface-lists; all lists are created here.
    }
private:
    static T kappa; //< Safety threshold for state-change, to prevent back-and-forth oscillations.
};

/** Input:
  *   - interface-to-fluid list: needed in bulk+1
  *   - interface-to-empty list: needed in bulk+1
  *   - density: needed in bulk+1
  *   - mass:    needed in bulk+1
  *   - flag:    needed in bulk+1
  * Output:
  *   - flag, dynamics, mass, volumeFraction, density, force, momentum
  *   - mass-excess-list: defined in bulk+1
  **/
template<typename T,template<typename U> class Descriptor>
class IniInterfaceToAnyNodes3D : public BoxProcessingFunctional3D {
public:
    IniInterfaceToAnyNodes3D(T rhoEmpty_);
    virtual void processGenericBlocks(Box3D domain, std::vector<AtomicBlock3D*> atomicBlocks);

    virtual IniInterfaceToAnyNodes3D<T,Descriptor>* clone() const {
        return new IniInterfaceToAnyNodes3D<T,Descriptor>(*this);
    }
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const {
        // Default-assign potential other parameters present in a multi-fluid system.
        std::fill(modified.begin(), modified.end(), modif::nothing);
        modified[0] = modif::nothing;           // Fluid. Gets assigned new dynamics. Should be: dataStructure
        modified[1] = modif::staticVariables;   // rhoBar.
        modified[2] = modif::nothing;           // j. Should be: staticVariables.
        modified[3] = modif::staticVariables;   // Mass. Is redistributed and initialized from neighborying density.
        modified[4] = modif::nothing;           // Volume fraction. Is default-initialized. Should be: staticVariables.
        modified[5] = modif::staticVariables;   // Flag-status. Is adapted according to cell-change lists.
        modified[6] = modif::nothing;           // Interface-lists. Read-only.
    }
private:
    T rhoEmpty; 
};


/// Based on the previously computed empty->interface list, initialize flow variables for
///   new interface cells.
/** Input:
  *   - Populations: needed in bulk+0
  *   - Momentum:    needed in bulk+1
  *   - Density:     needed in bulk+1
  *   - Flag-status: needed in bulk+0
  * Output:
  *   - flag-status:   initialized to "interface" on corresponding cells.
  *   - lattice:       initialized from neighbor averages on new interface cells.
  *   - mass:          initialized to zero on new interface cells.
  *   - mass-fraction: initialized to zero on new interface cells.
  *   - momentum
  **/
template<typename T,template<typename U> class Descriptor>
class IniEmptyToInterfaceNodes3D: public BoxProcessingFunctional3D {
public:
    IniEmptyToInterfaceNodes3D(Dynamics<T,Descriptor>* dynamicsTemplate_, Array<T,Descriptor<T>::d> force_)
        : dynamicsTemplate(dynamicsTemplate_), force(force_)
    { }
    IniEmptyToInterfaceNodes3D(IniEmptyToInterfaceNodes3D<T,Descriptor> const& rhs)
        : dynamicsTemplate(rhs.dynamicsTemplate->clone()),
          force(rhs.force)
    { }
    IniEmptyToInterfaceNodes3D<T,Descriptor>* operator=(IniEmptyToInterfaceNodes3D<T,Descriptor> const& rhs)
    { 
        IniEmptyToInterfaceNodes3D<T,Descriptor>(rhs).swap(*this);
        return *this;
    }
    void swap(IniEmptyToInterfaceNodes3D<T,Descriptor>& rhs) {
        std::swap(dynamicsTemplate, rhs.dynamicsTemplate);
        std::swap(force, rhs.force);
    }
    ~IniEmptyToInterfaceNodes3D() {
        delete dynamicsTemplate;
    }
    virtual void processGenericBlocks(Box3D domain, std::vector<AtomicBlock3D*> atomicBlocks);
    virtual IniEmptyToInterfaceNodes3D<T,Descriptor>* clone() const {
        return new IniEmptyToInterfaceNodes3D<T,Descriptor>(*this);
    }
    virtual void getTypeOfModification (std::vector<modif::ModifT>& modified) const {
        // Default-assign potential other parameters present in a multi-fluid system.
        std::fill(modified.begin(), modified.end(), modif::nothing);
        modified[0] = modif::nothing;         // Fluid. Should be: dataStructure
        modified[1] = modif::staticVariables; // rhoBar.
        modified[2] = modif::nothing;         // j. Should be: staticVariables.
        modified[3] = modif::staticVariables; // Mass.
        modified[4] = modif::nothing;         // Volume fraction, read-only. Should be: staticVariables
        modified[5] = modif::staticVariables; // Flag-status, read-only.
        modified[6] = modif::nothing;         // Interface-lists. Read access to gasCellToInitializeData.
    }
private:
    Dynamics<T,Descriptor>* dynamicsTemplate;
    Array<T,Descriptor<T>::d> force; // Body force, for initialization of the new interface cell.
};

/// Isolated cells cannot be part of the interface. This data processor spots and
/// removes them.
/** Input:
  *   - Flag-status: needed in bulk+2
  *   - mass:        needed in bulk+1
  *   - density:     needed in bulk+1
  * Output:
  *   - interfaceToFluidNodes:   initialized in bulk+1
  *   - interfaceToEmptyNodes:   initialized in bulk+1
  *   - massExcess list:         initialized in bulk+1
  *   - mass, density, mass-fraction, dynamics, force, momentum, flag: in bulk+1
  **/
template<typename T,template<typename U> class Descriptor>
class RemoveFalseInterfaceCells3D : public BoxProcessingFunctional3D {
public:
    RemoveFalseInterfaceCells3D(T rhoNSfluid_, T rhoEmpty_)
        : rhoNSfluid(rhoNSfluid_),
          rhoEmpty(rhoEmpty_)
    { }
    virtual void processGenericBlocks(Box3D domain, std::vector<AtomicBlock3D*> atomicBlocks);
    virtual RemoveFalseInterfaceCells3D<T,Descriptor>* clone() const {
        return new RemoveFalseInterfaceCells3D<T,Descriptor>(*this);
    }
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const {
        // Default-assign potential other parameters present in a multi-fluid system.
        std::fill(modified.begin(), modified.end(), modif::nothing);
        modified[0] = modif::nothing;          // Fluid: Gets NoDynamics when node changes to empty. Should be: dataStructure.
        modified[1] = modif::staticVariables;  // rhoBar.
        modified[2] = modif::nothing;          // j. Should be: staticVariables.
        modified[3] = modif::staticVariables;  // Mass
        modified[4] = modif::nothing;          // Volume fraction. Should be: staticVariables.
        modified[5] = modif::staticVariables;  // Flag-status
        modified[6] = modif::nothing;          // Interface lists
    }
private:
    T rhoNSfluid, rhoEmpty;  
};

/// Enforce exact mass balance when interface cells become fluid or empty.
/** Input:
  *   - mass-excess list: needed in bulk+1
  *   - Flag-status: needed in bulk+2
  *   - mass:        needed in bulk+2
  *   - density:     needed in bulk+2
  * Output:
  *   - mass, mass-fraction
  **/
template<typename T,template<typename U> class Descriptor>
class EqualMassExcessReDistribution3D : public BoxProcessingFunctional3D {
public:
    virtual void processGenericBlocks(Box3D domain, std::vector<AtomicBlock3D*> atomicBlocks);
    virtual EqualMassExcessReDistribution3D<T,Descriptor>* clone() const {
        return new EqualMassExcessReDistribution3D(*this);
    }
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const {
        // Default-assign potential other parameters present in a multi-fluid system.
        std::fill(modified.begin(), modified.end(), modif::nothing);
        modified[0] = modif::dataStructure;    // Fluid. Should be: nothing.
        modified[1] = modif::nothing;          // rhoBar.
        modified[2] = modif::staticVariables;  // j. Should be: nothing.
        modified[3] = modif::staticVariables;  // Mass
        modified[4] = modif::staticVariables;  // Volume fraction
        modified[5] = modif::nothing;          // Flag-status
        modified[6] = modif::nothing;          // Interface lists
    }
};

template<typename T,template<typename U> class Descriptor>
class InitializeInterfaceLists3D : public BoxProcessingFunctional3D {
    // Argument: (1) interface-lists
    virtual void processGenericBlocks(Box3D domain, std::vector<AtomicBlock3D*> atomicBlocks)
    {
        PLB_ASSERT(atomicBlocks.size()==1);
        
        AtomicContainerBlock3D* containerInterfaceLists = dynamic_cast<AtomicContainerBlock3D*>(atomicBlocks[0]);
        PLB_ASSERT(containerInterfaceLists);
        InterfaceLists<T,Descriptor>* interfaceLists = new InterfaceLists<T,Descriptor>;
        containerInterfaceLists->setData(interfaceLists);

    }
    virtual InitializeInterfaceLists3D<T,Descriptor>* clone() const {
        return new InitializeInterfaceLists3D<T,Descriptor>(*this);
    }
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const {
        // Default-assign potential other parameters present in a multi-fluid system.
        std::fill(modified.begin(), modified.end(), modif::nothing);
        modified[0] = modif::staticVariables;
    }
};

/// Wrapper for execution of InitializeInterfaceLists3D.
template<typename T,template<typename U> class Descriptor>
void initializeInterfaceLists3D(MultiContainerBlock3D& interfaceListBlock) {
    std::vector<MultiBlock3D*> arg;
    arg.push_back(&interfaceListBlock);
    applyProcessingFunctional (
            new InitializeInterfaceLists3D<T,Descriptor>,
            interfaceListBlock.getBoundingBox(), arg );
}


template<typename T, template<typename U> class Descriptor>
struct FreeSurfaceFields3D {
    static const int envelopeWidth = 4;
    static const int smallEnvelopeWidth = 1;

    /// This constructor is the most general one, and is to be used in most cases.
    /** Provide a blockManagement object to direct the sparse block distribution
     *  and parallel thread attribution of the blocks.
     **/
    FreeSurfaceFields3D( MultiBlockManagement3D const& blockManagement,
                         Dynamics<T,Descriptor>* dynamics_,
                         T rhoFluid_, T rhoEmpty_, Array<T,3> force_,
                         BoxProcessingFunctional3D_L<T,Descriptor>* postCollision=0 )
        : dynamics(dynamics_),
          rhoFluid(rhoFluid_), rhoEmpty(rhoEmpty_), force(force_),
          lattice (
                MultiBlockManagement3D (
                    blockManagement.getSparseBlockStructure(),
                    blockManagement.getThreadAttribution().clone(),
                    smallEnvelopeWidth,
                    blockManagement.getRefinementLevel() ),
                defaultMultiBlockPolicy3D().getBlockCommunicator(),
                defaultMultiBlockPolicy3D().getCombinedStatistics(),
                defaultMultiBlockPolicy3D().getMultiCellAccess<T,Descriptor>(), dynamics->clone() ),
          helperLists(lattice),
          mass(lattice),
          flag (
                MultiBlockManagement3D (
                    blockManagement.getSparseBlockStructure(),
                    blockManagement.getThreadAttribution().clone(),
                    envelopeWidth,
                    blockManagement.getRefinementLevel() ),
                defaultMultiBlockPolicy3D().getBlockCommunicator(),
                defaultMultiBlockPolicy3D().getCombinedStatistics(),
                defaultMultiBlockPolicy3D().getMultiScalarAccess<int>() ),
          volumeFraction((MultiBlock3D&)flag),
          rhoBar(lattice), j(lattice),
          freeSurfaceArgs(aggregateFreeSurfaceParams (
                  lattice, rhoBar, j, mass, volumeFraction, flag, helperLists ) )
    {
        initializeAll(postCollision);
    }
    FreeSurfaceFields3D( SparseBlockStructure3D const& blockStructure,
                         Dynamics<T,Descriptor>* dynamics_,
                         T rhoFluid_, T rhoEmpty_, Array<T,3> force_,
                         BoxProcessingFunctional3D_L<T,Descriptor>* postCollision=0 )
        : dynamics(dynamics_),
          rhoFluid(rhoFluid_), rhoEmpty(rhoEmpty_), force(force_),
          lattice (
                MultiBlockManagement3D (
                                blockStructure, defaultMultiBlockPolicy3D().getThreadAttribution(), smallEnvelopeWidth ),
                defaultMultiBlockPolicy3D().getBlockCommunicator(),
                defaultMultiBlockPolicy3D().getCombinedStatistics(),
                defaultMultiBlockPolicy3D().getMultiCellAccess<T,Descriptor>(), dynamics->clone() ),
          helperLists(lattice),
          mass(lattice),
          flag (
                MultiBlockManagement3D (
                            blockStructure,
                            defaultMultiBlockPolicy3D().getThreadAttribution(),
                            envelopeWidth ),
                defaultMultiBlockPolicy3D().getBlockCommunicator(),
                defaultMultiBlockPolicy3D().getCombinedStatistics(),
                defaultMultiBlockPolicy3D().getMultiScalarAccess<int>() ),
          volumeFraction((MultiBlock3D&)flag),
          rhoBar(lattice), j(lattice),
          freeSurfaceArgs(aggregateFreeSurfaceParams (
                  lattice, rhoBar, j, mass, volumeFraction, flag, helperLists ) )
    {
        initializeAll(postCollision);
    }
private:
    FreeSurfaceFields3D(FreeSurfaceFields3D<T,Descriptor> const& rhs) { }
    FreeSurfaceFields3D<T,Descriptor>& operator=(FreeSurfaceFields3D<T,Descriptor>& rhs) { return *this; }
    void initializeAll(BoxProcessingFunctional3D_L<T,Descriptor>* postCollision )
    {
        initializeInterfaceLists3D<T,Descriptor>(helperLists);
        lattice.periodicity().toggleAll(true);
        mass.periodicity().toggleAll(true);
        flag.periodicity().toggleAll(true);
        volumeFraction.periodicity().toggleAll(true);
        rhoBar.periodicity().toggleAll(true);
        j.periodicity().toggleAll(true);
        setToConstant(flag, flag.getBoundingBox(), (int)freeSurfaceFlag::empty);
        rhoBarJparam.push_back(&lattice);
        rhoBarJparam.push_back(&rhoBar);
        rhoBarJparam.push_back(&j);

        freeSurfaceDataProcessors(rhoFluid, rhoEmpty, force, *dynamics, postCollision);
    }
public:
    ~FreeSurfaceFields3D() {
        delete dynamics;
    }

    void defaultInitialize() {
        applyProcessingFunctional (
           new DefaultInitializeFreeSurface3D<T,Descriptor>(dynamics->clone(), force),
                   lattice.getBoundingBox(), freeSurfaceArgs );
    }

    void partiallyDefaultInitialize() {
        applyProcessingFunctional (
           new PartiallyDefaultInitializeFreeSurface3D<T,Descriptor>(dynamics->clone(), force),
                   lattice.getBoundingBox(), freeSurfaceArgs );
    }

    void freeSurfaceDataProcessors(T rhoFluid, T rhoEmpty, Array<T,3> force, Dynamics<T,Descriptor>& dynamics,
                                   BoxProcessingFunctional3D_L<T,Descriptor>* postCollision)
    {
        plint processorLevelZero  = 0;
        plint processorLevelOne   = 1;
        plint processorLevelTwo   = 2;
        plint processorLevelThree = 3;
        plint processorLevelFour  = 4;

        /***** Level 0 ******/
        integrateProcessingFunctional (
                new ExternalRhoJcollideAndStream3D<T,Descriptor>,
                lattice.getBoundingBox(), rhoBarJparam, processorLevelZero );

        if (postCollision) {
            integrateProcessingFunctional (
                    postCollision, lattice.getBoundingBox(), lattice, processorLevelZero );
        }

        /***** Level 1 ******/
        integrateProcessingFunctional (
            new MassChange3D<T,Descriptor>, lattice.getBoundingBox(),
            freeSurfaceArgs, processorLevelOne );
       
        integrateProcessingFunctional (
            new FreeSurfaceCompletion3D<T,Descriptor>(rhoEmpty),
            lattice.getBoundingBox(), freeSurfaceArgs, processorLevelOne );
                                    
        integrateProcessingFunctional (
            new FreeSurfaceMacroscopic3D<T,Descriptor>(rhoEmpty), 
            lattice.getBoundingBox(), freeSurfaceArgs, processorLevelOne );

        /***** Level 2 ******/
        integrateProcessingFunctional (
            new ComputeInterfaceLists3D<T,Descriptor>(),
            lattice.getBoundingBox(), freeSurfaceArgs, processorLevelTwo );

        integrateProcessingFunctional (
            new IniInterfaceToAnyNodes3D<T,Descriptor>(rhoEmpty),
            lattice.getBoundingBox(), freeSurfaceArgs, processorLevelTwo );
            
        integrateProcessingFunctional (
            new IniEmptyToInterfaceNodes3D<T,Descriptor>(dynamics.clone(), force),
                                    lattice.getBoundingBox(),
                                    freeSurfaceArgs, processorLevelTwo ); 

        /***** Level 3 ******/
        integrateProcessingFunctional (
            new RemoveFalseInterfaceCells3D<T,Descriptor>(rhoFluid, rhoEmpty),
            lattice.getBoundingBox(), freeSurfaceArgs, processorLevelThree);

        /***** Level 4 ******/
        integrateProcessingFunctional (
            new EqualMassExcessReDistribution3D<T,Descriptor>(),
            lattice.getBoundingBox(), freeSurfaceArgs, processorLevelFour );
    }

    Dynamics<T,Descriptor>* dynamics;
    T rhoFluid, rhoEmpty;
    Array<T,3> force;
    MultiBlockLattice3D<T, Descriptor> lattice;
    MultiContainerBlock3D helperLists;
    MultiScalarField3D<T> mass;
    MultiScalarField3D<int> flag;
    MultiScalarField3D<T> volumeFraction;
    MultiScalarField3D<T> rhoBar;
    MultiTensorField3D<T,3> j;
    std::vector<MultiBlock3D*> rhoBarJparam;
    std::vector<MultiBlock3D*> freeSurfaceArgs;
};

}  // namespace plb

#endif  // FREE_SURFACE_MODEL_3D_H

