/* This file is part of the Palabos library.
 *
 * Copyright (C) 2011 FlowKit Sarl
 * Avenue de Chailly 23
 * 1012 Lausanne, Switzerland
 * E-mail contact: contact@flowkit.com
 *
 * The most recent release of Palabos can be downloaded at 
 * <http://www.palabos.org/>
 *
 * The library Palabos is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * The library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/* Main author: Andrea */

#ifndef FREE_SURFACE_MODEL_3D_HH
#define FREE_SURFACE_MODEL_3D_HH

#include "core/globalDefs.h"
#include "core/block3D.h"
#include "latticeBoltzmann/geometricOperationTemplates.h"
#include "atomicBlock/dataProcessor3D.h"
#include "atomicBlock/blockLattice3D.h"
#include "atomicBlock/atomicContainerBlock3D.h"
#include "multiPhysics/freeSurfaceModel3D.h"
#include "multiPhysics/freeSurfaceTemplates.h"

namespace plb {

/* *************** Class MassChange3D ******************************************* */

template< typename T,template<typename U> class Descriptor>
void MassChange3D<T,Descriptor>::processGenericBlocks (
        Box3D domain,std::vector<AtomicBlock3D*> atomicBlocks )
{
    global::timer("FS_MassChange").start();
    typedef Descriptor<T> D;
    using namespace freeSurfaceFlag;
    FreeSurfaceProcessorParam3D<T,Descriptor> param(atomicBlocks);

    // This loop updates the mass, summarizing  Eq. 6/7, and Eq.8, in
    // the N. Thuerey e.a. technical report "Interactive Free Surface Fluids
    // with the Lattice Boltzmann Method".
    for (plint iX=domain.x0; iX<=domain.x1; ++iX) {
        for (plint iY=domain.y0; iY<=domain.y1; ++iY) {
            for (plint iZ=domain.z0; iZ<=domain.z1; ++iZ) {
                Cell<T,Descriptor>& cell = param.cell(iX,iY,iZ);
                int flag = param.flag(iX,iY,iZ);
                if(flag == fluid) {
                    freeSurfaceTemplates<T,Descriptor>::massExchangeFluidCell(param, iX,iY,iZ);
                }
                else if(flag==interface) {
                    for(plint iPop=0; iPop < D::q; ++iPop) {
                        plint nextX = iX + D::c[iPop][0];
                        plint nextY = iY + D::c[iPop][1];
                        plint nextZ = iZ + D::c[iPop][2];
                        int nextFlag = param.flag(nextX,nextY,nextZ);
                        plint opp = indexTemplates::opposite<D>(iPop);
                        // Calculate mass at time t+1 on interface cell --> eq 7 Thurey's paper.
                        if (nextFlag==fluid) {
                            param.mass(iX,iY,iZ) +=   
                                (cell[opp] - param.cell(nextX,nextY,nextZ)[iPop]);
                        }
                        else if (nextFlag==interface) {
                            param.mass(iX,iY,iZ) +=   
                                (cell[opp] - param.cell(nextX,nextY,nextZ)[iPop]) *
                                    0.5*(param.volumeFraction(nextX,nextY,nextZ) + param.volumeFraction(iX,iY,iZ));
                        } 
                    }
                }
            }
        }
    }
    global::timer("FS_MassChange").stop();
}


/* *************** Class FreeSurfaceCompletion3D ******************************************* */

template< typename T,template<typename U> class Descriptor, class Param>
void freeSurfaceCompletionImplementation3D(Param& param, Box3D domain)
{
    typedef Descriptor<T> D;
    using namespace freeSurfaceFlag;
    typedef typename InterfaceLists<T,Descriptor>::Node Node;

    // In this data processor, populations are both written locally and read non-locally.
    // To guarantee data consistency, a first loop makes only read accesses and stores
    // the necessary information into the list neighborOppositePop. A second loop reads
    // from this list and assigns values to populations.
    std::map<Node, Array<T,D::q> > neighborOppositePop;
    for (plint iX=domain.x0; iX<=domain.x1; ++iX) {
        for (plint iY=domain.y0; iY<=domain.y1; ++iY) {
            for (plint iZ=domain.z0; iZ<=domain.z1; ++iZ) {
                if(param.flag(iX,iY,iZ) == interface) {
                    // Here we are on an interface node. The entire set of fi's is reconstructed.
                    // The normal is recomputed as in eq. 10 of Thurey's paper.
                    Array<T,D::d> normalToInterface;                    
                    normalToInterface.resetToZero();
                    // Compute surface normal "the lattice Boltzmann way".
                    for (plint iPop=0; iPop < D::q; ++iPop ) {
                        plint nextX = iX + D::c[iPop][0];
                        plint nextY = iY + D::c[iPop][1];
                        plint nextZ = iZ + D::c[iPop][2];
                        if ( param.flag(nextX,nextY,nextZ) != empty &&  
                             param.flag(nextX,nextY,nextZ) != wall )
                        {
                            normalToInterface[0] -= D::t[iPop]*D::c[iPop][0]*param.volumeFraction(nextX,nextY,nextZ);
                            normalToInterface[1] -= D::t[iPop]*D::c[iPop][1]*param.volumeFraction(nextX,nextY,nextZ);
                            normalToInterface[2] -= D::t[iPop]*D::c[iPop][2]*param.volumeFraction(nextX,nextY,nextZ);
                        }
                    }
                    // None need to normalize the normal vector here, because it is only used to obtain
                    // information on the direction.
                    normalToInterface *= D::invCs2;                     
                    
                    bool needsModification = false;
                    Array<T,D::q> savedPop;
                    savedPop[0] = -2.;
                    for(plint iPop=1; iPop < D::q; ++iPop )
                    {
                        // This is one of the tricky points of the code
                        // we have to decide if the f_is from the neighborhood
                        // have to be re-update by using the Thurey's rule, which
                        // states that f_i's coming from nearest neighs. that are empy cells,
                        // have to be re-updated.
                        // I like the eq.   f^{in}_i(x,t+dt) = f^{out}_i(x-e_i,t);
                        // This eq. makes me think that the neigh. that I have to check 
                        // (to control is status e.g. empty or fluid ?) has to be pos-c_i
                        plint prevX = iX-D::c[iPop][0];
                        plint prevY = iY-D::c[iPop][1];
                        plint prevZ = iZ-D::c[iPop][2];
                        
                        plint opp = indexTemplates::opposite<D>(iPop);
                        T scalarProduct = D::c[opp][0]*normalToInterface[0] +
                                          D::c[opp][1]*normalToInterface[1] +
                                          D::c[opp][2]*normalToInterface[2];
                        
                        // Should I also change particle distribution function coming from 
                        // bounceBack nodes? Well ideally no ... but there is for sure some
                        // cell configuration where these f_is are not well defined because 
                        // they are probably coming from empty cells

                        // If the f_i[iPop] would be streamed from an empty cell, or whenever the scalar product is positive.
                        if ( scalarProduct > 0 || param.flag(prevX,prevY,prevZ) == empty ||
                             param.flag(prevX,prevY,prevZ) == wall )
                        {
                            savedPop[iPop] = param.cell(prevX,prevY,prevZ)[opp];
                            needsModification = true;
                        }
                        else {
                            savedPop[iPop] = (T)-2.;
                        }
                    }
                    if (needsModification) {
                        neighborOppositePop.insert(std::pair<Node,Array<T,D::q> >(Node(iX,iY,iZ), savedPop));
                    }
                }
            }
        }
    }

    typename std::map<Node, Array<T,D::q> >::const_iterator nodes = neighborOppositePop.begin();
    for (; nodes != neighborOppositePop.end(); ++nodes) {
        Node node = nodes->first;
        plint iX = node[0];
        plint iY = node[1];
        plint iZ = node[2];
        Array<T,D::q> neighborOppPop = nodes->second;
        for (plint iPop=1; iPop < D::q; ++iPop ) {
            if (neighborOppPop[iPop]>(T)-1.) {
                // Velocity is simply taken from the previous time step.
                Array<T,D::d> j = param.getMomentum(iX,iY,iZ);
                T jSqr = VectorTemplate<T,Descriptor>::normSqr(j);
                // Remember: the value of pressure on an interface node has been set in
                // F, and is equal to the ambient pressure for a
                // single free-surface fluid, or in the case of a binary pressure, an
                // averaged value.
                T rhoBar = Descriptor<T>::rhoBar(param.getDensity(iX,iY,iZ));
                T feq_i = param.cell(iX,iY,iZ).computeEquilibrium(iPop, rhoBar, j, jSqr);
                plint opp = indexTemplates::opposite<D>(iPop);
                T feq_opp_i = param.cell(iX,iY,iZ).computeEquilibrium(opp, rhoBar, j, jSqr);
                param.cell(iX,iY,iZ)[iPop] = feq_i + feq_opp_i - neighborOppPop[iPop];
            }
        }
    }
}

template< typename T,template<typename U> class Descriptor>
void FreeSurfaceCompletion3D<T,Descriptor>
        ::processGenericBlocks(Box3D domain,std::vector<AtomicBlock3D*> atomicBlocks)
{
    global::timer("FS_FreeSurfaceCompletion").start();
    FreeSurfaceProcessorParam3D<T,Descriptor> param(atomicBlocks);
    freeSurfaceCompletionImplementation3D
        <T,Descriptor,FreeSurfaceProcessorParam3D<T,Descriptor> > (param, domain);
    global::timer("FS_FreeSurfaceCompletion").stop();
}


/* *************** Class FreeSurfaceMacroscopic3D ******************************************* */

template< typename T,template<typename U> class Descriptor>
void FreeSurfaceMacroscopic3D<T,Descriptor>
        ::processGenericBlocks(Box3D domain,std::vector<AtomicBlock3D*> atomicBlocks)
{
    global::timer("FS_FreeSurfaceMacroscopic").start();
    typedef Descriptor<T> D;
    typedef typename InterfaceLists<T,Descriptor>::Node Node;
    using namespace freeSurfaceFlag;
    FreeSurfaceProcessorParam3D<T,Descriptor> param(atomicBlocks);


    // Save macroscopic fields in external scalars and update the mass-fraction.
    for (plint iX=domain.x0; iX<=domain.x1; ++iX) {
        for (plint iY=domain.y0; iY<=domain.y1; ++iY) {
            for (plint iZ=domain.z0; iZ<=domain.z1; ++iZ) {         
                if ( param.flag(iX,iY,iZ)==interface || param.flag(iX,iY,iZ)==fluid ) {
                    T rhoBar; 
                    Array<T,D::d> j;
                    momentTemplates<T,Descriptor>::get_rhoBar_j(param.cell(iX,iY,iZ), rhoBar, j);
                    T density = Descriptor<T>::fullRho(rhoBar);
                    param.setDensity(iX,iY,iZ, density);

                    if (param.flag(iX,iY,iZ)==interface) {
                        param.volumeFraction(iX,iY,iZ) = param.mass(iX,iY,iZ)/density;
                        // On interface cells, adjust the pressure to the ambient pressure.
                        param.setDensity(iX,iY,iZ, rhoEmpty);
                        j *= rhoEmpty/density;
                    }
                    else if (param.flag(iX,iY,iZ)==fluid) {
                        param.volumeFraction(iX,iY,iZ) = T(1);
                    }
                    
                    Array<T,D::d> force = param.getForce(iX,iY,iZ);
                    T tau = T(1)/param.cell(iX,iY,iZ).getDynamics().getOmega();
                    for (plint iD=0; iD<D::d; ++iD) {
                        j[iD] += density*force[iD]*tau; //*param.volumeFraction(iX,iY,iZ);
                    }
                    param.setMomentum(iX,iY,iZ, j);
                }
            }
        }
    }   
    global::timer("FS_FreeSurfaceMacroscopic").stop();
}


/* *************** Class ComputeInterfaceLists3D ******************************************* */
    
template< typename T, template<typename> class Descriptor>
T ComputeInterfaceLists3D<T,Descriptor>::kappa = 1.e-3;

template< typename T,template<typename U> class Descriptor>
void ComputeInterfaceLists3D<T,Descriptor>
        ::processGenericBlocks(Box3D domain,std::vector<AtomicBlock3D*> atomicBlocks)
{
    global::timer("FS_ComputeInterfaceLists").start();
    typedef Descriptor<T> D;
    typedef typename InterfaceLists<T,Descriptor>::Node Node;
    FreeSurfaceProcessorParam3D<T,Descriptor> param(atomicBlocks);
    using namespace freeSurfaceFlag;

    param.massExcess().clear();
    param.interfaceToFluid().clear();
    param.interfaceToEmpty().clear();
    param.emptyToInterface().clear();

    
    // interfaceToFluid needs to be computed in bulk+2.
    for (plint iX=domain.x0-2; iX<=domain.x1+2; ++iX) {
        for (plint iY=domain.y0-2; iY<=domain.y1+2; ++iY) {
            for (plint iZ=domain.z0-2; iZ<=domain.z1+2; ++iZ) { 
                Node node(iX,iY,iZ);
                // Eq. 11 in Thuerey's technical report.
                if (param.flag(iX,iY,iZ) == interface) { // Interface cell.
                    if (param.volumeFraction(iX,iY,iZ) > T(1)+kappa ) { // Interface cell is filled.
                        // Elements are added even if they belong to the envelope, because they may be
                        //   needed further down in the same data processor.
                        param.interfaceToFluid().insert(node);
                    }
                    else if (param.volumeFraction(iX,iY,iZ) < kappa) { // Interface cell is empty.
                        // Elements are added even if they belong to the envelope, because they may be
                        //   needed further down in the same data processor.
                        param.interfaceToEmpty().insert(node);
                    }
                }
            }
        }
    }
    
    // Where interface cells have become fluid, neighboring cells must be prevented from
    //   being empty, because otherwise there's no interface cell between empty and fluid.
    typename std::set<Node>::iterator iEle = param.interfaceToFluid().begin();
    for (; iEle != param.interfaceToFluid().end(); ++iEle) {
        // The node here may belong to the 1st envelope.
        Node node = *iEle;
        plint iX=node[0];
        plint iY=node[1];
        plint iZ=node[2];
        
        for(plint iPop=1; iPop < D::q; ++iPop) {
            plint nextX = iX+D::c[iPop][0];
            plint nextY = iY+D::c[iPop][1];
            plint nextZ = iZ+D::c[iPop][2];     
            Node nextNode(nextX,nextY,nextZ);
            
            // If one of my neighbors switches interface->fluid, then I shall be prevented
            //     from switching interface->empty at the same time step.
            if (contained(nextX,nextY,nextZ,domain.enlarge(1)) && param.flag(nextX,nextY,nextZ) == interface ) {  
                param.interfaceToEmpty().erase(nextNode);
            }
            // If one of my neighbors switches interface->fluid and I am empty I shall become
            //   interface.
            else if (contained(nextX,nextY,nextZ,domain) && param.flag(nextX,nextY,nextZ) == empty ) {
                param.emptyToInterface().insert(nextNode);
            }
        }
    }
    global::timer("FS_ComputeInterfaceLists").stop();
}


/* *************** Class IniInterfaceToAnyNodes3D ******************************************* */

template< typename T,template<typename U> class Descriptor>
IniInterfaceToAnyNodes3D<T,Descriptor>::IniInterfaceToAnyNodes3D(T rhoEmpty_)
    : rhoEmpty(rhoEmpty_)
{ }

template< typename T,template<typename U> class Descriptor>
void IniInterfaceToAnyNodes3D<T,Descriptor>
        ::processGenericBlocks(Box3D domain,std::vector<AtomicBlock3D*> atomicBlocks)
{
    global::timer("FS_IniInterfaceToAnyNodes").start();
    typedef Descriptor<T> D;
    typedef typename InterfaceLists<T,Descriptor>::Node Node;
    using namespace freeSurfaceFlag;

    FreeSurfaceProcessorParam3D<T,Descriptor> param(atomicBlocks);
    
    // 1. For interface->fluid nodes, update in the flag matrix,
    //   and compute and store mass excess from these cells.
    typename std::set<Node>::iterator iEle = param.interfaceToFluid().begin();
    for (; iEle != param.interfaceToFluid().end(); ++iEle) {
        Node node = *iEle;
        
        plint iX = node[0];
        plint iY = node[1];
        plint iZ = node[2];
 
        if (contained(iX,iY,iZ,domain.enlarge(1))) {
            T localMass = param.mass(iX,iY,iZ);
            param.mass(iX,iY,iZ) = param.getDensity(iX,iY,iZ);
            param.volumeFraction(iX,iY,iZ) = (T)1;
            
            // Make the interface->fluid update in the flag matrix.
            if (param.flag(iX,iY,iZ)==interface)
            {
                param.flag(iX,iY,iZ) = fluid;
            }
            else {
                pcout << "INTERNAL ERROR: Inside first part of IniInterfaceToAnyNodes3D, wanted to" << std::endl;
                pcout << "switch interface to fluid cell, but instead of being interface, the cell was "
                      << flagToString(param.flag(iX,iY,iZ)) << std::endl;
                PLB_ASSERT( false );
            }
                
            T massExcess = localMass - param.getDensity(iX,iY,iZ);
            param.massExcess().insert(std::pair<Node,T>(node,massExcess));
        }
    }
    
    // 2. For interface->empty nodes, update in the flag matrix,
    //   and compute and store mass excess from these cells.
    iEle = param.interfaceToEmpty().begin();
    for (; iEle != param.interfaceToEmpty().end(); ++iEle) 
    {
        Node node = *iEle;
        plint iX = node[0];
        plint iY = node[1];
        plint iZ = node[2];
        
        if (contained(iX,iY,iZ,domain.enlarge(1))) {
            param.flag(iX,iY,iZ) = empty;
            param.attributeDynamics(iX,iY,iZ, new NoDynamics<T,Descriptor>());

            T massExcess = param.mass(iX,iY,iZ);
            param.massExcess().insert(std::pair<Node,T>(node,massExcess));

            param.mass(iX,iY,iZ) = T();
            param.volumeFraction(iX,iY,iZ) = T();
            param.setDensity(iX,iY,iZ, rhoEmpty);
            param.setForce(iX,iY,iZ, Array<T,3>(T(),T(),T()));
            param.setMomentum(iX,iY,iZ, Array<T,3>(T(),T(),T()));
        }
        
        // Avoid the case where an empty cell has a fluid neighbor without
        // interface cell between them.
        for(plint iPop=1; iPop < D::q; ++iPop) {
            plint nextX = iX+D::c[iPop][0];
            plint nextY = iY+D::c[iPop][1];
            plint nextZ = iZ+D::c[iPop][2];

            // The concurrent read/write on param.flag is not an issue here, because the
            // result in any case is that all adjacent fluid cells have become interface.
            if (contained(nextX,nextY,nextZ,domain) && param.flag(nextX,nextY,nextZ) == fluid) {
                param.flag(nextX,nextY,nextZ) = interface;  
            }
        }
    }
    global::timer("FS_IniInterfaceToAnyNodes").stop();
}


/* *************** Class IniEmptyToInterfaceNodes3D ******************************************* */

template< typename T,template<typename U> class Descriptor>
void IniEmptyToInterfaceNodes3D<T,Descriptor>
        ::processGenericBlocks(Box3D domain,std::vector<AtomicBlock3D*> atomicBlocks)
{
    global::timer("FS_IniEmptyToInterfaceNodes").start();
    typedef Descriptor<T> D;
    typedef typename InterfaceLists<T,Descriptor>::Node Node;
    using namespace freeSurfaceFlag;
    FreeSurfaceProcessorParam3D<T,Descriptor> param(atomicBlocks);

    // In this data processor, density and momentum are potentially read and written
    //   from the same node, because nodes can switch state. The following two vectors
    //   store temporary variables to avoid read/write in undefined order.
    std::vector<T> newDensity(param.emptyToInterface().size());
    std::vector<Array<T,3> > newMomentum(param.emptyToInterface().size());
    std::fill(newDensity.begin(), newDensity.end(), T());
    std::fill(newMomentum.begin(), newMomentum.end(), Array<T,3>(T(),T(),T()));
    
    // Compute density and momentum for cells that will switch state empty->interface.
    //   It is sufficient to do this is bulk+0.
    //   This loop performs read-only access to the lattice.
    plint i=0;
    typename std::set<Node>::iterator iEle = param.emptyToInterface().begin();
    for (; iEle != param.emptyToInterface().end(); ++iEle, ++i ) 
    {
        Node node = *iEle;
        plint iX=node[0];
        plint iY=node[1];
        plint iZ=node[2];
        
        // If non-bulk elements are left in the list, disregard to avoid accessing undefined neighbors.
        if (contained(iX,iY,iZ,domain) ) {
            // For initialization of the new cell, compute average density
            //   and momentum on neighbors.
            T averageDensity = T(0);
            Array<T,3> averageMomentum(T(0),T(0),T(0));
            plint counter = 0; 
            for(plint iPop=1; iPop<D::q; ++iPop)
            {
                plint nextX = iX+D::c[iPop][0];
                plint nextY = iY+D::c[iPop][1];
                plint nextZ = iZ+D::c[iPop][2];
                
                // Warning: it is not accounted for the fact that neighbors can have excess mass. It
                //   might be good to account for this in the future.
                if (param.flag(nextX,nextY,nextZ) == interface || param.flag(nextX,nextY,nextZ) == fluid ) {
                    counter++;
                    averageDensity += param.getDensity(nextX,nextY,nextZ);
                    averageMomentum += param.getMomentum(nextX,nextY,nextZ);
                } 
            }
            T invCounter = T(1)/counter;
            averageDensity  *= invCounter;
            averageMomentum *= invCounter;
            newDensity[i] = averageDensity;
            newMomentum[i] = averageMomentum;
        }
    }
            
    // Elements that have switched state empty->interface are initialized at equilibrium.
    //   It is sufficient to initialize them in bulk+0.
    //   This loop performs write-only access on the lattice.
    i=0;
    iEle = param.emptyToInterface().begin();
    for (; iEle != param.emptyToInterface().end(); ++iEle, ++i ) 
    {
        Node node = *iEle;
        
        plint iX=node[0];
        plint iY=node[1];
        plint iZ=node[2];
        
        // If non-bulk elements are left in the list, disregard to avoid accessing undefined neighbors.
        if (contained(iX,iY,iZ,domain) ) 
        {
            T averageDensity = newDensity[i];
            Array<T,3> averageMomentum = newMomentum[i];

            param.attributeDynamics (
                    iX,iY,iZ, dynamicsTemplate->clone() );
            
            iniCellAtEquilibrium(param.cell(iX,iY,iZ), averageDensity, averageMomentum);
            param.setForce(iX,iY,iZ, force);
            // Change density, but leave mass and volumeFraction at 0, as they are later
            //   recomputed (Warning: this is probably correct, but there remains a small doubt).
            param.setMomentum(iX,iY,iZ, averageMomentum);
            param.setDensity(iX,iY,iZ, averageDensity);
            param.mass(iX,iY,iZ) = T();
            param.volumeFraction(iX,iY,iZ) = T();
            param.flag(iX,iY,iZ) = interface;
        }
    }
    global::timer("FS_IniEmptyToInterfaceNodes").stop();
}


/* *************** Class RemoveFalseInterfaceCells3D ******************************************* */

template< typename T,template<typename U> class Descriptor>
void RemoveFalseInterfaceCells3D<T,Descriptor>
        ::processGenericBlocks(Box3D domain,std::vector<AtomicBlock3D*> atomicBlocks)
{
    global::timer("FS_RemoveFalseInterfaceCells").start();
    typedef Descriptor<T> D;
    typedef typename InterfaceLists<T,Descriptor>::Node Node;
    using namespace freeSurfaceFlag;

    FreeSurfaceProcessorParam3D<T,Descriptor> param(atomicBlocks);

    /// In the following, the flag status of cells is read (non-locally) and
    /// modified (locally). To avoid conflict, two loops are made, the first
    /// of which reads only, and the second writes. The vectors "interfaceToFluidNodes"
    /// and "interfaceToEmptyNodes" store coordinates of nodes that will switch
    /// status.
    std::vector<Node> interfaceToFluidNodes, interfaceToEmptyNodes;
    for (plint iX=domain.x0-1; iX<=domain.x1+1; ++iX) {
        for (plint iY=domain.y0-1; iY<=domain.y1+1; ++iY) {
            for (plint iZ=domain.z0-1; iZ<=domain.z1+1; ++iZ) {    
                Node node(iX,iY,iZ);
                if (param.flag(iX,iY,iZ) == interface) {
                    bool noEmptyNeighbor = true;
                    bool noFluidNeighbor = true;
                    
                    for(plint iPop=1;iPop<D::q; iPop++) {
                        plint nextX = iX+D::c[iPop][0];
                        plint nextY = iY+D::c[iPop][1];
                        plint nextZ = iZ+D::c[iPop][2];
                
                        if (param.flag(nextX,nextY,nextZ) == empty) noEmptyNeighbor = false;
                        if (param.flag(nextX,nextY,nextZ) == fluid) noFluidNeighbor = false;
                    }
                    
                    if (noEmptyNeighbor) {
                        interfaceToFluidNodes.push_back(Node(iX,iY,iZ));
                        // Store the coordinates, so flag on this node
                        // can be changed in a loop outside the current one.
                        
                        T massExcess = param.mass(iX,iY,iZ) - param.getDensity(iX,iY,iZ);
                        param.massExcess().insert(std::pair<Node,T>(node,massExcess));

                        // I'm not sure about what follows actually.  I'm not changing the f_is
                        // (which is cool I would say) but I'm changing the density value (and also the mass)
                        // that are going to be used at the next time step for collision and mass change
                        // By doing that, I'm not sure that I have the control on the conservation of mass.
                        param.mass(iX,iY,iZ) = rhoNSfluid;
                        param.setDensity(iX,iY,iZ, rhoNSfluid);
                        param.volumeFraction(iX,iY,iZ) = T(1);
                    }
                    else if (noFluidNeighbor) {
                        interfaceToEmptyNodes.push_back(Node(iX,iY,iZ));
                        // Store the coordinates, so flag on this node
                        // can be changed in a loop outside the current one.

                        T massExcess = param.mass(iX,iY,iZ);
                        param.massExcess().insert(std::pair<Node,T>(node,massExcess));
                        
                        param.attributeDynamics(iX,iY,iZ,new NoDynamics<T,Descriptor>());
                        param.mass(iX,iY,iZ) = T();
                        param.setDensity(iX,iY,iZ, rhoEmpty);
                        param.volumeFraction(iX,iY,iZ) = T();
                        param.setForce(iX,iY,iZ, Array<T,3>(T(),T(),T()));
                        param.setMomentum(iX,iY,iZ, Array<T,3>(T(),T(),T()));
                    }
                }
            }
        }
    }

    for (pluint i=0; i<interfaceToFluidNodes.size(); ++i) {
        Node const& pos = interfaceToFluidNodes[i];
        param.flag(pos[0],pos[1],pos[2]) = fluid;
    }
    for (pluint i=0; i<interfaceToEmptyNodes.size(); ++i) {
        Node const& pos = interfaceToEmptyNodes[i];
        param.flag(pos[0],pos[1],pos[2]) = empty;
    }
    global::timer("FS_RemoveFalseInterfaceCells").stop();
}


/* *************** Class EqualMassExcessReDistribution3D ******************************************* */

template< typename T,template<typename U> class Descriptor>
void EqualMassExcessReDistribution3D<T,Descriptor>
        ::processGenericBlocks(Box3D domain,std::vector<AtomicBlock3D*> atomicBlocks)
{
    global::timer("FS_EqualMassExcessReDistribution").start();
    typedef Descriptor<T> D;
    typedef typename InterfaceLists<T,Descriptor>::Node Node;
    using namespace freeSurfaceFlag;
    FreeSurfaceProcessorParam3D<T,Descriptor> param(atomicBlocks);
        
    // First loop: compute new mass from mass-excess (non-local).
    plint i=0;
    typename std::map<Node,T>::iterator iEle = param.massExcess().begin();
    for (; iEle != param.massExcess().end(); ++iEle, ++i)
    {
        Array<plint,3> node = iEle->first;
        plint iX = node[0];
        plint iY = node[1];
        plint iZ = node[2];
        
        if( contained(iX,iY,iZ,domain.enlarge(1)) )  {
            plint numInterfaceNeighbors = 0;
            for (plint iPop=1; iPop<D::q; iPop++) {
                plint nextX = iX + D::c[iPop][0] ;
                plint nextY = iY + D::c[iPop][1] ;
                plint nextZ = iZ + D::c[iPop][2] ;
                if (param.flag(nextX,nextY,nextZ) == interface) numInterfaceNeighbors++;
            }
            
            T massToRedistribute = iEle->second/(T)numInterfaceNeighbors;
            for (plint iPop=1; iPop<D::q; iPop++) {
                plint nextX = iX+D::c[iPop][0];
                plint nextY = iY+D::c[iPop][1];
                plint nextZ = iZ+D::c[iPop][2];     
                if ( contained(nextX,nextY,nextZ,domain) &&
                     param.flag(nextX,nextY,nextZ) == interface )
                {
                    param.mass(nextX,nextY,nextZ) += massToRedistribute;
                }    
            }
        }
    }

    // Second loop: compute new mass-fraction from mass and density (non-local).
    i=0;
    iEle = param.massExcess().begin();
    for (; iEle != param.massExcess().end(); ++iEle, ++i) {
        Array<plint,3> node = iEle->first;
        plint iX = node[0];
        plint iY = node[1];
        plint iZ = node[2];
        
        if(contained(iX,iY,iZ,domain.enlarge(1)) )  {
            for (plint iPop=1; iPop<D::q; iPop++) {
                plint nextX = iX + D::c[iPop][0];
                plint nextY = iY + D::c[iPop][1];
                plint nextZ = iZ + D::c[iPop][2];
                
                if ( contained(nextX,nextY,nextZ,domain) &&
                     param.flag(nextX,nextY,nextZ) == interface )
                {
                    param.volumeFraction(nextX,nextY,nextZ) = 
                        param.mass(nextX,nextY,nextZ) / param.getDensity(nextX,nextY,nextZ);
                }    
            }
        }
    }
    global::timer("FS_EqualMassExcessReDistribution").stop();
}

}  // namespace plb

#endif  // FREE_SURFACE_MODEL_3D_HH

