/* This file is part of the Palabos library.
 *
 * Copyright (C) 2011 FlowKit Sarl
 * Avenue de Chailly 23
 * 1012 Lausanne, Switzerland
 * E-mail contact: contact@flowkit.com
 *
 * The most recent release of Palabos can be downloaded at 
 * <http://www.palabos.org/>
 *
 * The library Palabos is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * The library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef FREE_SURFACE_TENSION_3D_HH
#define FREE_SURFACE_TENSION_3D_HH

#include "core/globalDefs.h"
#include "core/geometry3D.h"
#include "core/block3D.h"
#include "latticeBoltzmann/geometricOperationTemplates.h"
#include "atomicBlock/dataProcessor3D.h"
#include "atomicBlock/blockLattice3D.h"
#include "atomicBlock/atomicContainerBlock3D.h"
#include "multiPhysics/freeSurfaceTension3D.h"

namespace plb {

template<typename T,template<typename U> class Descriptor>
void StoreSurfaceNormal3D<T,Descriptor>::process (
        Box3D domain, ScalarField3D<T>& volumeFraction, TensorField3D<T,3>& normal )
{
    T epsilon = 1.e-8;
    typedef Descriptor<T> D;
    Dot3D ofs = computeRelativeDisplacement(volumeFraction, normal);
    for (plint iX=domain.x0; iX<=domain.x1; ++iX) {
        for (plint iY=domain.y0; iY<=domain.y1; ++iY) {
            for (plint iZ=domain.z0; iZ<=domain.z1; ++iZ) {
                Array<T,D::d>& normalToInterface = normal.get(iX+ofs.x,iY+ofs.y,iZ+ofs.z);                    
                normalToInterface.resetToZero();
                // Compute surface normal "the lattice Boltzmann way".
                for (plint iPop=1; iPop < D::q; ++iPop ) {
                    plint nextX = iX + D::c[iPop][0];
                    plint nextY = iY + D::c[iPop][1];
                    plint nextZ = iZ + D::c[iPop][2];

                    normalToInterface[0] -= D::t[iPop]*D::c[iPop][0]*volumeFraction.get(nextX,nextY,nextZ);
                    normalToInterface[1] -= D::t[iPop]*D::c[iPop][1]*volumeFraction.get(nextX,nextY,nextZ);
                    normalToInterface[2] -= D::t[iPop]*D::c[iPop][2]*volumeFraction.get(nextX,nextY,nextZ);
                }
                normalToInterface *= D::invCs2;                     
                T length = norm(normalToInterface);
                if (length>epsilon) {
                    normalToInterface /= length;
                }
                else {
                    normalToInterface.resetToZero();
                }
            }
        }
    }
}


template<typename T>
bool IsoFreeSurfaceDefinition3D<T>::isInside (
            plint surfaceId, Array<plint,3> const& position ) const
{
    // This function checks if an edge-centered vertex is inside or not.
    // To do this it checks the surrounding 8 cell-centered vertices.
    Dot3D locFlag = flag->getLocation();
    using namespace freeSurfaceFlag;
    plint iX=position[0]-locFlag.x;
    plint iY=position[1]-locFlag.y;
    plint iZ=position[2]-locFlag.z;
    return isWet(flag->get(iX  ,iY  ,iZ  )) &&
           isWet(flag->get(iX-1,iY  ,iZ  )) &&
           isWet(flag->get(iX  ,iY-1,iZ  )) &&
           isWet(flag->get(iX-1,iY-1,iZ  )) &&
           isWet(flag->get(iX  ,iY  ,iZ-1)) &&
           isWet(flag->get(iX-1,iY  ,iZ-1)) &&
           isWet(flag->get(iX  ,iY-1,iZ-1)) &&
           isWet(flag->get(iX-1,iY-1,iZ-1));
}


template<typename T>
bool IsoFreeSurfaceDefinition3D<T>::isValid (
            Array<plint,3> const& position ) const
{
    // This function checks if an edge-centered vertex is valid or not.
    // To do this it checks the surrounding 8 cell-centered vertices.
    // It is not considered valid, if any of the surrounding
    // cell-centered vertices belongs to a wall.
    Dot3D locFlag = flag->getLocation();
    using namespace freeSurfaceFlag;
    plint iX=position[0]-locFlag.x;
    plint iY=position[1]-locFlag.y;
    plint iZ=position[2]-locFlag.z;

    return !(flag->get(iX  ,iY  ,iZ  ) == freeSurfaceFlag::wall ||
             flag->get(iX-1,iY  ,iZ  ) == freeSurfaceFlag::wall ||
             flag->get(iX  ,iY-1,iZ  ) == freeSurfaceFlag::wall ||
             flag->get(iX-1,iY-1,iZ  ) == freeSurfaceFlag::wall ||
             flag->get(iX  ,iY  ,iZ-1) == freeSurfaceFlag::wall ||
             flag->get(iX-1,iY  ,iZ-1) == freeSurfaceFlag::wall ||
             flag->get(iX  ,iY-1,iZ-1) == freeSurfaceFlag::wall ||
             flag->get(iX-1,iY-1,iZ-1) == freeSurfaceFlag::wall);
}

template<typename T>
Array<T,3> IsoFreeSurfaceDefinition3D<T>::getSurfacePosition (
             plint surfaceId, Array<plint,3> const& p1, Array<plint,3> const& p2 ) const
{
    using namespace freeSurfaceFlag;
    Dot3D locFlag = flag->getLocation();
    Dot3D locVF   = volumeFraction->getLocation();
    Dot3D locNorm = normals->getLocation();

    T epsilon = 1.e-8;
    // 1. Determine the ordering of vertices in inside->outside direction.
    Array<plint,3> outP1(p1), outP2(p2);
    if (isInside(surfaceId, p1)) {
        PLB_ASSERT(!isInside(surfaceId, p2));
        outP1 = p1;
        outP2 = p2;
    }
    else if (isInside(surfaceId, p2)) {
        PLB_ASSERT(!isInside(surfaceId, p1));
        outP1 = p2;
        outP2 = p1;
    }
    Array<T,3> outwardDirection(outP2-outP1);

    // 2. Determine the ordering of vertices in smallIndex->largeIndex direction.
    Array<plint,3> upP1(p1);
    Array<plint,3> perpend1, perpend2;
    if (p1[0] != p2[0]) {
        PLB_ASSERT(p1[1] == p2[1]);
        PLB_ASSERT(p1[2] == p2[2]);
        perpend1 = Array<plint,3>(0,1,0);
        perpend2 = Array<plint,3>(0,0,1);
        if (p2[0] < p1[0]) upP1=p2;
    }
    else if (p1[1] != p2[1]) {
        PLB_ASSERT(p1[0] == p2[0]);
        PLB_ASSERT(p1[2] == p2[2]);
        perpend1 = Array<plint,3>(1,0,0);
        perpend2 = Array<plint,3>(0,0,1);
        if (p2[1] < p1[1]) upP1=p2;
    }
    else if (p1[2] != p2[2]) {
        PLB_ASSERT(p1[0] == p2[0]);
        PLB_ASSERT(p1[1] == p2[1]);
        perpend1 = Array<plint,3>(1,0,0);
        perpend2 = Array<plint,3>(0,1,0);
        if (p2[2] < p1[2]) upP1=p2;
    }
    else {
        PLB_ASSERT( false );
    }

    // 3. Compute the weighted-averaged volume fraction.
    T totalWeight = T();
    T averageVolumeFraction = T();
    for (plint i1=-1; i1<=0; ++i1) {
        for (plint i2=-1; i2<=0; ++i2) {
            Array<plint,3> pos = upP1 + i1*perpend1 + i2*perpend2;
            PLB_ASSERT( isWet((*flag).get(pos[0]-locFlag.x,pos[1]-locFlag.y,pos[2]-locFlag.z) ) );
            T weight = dot (
                           normals->get(pos[0]-locNorm.x, pos[1]-locNorm.y, pos[2]-locNorm.z),
                           outwardDirection );
            weight = weight*weight;
            if (weight>(T)0) {
                totalWeight += weight;
                averageVolumeFraction += weight * volumeFraction->get(pos[0]-locVF.x, pos[1]-locVF.y, pos[2]-locVF.z);
            }
        }
    }
    if (totalWeight>epsilon) {
        averageVolumeFraction /= totalWeight;
    }
    else {
        averageVolumeFraction = (T)0.5;
    }

    T eps=1.e-5;
    if (averageVolumeFraction>1.-eps) {
        averageVolumeFraction = 1.-eps;
    } else if (averageVolumeFraction < eps) {
        averageVolumeFraction = eps;
    }
    else if (averageVolumeFraction<eps) {
        averageVolumeFraction = eps;
    }
    // 4. Position the interface intersection, in outward direction.
    //    Subtract 0.5, because the edge-centered grid is staggered wrt
    //    the cell-centered grid.
    return Array<T,3>(outP1) + averageVolumeFraction*outwardDirection
           - Array<T,3>((T)0.5,(T)0.5,(T)0.5);
}

template<typename T>
void IsoFreeSurfaceDefinition3D<T>::setArguments(std::vector<AtomicBlock3D*> const& arguments)
{
    PLB_ASSERT( arguments.size()==3 );

    flag = dynamic_cast<ScalarField3D<int>*>(arguments[0]);
    PLB_ASSERT( flag );
    volumeFraction = dynamic_cast<ScalarField3D<T>*>(arguments[1]);
    PLB_ASSERT( volumeFraction );
    normals = dynamic_cast<TensorField3D<T,3>*>(arguments[2]);
    PLB_ASSERT( normals );
}

template<typename T>
IsoFreeSurfaceDefinition3D<T>* IsoFreeSurfaceDefinition3D<T>::clone() const {
    return new IsoFreeSurfaceDefinition3D(*this);
}

template<typename T>
plint IsoFreeSurfaceDefinition3D<T>::getNumArgs() const {
    return 3;
}

template<typename T>
std::vector<plint> IsoFreeSurfaceDefinition3D<T>::getSurfaceIds() const {
    std::vector<plint> surfaceIds;
    surfaceIds.push_back(0); // Only one surface can be produced, ID does not matter.
    return surfaceIds;
}


template<typename T>
std::auto_ptr<MultiContainerBlock3D> storeTrianglesOnEdges(MultiScalarField3D<T>& volumeFraction, MultiScalarField3D<int>& flag)
{
    // Use a smoothen volume fraction for computing the normal (but not for computing the triangles).
    std::auto_ptr<MultiScalarField3D<T> > smoothVF(lbmSmoothen<T,descriptors::D3Q19Descriptor>(volumeFraction, volumeFraction.getBoundingBox()));
    MultiTensorField3D<T,3> normals(volumeFraction);
    applyProcessingFunctional(new StoreSurfaceNormal3D<T,descriptors::D3Q19Descriptor>(), normals.getBoundingBox(), *smoothVF, normals);

    std::auto_ptr<MultiContainerBlock3D> triangleContainer(new MultiContainerBlock3D(volumeFraction));

    std::vector<MultiBlock3D*> args;
    // First argument is for MarchingCubeSurfaces3D.
    args.push_back(triangleContainer.get());
    // Following arguments are for IsoFreeSurfaceDefinition3D.
    args.push_back(&flag);
    args.push_back(&volumeFraction);
    args.push_back(&normals);

    std::vector<plint> ids;
    ids.push_back(0);
    bool edgeOriented = true;
    applyProcessingFunctional (
            new MarchingCubeSurfaces3D<T>(ids, new IsoFreeSurfaceDefinition3D<T>(), edgeOriented), volumeFraction.getBoundingBox(), args );

    return triangleContainer;
}

}  // namespace plb

#endif  // FREE_SURFACE_TENSION_3D_HH

