/* This file is part of the Palabos library.
 *
 * Copyright (C) 2011 FlowKit Sarl
 * Avenue de Chailly 23
 * 1012 Lausanne, Switzerland
 * E-mail contact: contact@flowkit.com
 *
 * The most recent release of Palabos can be downloaded at 
 * <http://www.palabos.org/>
 *
 * The library Palabos is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * The library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/* Main author: Andrea */

#ifndef FREE_SURFACE_UTIL_3D_H
#define FREE_SURFACE_UTIL_3D_H

#include "core/globalDefs.h"
#include "core/plbDebug.h"
#include "multiBlock/multiContainerBlock3D.h"
#include "multiBlock/multiDataField3D.h"
#include "multiBlock/multiBlockLattice3D.h"
#include <vector>
#include <set>
#include <string>

namespace plb {

/// Constants used in a free surface flag matrix for cell tagging.
namespace freeSurfaceFlag {
    enum Flag {empty=0, interface=1, fluid=2, wall=4};
    inline std::string flagToString(int flag) {
        switch(flag) {
            case empty:     return "empty";
            case interface: return "interface";
            case fluid:     return "fluid";
            case wall:      return "wall";
            default: PLB_ASSERT( false );
        }
    }
    inline Flag invert(int flag) {
        switch(flag) {
            case empty:     return fluid;
            case interface: return interface;
            case fluid:     return empty;
            case wall:      return wall;
            default: PLB_ASSERT( false );
        }
    }
    inline bool isWet(int flag) {
        return flag==interface || flag==fluid;
    }
}

/// Create a parameter-list for most free-surface data processors.
template< typename T,template<typename U> class FluidDescriptor>
std::vector<MultiBlock3D*> aggregateFreeSurfaceParams (
        MultiBlockLattice3D<T,FluidDescriptor>& fluid, MultiScalarField3D<T>& rhoBar,
        MultiTensorField3D<T,3>& j, MultiScalarField3D<T>& mass,
        MultiScalarField3D<T>& volumeFraction, MultiScalarField3D<int>& flag,
        MultiContainerBlock3D& interfaceLists )
{
    std::vector<MultiBlock3D*> aggregation;
    aggregation.push_back(&fluid);
    aggregation.push_back(&rhoBar);
    aggregation.push_back(&j);
    aggregation.push_back(&mass);
    aggregation.push_back(&volumeFraction);
    aggregation.push_back(&flag);
    aggregation.push_back(&interfaceLists);
    return aggregation;
}
    
/// Create a parameter-list for most free-surface data processors, in a two-fluid system.
template< typename T,template<typename U> class FluidDescriptor>
std::vector<MultiBlock3D*> aggregateFreeSurfaceParams (
        MultiBlockLattice3D<T,FluidDescriptor>& fluid1, MultiBlockLattice3D<T,FluidDescriptor>& fluid2,
        MultiScalarField3D<T>& rhoBar1, MultiScalarField3D<T>& rhoBar2,
        MultiTensorField3D<T,3>& j1, MultiTensorField3D<T,3>& j2, MultiScalarField3D<T>& mass,
        MultiScalarField3D<T>& volumeFraction, MultiScalarField3D<int>& flag,
        MultiContainerBlock3D& interfaceLists )
{
    std::vector<MultiBlock3D*> aggregation;
    aggregation.push_back(&fluid1);
    aggregation.push_back(&rhoBar1);
    aggregation.push_back(&j1);
    aggregation.push_back(&mass);
    aggregation.push_back(&volumeFraction);
    aggregation.push_back(&flag);
    aggregation.push_back(&interfaceLists);
    aggregation.push_back(&fluid2);
    aggregation.push_back(&rhoBar2);
    aggregation.push_back(&j2);
    return aggregation;
}
    
/// Data structure for holding lists of cells along the free surface in an AtomicContainerBlock.
template< typename T,template<typename U> class FluidDescriptor>
struct InterfaceLists : public ContainerBlockData {
    typedef Array<plint,FluidDescriptor<T>::d> Node;
    /// Holds all nodes which have excess mass.
    std::map<Node,T> massExcess;
    /// Holds all nodes that need to change status from interface to fluid.
    std::set<Node>   interfaceToFluid;
    /// Holds all nodes that need to change status from interface to empty.
    std::set<Node>   interfaceToEmpty;
    /// Holds all nodes that need to change status from empty to interface.
    std::set<Node>   emptyToInterface;

    virtual InterfaceLists<T,FluidDescriptor>* clone() const {
        return new InterfaceLists<T,FluidDescriptor>(*this);
    }
};

/// A wrapper offering convenient access to the free-surface data provided to
/// data processors. Avoids verbous casting, asserting, etc.
template< typename T,template<typename U> class FluidDescriptor>
class FreeSurfaceProcessorParam3D {
public:
    typedef typename InterfaceLists<T,FluidDescriptor>::Node Node;
    FreeSurfaceProcessorParam3D(std::vector<AtomicBlock3D*>& atomicBlocks)
    {
        // Seven blocks for a single free-surface system,
        // and ten blocks for a binary system.
        PLB_ASSERT( atomicBlocks.size()==7 || atomicBlocks.size()==10 ); 

        fluid_ = dynamic_cast<BlockLattice3D<T,FluidDescriptor>*>(atomicBlocks[0]);
        PLB_ASSERT(fluid_);

        rhoBar_ = dynamic_cast<ScalarField3D<T>*>(atomicBlocks[1]);
        PLB_ASSERT(rhoBar_);

        j_ = dynamic_cast<TensorField3D<T,3>*>(atomicBlocks[2]);
        PLB_ASSERT(j_);

        mass_ = dynamic_cast<ScalarField3D<T>*>(atomicBlocks[3]);               
        PLB_ASSERT(mass_);

        volumeFraction_ = dynamic_cast<ScalarField3D<T>*>(atomicBlocks[4]);
        PLB_ASSERT(volumeFraction_);

        flag_ = dynamic_cast<ScalarField3D<int>*>(atomicBlocks[5]);
        PLB_ASSERT(flag_);

        containerInterfaceLists_ = dynamic_cast<AtomicContainerBlock3D*>(atomicBlocks[6]);
        PLB_ASSERT(containerInterfaceLists_);

        interfaceLists_ = dynamic_cast<InterfaceLists<T,FluidDescriptor>*>(containerInterfaceLists_->getData());
        PLB_ASSERT(interfaceLists_);

        absoluteOffset     = fluid_->getLocation();
        relativeOffsetRhoBar = computeRelativeDisplacement(*fluid_,*rhoBar_);      
        relativeOffsetJ    = computeRelativeDisplacement(*fluid_,*j_);      
        relativeOffsetMass = computeRelativeDisplacement(*fluid_,*mass_);      
        relativeOffsetVF   = computeRelativeDisplacement(*fluid_,*volumeFraction_);      
        relativeOffsetFS   = computeRelativeDisplacement(*fluid_,*flag_); 
    }
    Cell<T,FluidDescriptor>& cell(plint iX, plint iY, plint iZ) { return fluid_->get(iX,iY,iZ); }
    T& mass(plint iX, plint iY, plint iZ) {
        return mass_->get(iX+relativeOffsetMass.x,iY+relativeOffsetMass.y,iZ+relativeOffsetMass.z);
    }
    T& volumeFraction(plint iX, plint iY, plint iZ) {
        return volumeFraction_->get(iX+relativeOffsetVF.x,iY+relativeOffsetVF.y,iZ+relativeOffsetVF.z);
    }
    int& flag(plint iX, plint iY, plint iZ) {
        return flag_->get(iX+relativeOffsetFS.x,iY+relativeOffsetFS.y,iZ+relativeOffsetFS.z);
    }
    void setForce(plint iX, plint iY, plint iZ, Array<T,3> const& force) {
            force.to_cArray(cell(iX,iY,iZ).getExternal(forceOffset));
    }
    Array<T,3> getForce(plint iX, plint iY, plint iZ) {
        Array<T,3> force; force.from_cArray(cell(iX,iY,iZ).getExternal(forceOffset));
        return force;
    }
    void setMomentum(plint iX, plint iY, plint iZ, Array<T,3> const& momentum) {
        j_->get(iX+relativeOffsetJ.x,iY+relativeOffsetJ.y,iZ+relativeOffsetJ.z) = momentum;
    }
    Array<T,3> getMomentum(plint iX, plint iY, plint iZ) {
        return j_->get(iX+relativeOffsetJ.x,iY+relativeOffsetJ.y,iZ+relativeOffsetJ.z);
    }
    T getDensity(plint iX, plint iY, plint iZ) {
        return FluidDescriptor<T>::fullRho (
                    rhoBar_->get(iX+relativeOffsetRhoBar.x, iY+relativeOffsetRhoBar.y, iZ+relativeOffsetRhoBar.z) );
    }
    void setDensity(plint iX, plint iY, plint iZ, T rho) {
        rhoBar_->get(iX+relativeOffsetRhoBar.x, iY+relativeOffsetRhoBar.y, iZ+relativeOffsetRhoBar.z)
            = FluidDescriptor<T>::rhoBar(rho);
    }

    void attributeDynamics(plint iX, plint iY, plint iZ, Dynamics<T,FluidDescriptor>* dynamics) {
        fluid_->attributeDynamics(iX,iY,iZ, dynamics);
    }

    std::map<Node,T>& massExcess() { return interfaceLists_ -> massExcess; }
    std::set<Node>& interfaceToFluid() { return interfaceLists_ -> interfaceToFluid; }
    std::set<Node>& interfaceToEmpty() { return interfaceLists_ -> interfaceToEmpty; }
    std::set<Node>& emptyToInterface() { return  interfaceLists_ -> emptyToInterface; }

    Dot3D const& absOffset() const { return absoluteOffset; }
    Box3D getBoundingBox() const { return volumeFraction_->getBoundingBox(); }
private:
    BlockLattice3D<T,FluidDescriptor>* fluid_;
    ScalarField3D<T>* rhoBar_;
    TensorField3D<T,3>* j_;
    ScalarField3D<T>* mass_;
    ScalarField3D<T>* volumeFraction_;
    ScalarField3D<int>* flag_;
    AtomicContainerBlock3D* containerInterfaceLists_;
    InterfaceLists<T,FluidDescriptor>* interfaceLists_;

    Dot3D absoluteOffset, relativeOffsetRhoBar, relativeOffsetJ, relativeOffsetMass, relativeOffsetVF, relativeOffsetFS;

    static const int forceOffset = FluidDescriptor<T>::ExternalField::forceBeginsAt;
};

/// A wrapper offering convenient access to the free-surface data provided to
/// data processors. Avoids verbous casting, asserting, etc.
template< typename T,template<typename U> class FluidDescriptor>
class BinaryFreeSurfaceProcessorParam3D {
public:
    typedef typename InterfaceLists<T,FluidDescriptor>::Node Node;
    BinaryFreeSurfaceProcessorParam3D(std::vector<AtomicBlock3D*>& atomicBlocks)
    {
        // Seven blocks for the first fluid, three blocks for the second.
        PLB_ASSERT( atomicBlocks.size()==10 ); 

        fluid1_ = dynamic_cast<BlockLattice3D<T,FluidDescriptor>*>(atomicBlocks[0]);
        PLB_ASSERT(fluid1_);

        volumeFraction_ = dynamic_cast<ScalarField3D<T>*>(atomicBlocks[4]);
        PLB_ASSERT(volumeFraction_);

        flag_ = dynamic_cast<ScalarField3D<int>*>(atomicBlocks[5]);
        PLB_ASSERT(flag_);

        fluid2_ = dynamic_cast<BlockLattice3D<T,FluidDescriptor>*>(atomicBlocks[7]);
        PLB_ASSERT(fluid2_);

        rhoBar_ = dynamic_cast<ScalarField3D<T>*>(atomicBlocks[8]);
        PLB_ASSERT(rhoBar_);

        j_ = dynamic_cast<TensorField3D<T,3>*>(atomicBlocks[9]);
        PLB_ASSERT(j_);

        relativeOffsetFluid2 = computeRelativeDisplacement(*fluid1_,*fluid2_);      
        relativeOffsetVF   = computeRelativeDisplacement(*fluid1_,*volumeFraction_);      
        relativeOffsetFS   = computeRelativeDisplacement(*fluid1_,*flag_); 
        relativeOffsetRhoBar = computeRelativeDisplacement(*fluid1_,*rhoBar_);      
        relativeOffsetJ = computeRelativeDisplacement(*fluid1_,*j_);      
    }

    Cell<T,FluidDescriptor>& cell(plint iX, plint iY, plint iZ) {
        PLB_ASSERT(fluid2_);
        return fluid2_->get(iX+relativeOffsetFluid2.x,iY+relativeOffsetFluid2.y,iZ+relativeOffsetFluid2.z);
    }
    void setMomentum(plint iX, plint iY, plint iZ, Array<T,3> const& momentum) {
        PLB_ASSERT(j_);
        j_->get(iX+relativeOffsetJ.x,iY+relativeOffsetJ.y,iZ+relativeOffsetJ.z) = momentum;
    }
    Array<T,3> getMomentum(plint iX, plint iY, plint iZ) {
        PLB_ASSERT(j_);
        return j_->get(iX+relativeOffsetJ.x,iY+relativeOffsetJ.y,iZ+relativeOffsetJ.z);
    }
    T getDensity(plint iX, plint iY, plint iZ) {
        PLB_ASSERT(rhoBar_);
        return FluidDescriptor<T>::fullRho (
                    rhoBar_->get(iX+relativeOffsetRhoBar.x, iY+relativeOffsetRhoBar.y, iZ+relativeOffsetRhoBar.z) );
    }
    void setDensity(plint iX, plint iY, plint iZ, T rho) {
        PLB_ASSERT(rhoBar_);
        rhoBar_->get(iX+relativeOffsetRhoBar.x, iY+relativeOffsetRhoBar.y, iZ+relativeOffsetRhoBar.z)
            = FluidDescriptor<T>::rhoBar(rho);
    }
    void attributeDynamics(plint iX, plint iY, plint iZ, Dynamics<T,FluidDescriptor>* dynamics) {
        fluid2_->attributeDynamics(iX+relativeOffsetFluid2.x,iY+relativeOffsetFluid2.y,iZ+relativeOffsetFluid2.z, dynamics);
    }
    void setForce(plint iX, plint iY, plint iZ, Array<T,3> const& force) {
            force.to_cArray(cell(iX,iY,iZ).getExternal(forceOffset));
    }
    Array<T,3> getForce(plint iX, plint iY, plint iZ) {
        Array<T,3> force; force.from_cArray(cell(iX,iY,iZ).getExternal(forceOffset));
        return force;
    }
    // The flag matrix is defined for fluid 1. It is therefore inverted here.
    int flag(plint iX, plint iY, plint iZ) {
        return freeSurfaceFlag::invert(flag_->get(iX+relativeOffsetFS.x,iY+relativeOffsetFS.y,iZ+relativeOffsetFS.z));
    }
    // The volume fraction is defined for fluid 1. It is therefore inverted here.
    T volumeFraction(plint iX, plint iY, plint iZ) {
        return (T)1. - volumeFraction_->get(iX+relativeOffsetVF.x,iY+relativeOffsetVF.y,iZ+relativeOffsetVF.z);
    }
private:
    BlockLattice3D<T,FluidDescriptor> *fluid1_, *fluid2_;
    ScalarField3D<T>* volumeFraction_;
    ScalarField3D<int>* flag_;
    ScalarField3D<T>* rhoBar_;
    TensorField3D<T,3>* j_;

    Dot3D relativeOffsetVF, relativeOffsetFS;
    Dot3D relativeOffsetFluid2, relativeOffsetRhoBar, relativeOffsetJ;

    static const int forceOffset = FluidDescriptor<T>::ExternalField::forceBeginsAt;
};


inline void freeSurfaceProfilerOutput() {
    double collideAndStream = global::timer("collideAndStream").getTime();
    double FS_MassChange = global::timer("FS_MassChange").getTime();
    double FS_FreeSurfaceCompletion = global::timer("FS_FreeSurfaceCompletion").getTime();
    double FS_FreeSurfaceMacroscopic = global::timer("FS_FreeSurfaceMacroscopic").getTime();
    double FS_ComputeInterfaceLists = global::timer("FS_ComputeInterfaceLists").getTime();
    double FS_IniInterfaceToAnyNodes = global::timer("FS_IniInterfaceToAnyNodes").getTime();
    double FS_IniEmptyToInterfaceNodes = global::timer("FS_IniEmptyToInterfaceNodes").getTime();
    double FS_RemoveFalseInterfaceCells = global::timer("FS_RemoveFalseInterfaceCells").getTime();
    double FS_EqualMassExcessReDistribution = global::timer("FS_EqualMassExcessReDistribution").getTime();

    double total = collideAndStream + FS_MassChange + FS_FreeSurfaceCompletion + FS_FreeSurfaceMacroscopic + FS_ComputeInterfaceLists +
              FS_IniInterfaceToAnyNodes + FS_IniEmptyToInterfaceNodes + FS_RemoveFalseInterfaceCells + FS_EqualMassExcessReDistribution;

    pcout << "PROFILER" << std::endl;
    pcout << "collideAndStream " << collideAndStream/total << std::endl;
    pcout << "FS_MassChange " << FS_MassChange/total << std::endl;
    pcout << "FS_FreeSurfaceCompletion " << FS_FreeSurfaceCompletion/total << std::endl;
    pcout << "FS_FreeSurfaceMacroscopic " << FS_FreeSurfaceMacroscopic/total << std::endl;
    pcout << "FS_ComputeInterfaceLists " << FS_ComputeInterfaceLists/total << std::endl;
    pcout << "FS_IniInterfaceToAnyNodes " << FS_IniInterfaceToAnyNodes/total << std::endl;
    pcout << "FS_IniEmptyToInterfaceNodes " << FS_IniEmptyToInterfaceNodes/total << std::endl;
    pcout << "FS_RemoveFalseInterfaceCells " << FS_RemoveFalseInterfaceCells/total << std::endl;
    pcout << "FS_EqualMassExcessReDistribution " << FS_EqualMassExcessReDistribution/total << std::endl;
    pcout << "END PROFILER" << std::endl;

    global::timer("collideAndStream").reset();
    global::timer("FS_MassChange").reset();
    global::timer("FS_FreeSurfaceCompletion").reset();
    global::timer("FS_FreeSurfaceMacroscopic").reset();
    global::timer("FS_ComputeInterfaceLists").reset();
    global::timer("FS_IniInterfaceToAnyNodes").reset();
    global::timer("FS_IniEmptyToInterfaceNodes").reset();
    global::timer("FS_RemoveFalseInterfaceCells").reset();
    global::timer("FS_EqualMassExcessReDistribution").reset();
}

}  // namespace plb

#endif  // FREE_SURFACE_UTIL_3D_H

