/* This file is part of the Palabos library.
 *
 * Copyright (C) 2011 FlowKit Sarl
 * Avenue de Chailly 23
 * 1012 Lausanne, Switzerland
 * E-mail contact: contact@flowkit.com
 *
 * The most recent release of Palabos can be downloaded at 
 * <http://www.palabos.org/>
 *
 * The library Palabos is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * The library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef PARTICLE_2D_H
#define PARTICLE_2D_H

#include "core/globalDefs.h"
#include "core/array.h"
#include "atomicBlock/blockLattice2D.h"
#include <vector>

namespace plb {

template<typename T, template<typename U> class Descriptor>
class Particle2D {
public:
    Particle2D();
    /// The tag does not need to be unique and exists only for the
    ///   user's convenience.
    Particle2D(plint tag_, Array<T,2> const& position_);
    virtual ~Particle2D() { }
    virtual void fluidToParticle(BlockLattice2D<T,Descriptor>& fluid) =0;
    virtual void advance() =0;
    Array<T,2> const& getPosition() const { return position; }
    Array<T,2>& getPosition() { return position; }
    virtual void reset(Array<T,2> const& position);
    virtual void serialize(HierarchicSerializer& serializer) const;
    virtual void unserialize(HierarchicUnserializer& unserializer);
    /// Return a unique ID for this class.
    virtual int getId() const =0;
    /// Return this particle's tag (not necessarily unique).
    plint getTag() const;
    /// Set a tag for this particle (not necessarily unique).
    void setTag(plint tag_);
    virtual Particle2D<T,Descriptor>* clone() const =0;
    virtual bool getScalar(plint whichScalar, T& scalar) const;
    virtual bool getVector(plint whichVector, Array<T,2>& vector) const;
    virtual bool getTensor(plint whichVector, Array<T,SymmetricTensorImpl<T,2>::n>& tensor) const;
    virtual bool setScalars(std::vector<T> const& scalars);
    virtual bool setVectors(std::vector<Array<T,2> > const& vectors);
    virtual bool setTensors(std::vector<Array<T,SymmetricTensorImpl<T,2>::n> > const& tensors);
private:
    plint tag;
    Array<T,2> position;
};

/// Serialize the whole particle chain into a byte-stream, and append it to
///   the vector data.
template<typename T, template<typename U> class Descriptor>
void serialize(Particle2D<T,Descriptor> const& particle, std::vector<char>& data);

/// Unserialize all data into newly generated particle objects.
template<typename T, template<typename U> class Descriptor>
void generateAndUnserializeParticles (
        std::vector<char> const& data,
        std::vector<Particle2D<T,Descriptor>*>& particle );

template<typename T, template<typename U> class Descriptor>
class PointParticle2D : public Particle2D<T,Descriptor> {
public:
    PointParticle2D();
    PointParticle2D(plint tag_, Array<T,2> const& position_, Array<T,2> const& velocity_);
    virtual void fluidToParticle(BlockLattice2D<T,Descriptor>& fluid);
    virtual void advance();
    virtual void serialize(HierarchicSerializer& serializer) const;
    virtual void unserialize(HierarchicUnserializer& unserializer);
    virtual int getId() const;
    virtual void reset(Array<T,2> const& position);
    virtual PointParticle2D<T,Descriptor>* clone() const;
    virtual bool getVector(plint whichVector, Array<T,2>& vector) const;
    virtual bool setVectors(std::vector<Array<T,2> > const& vectors);
private:
    Array<T,2> velocity;
    static int id;
};

}  // namespace plb

#endif  // PARTICLE_2D_H
