/* This file is part of the Palabos library.
 *
 * Copyright (C) 2011 FlowKit Sarl
 * Avenue de Chailly 23
 * 1012 Lausanne, Switzerland
 * E-mail contact: contact@flowkit.com
 *
 * The most recent release of Palabos can be downloaded at 
 * <http://www.palabos.org/>
 *
 * The library Palabos is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * The library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef PARTICLE_2D_HH
#define PARTICLE_2D_HH

#include "core/globalDefs.h"
#include "particles/particle2D.h"
#include "particles/particleIdentifiers2D.h"
#include "finiteDifference/interpolations2D.h"

namespace plb {


/* *************** class Particle2D ***************************************** */

template<typename T, template<typename U> class Descriptor>
Particle2D<T,Descriptor>::Particle2D()
    : tag(0),
      position(T(),T(),T())
{ }

template<typename T, template<typename U> class Descriptor>
Particle2D<T,Descriptor>::Particle2D(plint tag_, Array<T,2> const& position_)
    : tag(tag_),
      position(position_)
{ }

template<typename T, template<typename U> class Descriptor>
void Particle2D<T,Descriptor>::reset(Array<T,2> const& position_)
{
    position = position_;
}


template<typename T, template<typename U> class Descriptor>
void Particle2D<T,Descriptor>::serialize(HierarchicSerializer& serializer) const
{
    serializer.addValue(tag);
    serializer.addValues<T,2>(position);
}

template<typename T, template<typename U> class Descriptor>
void Particle2D<T,Descriptor>::unserialize(HierarchicUnserializer& unserializer)
{
    unserializer.readValue(tag);
    unserializer.readValues<T,2>(position);
}

template<typename T, template<typename U> class Descriptor>
plint Particle2D<T,Descriptor>::getTag() const {
    return tag;
}

template<typename T, template<typename U> class Descriptor>
void Particle2D<T,Descriptor>::setTag(plint tag_) {
    tag = tag_;
}


template<typename T, template<typename U> class Descriptor>
bool Particle2D<T,Descriptor>::getVector(plint whichVector, Array<T,2>& vector) const {
    return false;
}

template<typename T, template<typename U> class Descriptor>
bool Particle2D<T,Descriptor>::getScalar(plint whichScalar, T& scalar) const {
    return false;
}

template<typename T, template<typename U> class Descriptor>
bool Particle2D<T,Descriptor>::getTensor(plint whichVector, Array<T,SymmetricTensorImpl<T,2>::n>& tensor) const
{
    return false;
}

template<typename T, template<typename U> class Descriptor>
bool Particle2D<T,Descriptor>::setScalars(std::vector<T> const& scalars)
{
    return false;
}

template<typename T, template<typename U> class Descriptor>
bool Particle2D<T,Descriptor>::setVectors(std::vector<Array<T,2> > const& vectors)
{
    return false;
}

template<typename T, template<typename U> class Descriptor>
bool Particle2D<T,Descriptor>::setTensors (
        std::vector<Array<T,SymmetricTensorImpl<T,2>::n> > const& tensors )
{
    return false;
}

/* *************** class PointParticle2D ************************************ */

template<typename T, template<typename U> class Descriptor>
int PointParticle2D<T,Descriptor>::id =
        meta::registerPointParticle2D<T,Descriptor,PointParticle2D<T,Descriptor> >("Point");

template<typename T, template<typename U> class Descriptor>
PointParticle2D<T,Descriptor>::PointParticle2D()
    : Particle2D<T,Descriptor>(),
      velocity(T(),T(),T())
{ }

template<typename T, template<typename U> class Descriptor>
PointParticle2D<T,Descriptor>::PointParticle2D(plint tag_, Array<T,2> const& position_, Array<T,2> const& velocity_)
    : Particle2D<T,Descriptor>(tag_, position_),
      velocity(velocity_)
{ }

template<typename T, template<typename U> class Descriptor>
int PointParticle2D<T,Descriptor>::getId() const {
    return id;
}

template<typename T, template<typename U> class Descriptor>
void PointParticle2D<T,Descriptor>::reset(Array<T,2> const& position_)
{
    Particle2D<T,Descriptor>::reset(position_);
    velocity.resetToZero();
}

template<typename T, template<typename U> class Descriptor>
void PointParticle2D<T,Descriptor>::fluidToParticle(BlockLattice2D<T,Descriptor>& fluid)
{
    Array<T,2> position(this->getPosition());
    std::vector<Dot2D> pos(4);
    std::vector<T> weights(4);
    linearInterpolationCoefficients(fluid, position, pos, weights);
    Array<T,2> j;
    T rhoBar;
    velocity.resetToZero();
    for (plint iCell=0; iCell<4; ++iCell) {
        getRhoBarJ(fluid.get(pos[iCell].x,pos[iCell].y), rhoBar, j);
        velocity += weights[iCell]*j;
    }
}

template<typename T, template<typename U> class Descriptor>
void PointParticle2D<T,Descriptor>::advance() {
    PLB_ASSERT( norm(velocity)<1. );
    this->getPosition() += velocity;
}

template<typename T, template<typename U> class Descriptor>
void PointParticle2D<T,Descriptor>::serialize(HierarchicSerializer& serializer) const
{
    Particle2D<T,Descriptor>::serialize(serializer);
    serializer.addValues<T,2>(velocity);
}

template<typename T, template<typename U> class Descriptor>
void PointParticle2D<T,Descriptor>::unserialize(HierarchicUnserializer& unserializer)
{
    Particle2D<T,Descriptor>::unserialize(unserializer);
    unserializer.readValues<T,2>(velocity);
}

template<typename T, template<typename U> class Descriptor>
PointParticle2D<T,Descriptor>* PointParticle2D<T,Descriptor>::clone() const {
    return new PointParticle2D<T,Descriptor>(*this);
}

template<typename T, template<typename U> class Descriptor>
void serialize(Particle2D<T,Descriptor> const& particle, std::vector<char>& data)
{
    HierarchicSerializer serializer(data, particle.getId());
    particle.serialize(serializer);
}

template<typename T, template<typename U> class Descriptor>
void generateAndUnserializeParticles (
        std::vector<char> const& data,
        std::vector<Particle2D<T,Descriptor>*>& particles )
{
    pluint serializerPos = 0;
    while (serializerPos < data.size()) {
        HierarchicUnserializer unserializer(data, serializerPos);
        particles.push_back (
                meta::particleRegistration2D<T,Descriptor>().generate(unserializer) );
        serializerPos = unserializer.getCurrentPos();
    }
}


template<typename T, template<typename U> class Descriptor>
bool PointParticle2D<T,Descriptor>::getVector(plint whichVector, Array<T,2>& vector) const {
    if (whichVector==0) {
        vector = velocity;
        return true;
    }
    return Particle2D<T,Descriptor>::getVector(whichVector, vector);
}


template<typename T, template<typename U> class Descriptor>
bool PointParticle2D<T,Descriptor>::setVectors(std::vector<Array<T,2> > const& vectors)
{
    if (vectors.size()==1) {
        velocity = vectors[0];
        return true;
    }
    return false;
}


}  // namespace plb

#endif  // PARTICLE_2D_H
