/* This file is part of the Palabos library.
 *
 * Copyright (C) 2011 FlowKit Sarl
 * Avenue de Chailly 23
 * 1012 Lausanne, Switzerland
 * E-mail contact: contact@flowkit.com
 *
 * The most recent release of Palabos can be downloaded at 
 * <http://www.palabos.org/>
 *
 * The library Palabos is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * The library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef PARTICLE_PROCESSING_FUNCTIONAL_2D_HH
#define PARTICLE_PROCESSING_FUNCTIONAL_2D_HH

#include "particles/particleProcessingFunctional2D.h"
#include "particles/particleField2D.h"
#include "dataProcessors/metaStuffFunctional2D.h"
#include "core/plbDebug.h"
#include "core/blockStatistics.h"
#include "atomicBlock/atomicBlock2D.h"
#include "atomicBlock/blockLattice2D.h"

namespace plb {

/* ******** CountParticlesFunctional2D *********************************** */

template<typename T, template<typename U> class Descriptor>
CountParticlesFunctional2D<T,Descriptor>::CountParticlesFunctional2D()
    : numParticlesId(this->getStatistics().subscribeIntSum())
{ }

template<typename T, template<typename U> class Descriptor>
void CountParticlesFunctional2D<T,Descriptor>::processGenericBlocks (
        Box2D domain, std::vector<AtomicBlock2D*> blocks )
{
    PLB_PRECONDITION( blocks.size()==1 );
    ParticleField2D<T,Descriptor>& particleField
        = *dynamic_cast<ParticleField2D<T,Descriptor>*>(blocks[0]);
    std::vector<Particle2D<T,Descriptor>*> particles;
    particleField.findParticles(domain, particles);
    this->getStatistics().gatherIntSum(numParticlesId, (plint)particles.size());
}

template<typename T, template<typename U> class Descriptor>
CountParticlesFunctional2D<T,Descriptor>* CountParticlesFunctional2D<T,Descriptor>::clone() const {
    return new CountParticlesFunctional2D<T,Descriptor>(*this);
}

template<typename T, template<typename U> class Descriptor>
void CountParticlesFunctional2D<T,Descriptor>::getTypeOfModification(std::vector<modif::ModifT>& modified) const {
    modified[0] = modif::nothing;
}

template<typename T, template<typename U> class Descriptor>
BlockDomain::DomainT CountParticlesFunctional2D<T,Descriptor>::appliesTo() const {
    return BlockDomain::bulk;
}

template<typename T, template<typename U> class Descriptor>
plint CountParticlesFunctional2D<T,Descriptor>::getNumParticles() const {
    return this->getStatistics().getIntSum(numParticlesId);
}


/* ******** AverageParticleVelocityFunctional2D *********************************** */

template<typename T, template<typename U> class Descriptor>
AverageParticleVelocityFunctional2D<T,Descriptor>::AverageParticleVelocityFunctional2D()
    : averageVelocityId (
            Array<T,2>( this->getStatistics().subscribeAverage(),
                        this->getStatistics().subscribeAverage() ) )
{ }

template<typename T, template<typename U> class Descriptor>
void AverageParticleVelocityFunctional2D<T,Descriptor>::processGenericBlocks (
        Box2D domain, std::vector<AtomicBlock2D*> blocks )
{
    PLB_PRECONDITION( blocks.size()==1 );
    ParticleField2D<T,Descriptor>& particleField = *dynamic_cast<ParticleField2D<T,Descriptor>*>(blocks[0]);
    std::vector<Particle2D<T,Descriptor>*> particles;
    particleField.findParticles(domain, particles);
    for (pluint iParticle=0; iParticle<particles.size(); ++iParticle) {
        Array<T,2> velocity;
        particles[iParticle]->getVelocity(velocity);
        this->getStatistics().gatherAverage(averageVelocityId[0], velocity[0]);
        this->getStatistics().gatherAverage(averageVelocityId[1], velocity[1]);
        this->getStatistics().incrementStats();
    }
}

template<typename T, template<typename U> class Descriptor>
AverageParticleVelocityFunctional2D<T,Descriptor>* AverageParticleVelocityFunctional2D<T,Descriptor>::clone() const {
    return new AverageParticleVelocityFunctional2D<T,Descriptor>(*this);
}

template<typename T, template<typename U> class Descriptor>
void AverageParticleVelocityFunctional2D<T,Descriptor>::getTypeOfModification(std::vector<modif::ModifT>& modified) const {
    modified[0] = modif::nothing;
}

template<typename T, template<typename U> class Descriptor>
BlockDomain::DomainT AverageParticleVelocityFunctional2D<T,Descriptor>::appliesTo() const {
    return BlockDomain::bulk;
}

template<typename T, template<typename U> class Descriptor>
Array<T,2> AverageParticleVelocityFunctional2D<T,Descriptor>::getAverageParticleVelocity() const {
    return Array<T,2> (
        this->getStatistics().getAverage(averageVelocityId[0]),
        this->getStatistics().getAverage(averageVelocityId[1]) );
}


/* ******** InjectParticlesFunctional2D *********************************** */

template<typename T, template<typename U> class Descriptor>
InjectParticlesFunctional2D<T,Descriptor>::InjectParticlesFunctional2D(std::vector<Particle2D<T,Descriptor>*>& particles_)
    : particles(particles_)
{ }

template<typename T, template<typename U> class Descriptor>
void InjectParticlesFunctional2D<T,Descriptor>::processGenericBlocks (
        Box2D domain, std::vector<AtomicBlock2D*> blocks )
{
    PLB_PRECONDITION( blocks.size()==1 );
    ParticleField2D<T,Descriptor>& particleField = *dynamic_cast<ParticleField2D<T,Descriptor>*>(blocks[0]);
    for (pluint iParticle=0; iParticle<particles.size(); ++iParticle) {
        // The function addParticle tests automatically if the particle corresponds to the domain
        //   and needs to be added. In either case, the particle is consumed by the function.
        particleField.addParticle(domain, particles[iParticle]->clone());
    }
}

template<typename T, template<typename U> class Descriptor>
InjectParticlesFunctional2D<T,Descriptor>* InjectParticlesFunctional2D<T,Descriptor>::clone() const {
    return new InjectParticlesFunctional2D<T,Descriptor>(*this);
}

template<typename T, template<typename U> class Descriptor>
BlockDomain::DomainT InjectParticlesFunctional2D<T,Descriptor>::appliesTo() const {
    return BlockDomain::bulk;
}

template<typename T, template<typename U> class Descriptor>
void InjectParticlesFunctional2D<T,Descriptor>::getTypeOfModification (
        std::vector<modif::ModifT>& modified ) const
{
    modified[0] = modif::dynamicVariables;  // Particle field.
}


/* ******** InjectRandomPointParticlesFunctional2D *********************************** */

template<typename T, template<typename U> class Descriptor>
InjectRandomPointParticlesFunctional2D<T,Descriptor>::InjectRandomPointParticlesFunctional2D (
        plint tag_, T probabilityPerCell_ )
    : tag(tag_),
      probabilityPerCell(probabilityPerCell_)
{ }

template<typename T, template<typename U> class Descriptor>
void InjectRandomPointParticlesFunctional2D<T,Descriptor>::processGenericBlocks (
        Box2D domain, std::vector<AtomicBlock2D*> blocks )
{
    PLB_PRECONDITION( blocks.size()==1 );
    ParticleField2D<T,Descriptor>& particleField = *dynamic_cast<ParticleField2D<T,Descriptor>*>(blocks[0]);
    for (plint iX=domain.x0; iX<=domain.x1; ++iX) {
        for (plint iY=domain.y0; iY<=domain.y1; ++iY) {
            T randNumber = (T)rand() / (T)RAND_MAX;
            if(randNumber<probabilityPerCell) {
                T randX = (T)rand() / (T)RAND_MAX - (T)1;
                T randY = (T)rand() / (T)RAND_MAX - (T)1;
                Array<T,2> position (
                        particleField.getLocation().x + iX + randX,
                        particleField.getLocation().y + iY + randY );
                Array<T,2> velocity; velocity.resetToZero();
                particleField.addParticle(
                        domain,
                        new PointParticle2D<T,Descriptor>(tag, position, velocity) );
            }
        }
    }
}

template<typename T, template<typename U> class Descriptor>
InjectRandomPointParticlesFunctional2D<T,Descriptor>* InjectRandomPointParticlesFunctional2D<T,Descriptor>::clone() const {
    return new InjectRandomPointParticlesFunctional2D<T,Descriptor>(*this);
}

template<typename T, template<typename U> class Descriptor>
BlockDomain::DomainT InjectRandomPointParticlesFunctional2D<T,Descriptor>::appliesTo() const {
    return BlockDomain::bulk;
}

template<typename T, template<typename U> class Descriptor>
void InjectRandomPointParticlesFunctional2D<T,Descriptor>::getTypeOfModification (
        std::vector<modif::ModifT>& modified ) const
{
    modified[0] = modif::dynamicVariables;  // Particle field.
}


/* ******** AnalyticalInjectRandomPointParticlesFunctional2D *********************************** */

template<typename T, template<typename U> class Descriptor, class DomainFunctional>
AnalyticalInjectRandomPointParticlesFunctional2D<T,Descriptor,DomainFunctional>::AnalyticalInjectRandomPointParticlesFunctional2D (
        plint tag_, T probabilityPerCell_, DomainFunctional functional_ )
    : tag(tag_),
      probabilityPerCell(probabilityPerCell_),
      functional(functional_)
{ }

template<typename T, template<typename U> class Descriptor, class DomainFunctional>
void AnalyticalInjectRandomPointParticlesFunctional2D<T,Descriptor,DomainFunctional>::processGenericBlocks (
        Box2D domain, std::vector<AtomicBlock2D*> blocks )
{
    PLB_PRECONDITION( blocks.size()==1 );
    ParticleField2D<T,Descriptor>& particleField = *dynamic_cast<ParticleField2D<T,Descriptor>*>(blocks[0]);
    for (plint iX=domain.x0; iX<=domain.x1; ++iX) {
        for (plint iY=domain.y0; iY<=domain.y1; ++iY) {
            T randNumber = (T)rand() / (T)RAND_MAX;
            if(randNumber<probabilityPerCell) {
                T randX = (T)rand() / (T)RAND_MAX - (T)1;
                T randY = (T)rand() / (T)RAND_MAX - (T)1;
                Array<T,2> position (
                        particleField.getLocation().x + iX + randX,
                        particleField.getLocation().y + iY + randY );
                if (functional(position)) {
                    Array<T,2> velocity; velocity.resetToZero();
                    particleField.addParticle(
                            domain,
                            new PointParticle2D<T,Descriptor>(tag, position, velocity) );
                }
            }
        }
    }
}

template<typename T, template<typename U> class Descriptor, class DomainFunctional>
AnalyticalInjectRandomPointParticlesFunctional2D<T,Descriptor,DomainFunctional>*
    AnalyticalInjectRandomPointParticlesFunctional2D<T,Descriptor,DomainFunctional>::clone() const {
    return new AnalyticalInjectRandomPointParticlesFunctional2D<T,Descriptor,DomainFunctional>(*this);
}

template<typename T, template<typename U> class Descriptor, class DomainFunctional>
BlockDomain::DomainT AnalyticalInjectRandomPointParticlesFunctional2D<T,Descriptor,DomainFunctional>::appliesTo() const {
    return BlockDomain::bulk;
}

template<typename T, template<typename U> class Descriptor, class DomainFunctional>
void AnalyticalInjectRandomPointParticlesFunctional2D<T,Descriptor,DomainFunctional>::getTypeOfModification (
        std::vector<modif::ModifT>& modified ) const
{
    modified[0] = modif::dynamicVariables;  // Particle field.
}


/* ******** AbsorbParticlesFunctional2D *********************************** */

template<typename T, template<typename U> class Descriptor>
void AbsorbParticlesFunctional2D<T,Descriptor>::processGenericBlocks (
        Box2D domain, std::vector<AtomicBlock2D*> blocks )
{
    PLB_PRECONDITION( blocks.size()==1 );
    ParticleField2D<T,Descriptor>& particleField =
        *dynamic_cast<ParticleField2D<T,Descriptor>*>(blocks[0]);
    particleField.removeParticles(domain);
}

template<typename T, template<typename U> class Descriptor>
AbsorbParticlesFunctional2D<T,Descriptor>* AbsorbParticlesFunctional2D<T,Descriptor>::clone() const {
    return new AbsorbParticlesFunctional2D<T,Descriptor>(*this);
}

template<typename T, template<typename U> class Descriptor>
BlockDomain::DomainT AbsorbParticlesFunctional2D<T,Descriptor>::appliesTo() const {
    return BlockDomain::bulk;
}

template<typename T, template<typename U> class Descriptor>
void AbsorbParticlesFunctional2D<T,Descriptor>::getTypeOfModification (
        std::vector<modif::ModifT>& modified ) const
{
    modified[0] = modif::dynamicVariables;  // Particle field.
}


/* ******** FluidToParticleCoupling2D *********************************** */

template<typename T, template<typename U> class Descriptor>
void FluidToParticleCoupling2D<T,Descriptor>::processGenericBlocks (
        Box2D domain, std::vector<AtomicBlock2D*> blocks )
{
    PLB_PRECONDITION( blocks.size()==2 );
    ParticleField2D<T,Descriptor>& particleField =
        *dynamic_cast<ParticleField2D<T,Descriptor>*>(blocks[0]);
    BlockLattice2D<T,Descriptor>& fluid =
        *dynamic_cast<BlockLattice2D<T,Descriptor>*>(blocks[1]);
    particleField.fluidToParticleCoupling(domain, fluid);
}

template<typename T, template<typename U> class Descriptor>
FluidToParticleCoupling2D<T,Descriptor>* FluidToParticleCoupling2D<T,Descriptor>::clone() const {
    return new FluidToParticleCoupling2D<T,Descriptor>(*this);
}

template<typename T, template<typename U> class Descriptor>
BlockDomain::DomainT FluidToParticleCoupling2D<T,Descriptor>::appliesTo() const {
    return BlockDomain::bulk; // Important: must be bulk-only,
    // because non-local access to fluid is made.
}

template<typename T, template<typename U> class Descriptor>
void FluidToParticleCoupling2D<T,Descriptor>::getTypeOfModification (
        std::vector<modif::ModifT>& modified ) const
{
    modified[0] = modif::dynamicVariables; // Particle field.
    modified[1] = modif::nothing;  // Fluid.
}


/* ******** AdvanceParticlesFunctional2D *********************************** */

template<typename T, template<typename U> class Descriptor>
AdvanceParticlesFunctional2D<T,Descriptor>::AdvanceParticlesFunctional2D (
        T cutOffValue_ )
  : cutOffValue(cutOffValue_)
{ }

template<typename T, template<typename U> class Descriptor>
void AdvanceParticlesFunctional2D<T,Descriptor>::processGenericBlocks (
        Box2D domain, std::vector<AtomicBlock2D*> blocks )
{
    PLB_PRECONDITION( blocks.size()==1 );
    ParticleField2D<T,Descriptor>& particleField =
        *dynamic_cast<ParticleField2D<T,Descriptor>*>(blocks[0]);
    particleField.advanceParticles(domain, cutOffValue);
}

template<typename T, template<typename U> class Descriptor>
AdvanceParticlesFunctional2D<T,Descriptor>* AdvanceParticlesFunctional2D<T,Descriptor>::clone() const {
    return new AdvanceParticlesFunctional2D<T,Descriptor>(*this);
}

template<typename T, template<typename U> class Descriptor>
BlockDomain::DomainT AdvanceParticlesFunctional2D<T,Descriptor>::appliesTo() const {
    return BlockDomain::bulkAndEnvelope;  // Important: access envelope as well,
                                          // because particles are streamed from the
                                          // envelope into the bulk.
}

template<typename T, template<typename U> class Descriptor>
void AdvanceParticlesFunctional2D<T,Descriptor>::getTypeOfModification (
        std::vector<modif::ModifT>& modified ) const
{
    modified[0] = modif::dynamicVariables;  // Particle field.
}


/* ******** CountAndAccumulateParticles2D *********************************** */

template<typename T, template<typename U> class Descriptor>
void CountAndAccumulateParticles2D<T,Descriptor>::processGenericBlocks (
        Box2D domain, std::vector<AtomicBlock2D*> blocks )
{
    PLB_PRECONDITION( blocks.size()==2 );
    ParticleField2D<T,Descriptor>& particleField =
        *dynamic_cast<ParticleField2D<T,Descriptor>*>(blocks[0]);
    ScalarField2D<plint>& numParticlefield =
        *dynamic_cast<ScalarField2D<plint>*>(blocks[1]);
    Dot2D offset = computeRelativeDisplacement(particleField, numParticlefield);
    std::vector<Particle2D<T,Descriptor>*> particles;
    for (plint iX=domain.x0; iX<=domain.x1; ++iX) {
        for (plint iY=domain.y0; iY<=domain.y1; ++iY) {
            particleField.findParticles(Box2D(iX,iX,iY,iY), particles);
            numParticlefield.get(iX+offset.x,iY+offset.y) += particles.size();
        }
    }
}

template<typename T, template<typename U> class Descriptor>
CountAndAccumulateParticles2D<T,Descriptor>* CountAndAccumulateParticles2D<T,Descriptor>::clone() const {
    return new CountAndAccumulateParticles2D<T,Descriptor>(*this);
}

template<typename T, template<typename U> class Descriptor>
BlockDomain::DomainT CountAndAccumulateParticles2D<T,Descriptor>::appliesTo() const {
    return BlockDomain::bulk;
}

template<typename T, template<typename U> class Descriptor>
void CountAndAccumulateParticles2D<T,Descriptor>::getTypeOfModification (
        std::vector<modif::ModifT>& modified ) const
{
    modified[0] = modif::nothing;          // Particle field.
    modified[1] = modif::staticVariables;  // Scalar field.
}

/* ******** CountAndAccumulateTaggedParticles2D *********************************** */

template<typename T, template<typename U> class Descriptor>
CountAndAccumulateTaggedParticles2D<T,Descriptor>::CountAndAccumulateTaggedParticles2D(plint tag_)
    : tag(tag_)
{ }

template<typename T, template<typename U> class Descriptor>
void CountAndAccumulateTaggedParticles2D<T,Descriptor>::processGenericBlocks (
        Box2D domain, std::vector<AtomicBlock2D*> blocks )
{
    PLB_PRECONDITION( blocks.size()==2 );
    ParticleField2D<T,Descriptor>& particleField =
        *dynamic_cast<ParticleField2D<T,Descriptor>*>(blocks[0]);
    ScalarField2D<plint>& numParticlefield =
        *dynamic_cast<ScalarField2D<plint>*>(blocks[1]);
    Dot2D offset = computeRelativeDisplacement(particleField, numParticlefield);
    std::vector<Particle2D<T,Descriptor>*> particles;
    for (plint iX=domain.x0; iX<=domain.x1; ++iX) {
        for (plint iY=domain.y0; iY<=domain.y1; ++iY) {
            particleField.findParticles(Box2D(iX,iX,iY,iY), particles);
            for (pluint iParticle=0; iParticle<particles.size(); ++iParticle) {
                if (particles[iParticle]->getTag() == tag) {
                    ++numParticlefield.get(iX+offset.x,iY+offset.y);
                }
            }
        }
    }
}

template<typename T, template<typename U> class Descriptor>
CountAndAccumulateTaggedParticles2D<T,Descriptor>* CountAndAccumulateTaggedParticles2D<T,Descriptor>::clone() const {
    return new CountAndAccumulateTaggedParticles2D<T,Descriptor>(*this);
}

template<typename T, template<typename U> class Descriptor>
BlockDomain::DomainT CountAndAccumulateTaggedParticles2D<T,Descriptor>::appliesTo() const {
    return BlockDomain::bulk;
}

template<typename T, template<typename U> class Descriptor>
void CountAndAccumulateTaggedParticles2D<T,Descriptor>::getTypeOfModification (
        std::vector<modif::ModifT>& modified ) const
{
    modified[0] = modif::nothing;          // Particle field.
    modified[1] = modif::staticVariables;  // Scalar field.
}

}  // namespace plb

#endif  // PARTICLE_PROCESSING_FUNCTIONAL_2D_HH
