/* This file is part of the Palabos library.
 *
 * Copyright (C) 2011 FlowKit Sarl
 * Avenue de Chailly 23
 * 1012 Lausanne, Switzerland
 * E-mail contact: contact@flowkit.com
 *
 * The most recent release of Palabos can be downloaded at 
 * <http://www.palabos.org/>
 *
 * The library Palabos is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * The library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef PARTICLE_PROCESSING_FUNCTIONAL_3D_HH
#define PARTICLE_PROCESSING_FUNCTIONAL_3D_HH

#include "particles/particleProcessingFunctional3D.h"
#include "particles/particleField3D.h"
#include "dataProcessors/metaStuffFunctional3D.h"
#include "core/plbDebug.h"
#include "core/blockStatistics.h"
#include "atomicBlock/atomicBlock3D.h"
#include "atomicBlock/blockLattice3D.h"
#include <algorithm>

namespace plb {

/* ******** CountParticlesFunctional3D *********************************** */

template<typename T, template<typename U> class Descriptor>
CountParticlesFunctional3D<T,Descriptor>::CountParticlesFunctional3D()
    : numParticlesId(this->getStatistics().subscribeIntSum())
{ }

template<typename T, template<typename U> class Descriptor>
void CountParticlesFunctional3D<T,Descriptor>::processGenericBlocks (
        Box3D domain, std::vector<AtomicBlock3D*> blocks )
{
    PLB_PRECONDITION( blocks.size()==1 );
    ParticleField3D<T,Descriptor>& particleField
        = *dynamic_cast<ParticleField3D<T,Descriptor>*>(blocks[0]);
    std::vector<Particle3D<T,Descriptor>*> particles;
    particleField.findParticles(domain, particles);
    this->getStatistics().gatherIntSum(numParticlesId, (plint)particles.size());
}

template<typename T, template<typename U> class Descriptor>
CountParticlesFunctional3D<T,Descriptor>* CountParticlesFunctional3D<T,Descriptor>::clone() const {
    return new CountParticlesFunctional3D<T,Descriptor>(*this);
}

template<typename T, template<typename U> class Descriptor>
void CountParticlesFunctional3D<T,Descriptor>::getTypeOfModification(std::vector<modif::ModifT>& modified) const {
    modified[0] = modif::nothing;
}

template<typename T, template<typename U> class Descriptor>
plint CountParticlesFunctional3D<T,Descriptor>::getNumParticles() const {
    return this->getStatistics().getIntSum(numParticlesId);
}


/* ******** AverageParticleVelocityFunctional3D *********************************** */

template<typename T, template<typename U> class Descriptor>
AverageParticleVelocityFunctional3D<T,Descriptor>::AverageParticleVelocityFunctional3D()
    : averageVelocityId (
            Array<T,3>( this->getStatistics().subscribeAverage(),
                        this->getStatistics().subscribeAverage(),
                        this->getStatistics().subscribeAverage()) )
{ }

template<typename T, template<typename U> class Descriptor>
void AverageParticleVelocityFunctional3D<T,Descriptor>::processGenericBlocks (
        Box3D domain, std::vector<AtomicBlock3D*> blocks )
{
    PLB_PRECONDITION( blocks.size()==1 );
    ParticleField3D<T,Descriptor>& particleField = *dynamic_cast<ParticleField3D<T,Descriptor>*>(blocks[0]);
    std::vector<Particle3D<T,Descriptor>*> particles;
    particleField.findParticles(domain, particles);
    for (pluint iParticle=0; iParticle<particles.size(); ++iParticle) {
        Array<T,3> velocity;
        particles[iParticle]->getVelocity(velocity);
        this->getStatistics().gatherAverage(averageVelocityId[0], velocity[0]);
        this->getStatistics().gatherAverage(averageVelocityId[1], velocity[1]);
        this->getStatistics().gatherAverage(averageVelocityId[2], velocity[2]);
        this->getStatistics().incrementStats();
    }
}

template<typename T, template<typename U> class Descriptor>
AverageParticleVelocityFunctional3D<T,Descriptor>* AverageParticleVelocityFunctional3D<T,Descriptor>::clone() const {
    return new AverageParticleVelocityFunctional3D<T,Descriptor>(*this);
}

template<typename T, template<typename U> class Descriptor>
void AverageParticleVelocityFunctional3D<T,Descriptor>::getTypeOfModification(std::vector<modif::ModifT>& modified) const {
    modified[0] = modif::nothing;
}

template<typename T, template<typename U> class Descriptor>
Array<T,3> AverageParticleVelocityFunctional3D<T,Descriptor>::getAverageParticleVelocity() const {
    return Array<T,3> (
        this->getStatistics().getAverage(averageVelocityId[0]),
        this->getStatistics().getAverage(averageVelocityId[1]),
        this->getStatistics().getAverage(averageVelocityId[2]) );
}


/* ******** InjectParticlesFunctional3D *********************************** */

template<typename T, template<typename U> class Descriptor>
InjectParticlesFunctional3D<T,Descriptor>::InjectParticlesFunctional3D(std::vector<Particle3D<T,Descriptor>*>& particles_)
    : particles(particles_)
{ }

template<typename T, template<typename U> class Descriptor>
void InjectParticlesFunctional3D<T,Descriptor>::processGenericBlocks (
        Box3D domain, std::vector<AtomicBlock3D*> blocks )
{
    PLB_PRECONDITION( blocks.size()==1 );
    ParticleField3D<T,Descriptor>& particleField = *dynamic_cast<ParticleField3D<T,Descriptor>*>(blocks[0]);
    for (pluint iParticle=0; iParticle<particles.size(); ++iParticle) {
        // The function addParticle tests automatically if the particle corresponds to the domain
        //   and needs to be added. In either case, the particle is consumed by the function.
        particleField.addParticle(domain, particles[iParticle]->clone());
    }
}

template<typename T, template<typename U> class Descriptor>
InjectParticlesFunctional3D<T,Descriptor>* InjectParticlesFunctional3D<T,Descriptor>::clone() const {
    return new InjectParticlesFunctional3D<T,Descriptor>(*this);
}

template<typename T, template<typename U> class Descriptor>
void InjectParticlesFunctional3D<T,Descriptor>::getTypeOfModification (
        std::vector<modif::ModifT>& modified ) const
{
    modified[0] = modif::dynamicVariables;  // Particle field.
}


/* ******** InjectRandomParticlesFunctional3D *********************************** */

template<typename T, template<typename U> class Descriptor>
InjectRandomParticlesFunctional3D<T,Descriptor>::InjectRandomParticlesFunctional3D (
        Particle3D<T,Descriptor>* particleTemplate_, T probabilityPerCell_ )
    : particleTemplate(particleTemplate_),
      probabilityPerCell(probabilityPerCell_)
{ }

template<typename T, template<typename U> class Descriptor>
InjectRandomParticlesFunctional3D<T,Descriptor>::InjectRandomParticlesFunctional3D (
        InjectRandomParticlesFunctional3D<T,Descriptor> const& rhs)
    : particleTemplate(rhs.particleTemplate->clone()),
      probabilityPerCell(rhs.probabilityPerCell)
{ }

template<typename T, template<typename U> class Descriptor>
InjectRandomParticlesFunctional3D<T,Descriptor>&
    InjectRandomParticlesFunctional3D<T,Descriptor>::operator= (
        InjectRandomParticlesFunctional3D<T,Descriptor> const& rhs )
{
    InjectRandomParticlesFunctional3D<T,Descriptor>(rhs).swap(*this);
}

template<typename T, template<typename U> class Descriptor>
void InjectRandomParticlesFunctional3D<T,Descriptor>::swap (
        InjectRandomParticlesFunctional3D<T,Descriptor>& rhs )
{
    std::swap(particleTemplate, rhs.particleTemplate);
    std::swap(probabilityPerCell, rhs.probabilityPerCell);
}

template<typename T, template<typename U> class Descriptor>
InjectRandomParticlesFunctional3D<T,Descriptor>::~InjectRandomParticlesFunctional3D()
{
    delete particleTemplate;
}

template<typename T, template<typename U> class Descriptor>
void InjectRandomParticlesFunctional3D<T,Descriptor>::processGenericBlocks (
        Box3D domain, std::vector<AtomicBlock3D*> blocks )
{
    PLB_PRECONDITION( blocks.size()==1 );
    ParticleField3D<T,Descriptor>& particleField = *dynamic_cast<ParticleField3D<T,Descriptor>*>(blocks[0]);
    for (plint iX=domain.x0; iX<=domain.x1; ++iX) {
        for (plint iY=domain.y0; iY<=domain.y1; ++iY) {
            for (plint iZ=domain.z0; iZ<=domain.z1; ++iZ) {
                T randNumber = (T)rand() / (T)RAND_MAX;
                if(randNumber<probabilityPerCell) {
                    T randX = (T)rand() / (T)RAND_MAX - (T)1;
                    T randY = (T)rand() / (T)RAND_MAX - (T)1;
                    T randZ = (T)rand() / (T)RAND_MAX - (T)1;
                    Particle3D<T,Descriptor>* newparticle = particleTemplate->clone();
                    newparticle->getPosition() =
                            Array<T,3> (
                                    particleField.getLocation().x + iX + randX,
                                    particleField.getLocation().y + iY + randY,
                                    particleField.getLocation().z + iZ + randZ );
                    particleField.addParticle(domain, newparticle);
                }
            }
        }
    }
}

template<typename T, template<typename U> class Descriptor>
InjectRandomParticlesFunctional3D<T,Descriptor>* InjectRandomParticlesFunctional3D<T,Descriptor>::clone() const {
    return new InjectRandomParticlesFunctional3D<T,Descriptor>(*this);
}

template<typename T, template<typename U> class Descriptor>
void InjectRandomParticlesFunctional3D<T,Descriptor>::getTypeOfModification (
        std::vector<modif::ModifT>& modified ) const
{
    modified[0] = modif::dynamicVariables;  // Particle field.
}


/* ******** AnalyticalInjectRandomParticlesFunctional3D *********************************** */

template<typename T, template<typename U> class Descriptor, class DomainFunctional>
AnalyticalInjectRandomParticlesFunctional3D<T,Descriptor,DomainFunctional>::
    AnalyticalInjectRandomParticlesFunctional3D (
         Particle3D<T,Descriptor>* particleTemplate_, T probabilityPerCell_, DomainFunctional functional_ )
    : particleTemplate(particleTemplate_),
      probabilityPerCell(probabilityPerCell_),
      functional(functional_)
{ }

template<typename T, template<typename U> class Descriptor, class DomainFunctional>
AnalyticalInjectRandomParticlesFunctional3D<T,Descriptor,DomainFunctional>::AnalyticalInjectRandomParticlesFunctional3D (
        AnalyticalInjectRandomParticlesFunctional3D<T,Descriptor,DomainFunctional> const& rhs)
    : particleTemplate(rhs.particleTemplate->clone()),
      probabilityPerCell(rhs.probabilityPerCell),
      functional(rhs.functional)
{ }

template<typename T, template<typename U> class Descriptor, class DomainFunctional>
AnalyticalInjectRandomParticlesFunctional3D<T,Descriptor,DomainFunctional>&
    AnalyticalInjectRandomParticlesFunctional3D<T,Descriptor,DomainFunctional>::operator= (
        AnalyticalInjectRandomParticlesFunctional3D<T,Descriptor,DomainFunctional> const& rhs )
{
    AnalyticalInjectRandomParticlesFunctional3D<T,Descriptor,DomainFunctional>(rhs).swap(*this);
}

template<typename T, template<typename U> class Descriptor, class DomainFunctional>
void AnalyticalInjectRandomParticlesFunctional3D<T,Descriptor,DomainFunctional>::swap (
        AnalyticalInjectRandomParticlesFunctional3D<T,Descriptor,DomainFunctional>& rhs )
{
    std::swap(particleTemplate, rhs.particleTemplate);
    std::swap(probabilityPerCell, rhs.probabilityPerCell);
    std::swap(functional, rhs.functional);
}

template<typename T, template<typename U> class Descriptor, class DomainFunctional>
AnalyticalInjectRandomParticlesFunctional3D<T,Descriptor,DomainFunctional>::~AnalyticalInjectRandomParticlesFunctional3D()
{
    delete particleTemplate;
}

template<typename T, template<typename U> class Descriptor, class DomainFunctional>
void AnalyticalInjectRandomParticlesFunctional3D<T,Descriptor,DomainFunctional>::processGenericBlocks (
        Box3D domain, std::vector<AtomicBlock3D*> blocks )
{
    PLB_PRECONDITION( blocks.size()==1 );
    ParticleField3D<T,Descriptor>& particleField = *dynamic_cast<ParticleField3D<T,Descriptor>*>(blocks[0]);
    for (plint iX=domain.x0; iX<=domain.x1; ++iX) {
        for (plint iY=domain.y0; iY<=domain.y1; ++iY) {
            for (plint iZ=domain.z0; iZ<=domain.z1; ++iZ) {
                T randNumber = (T)rand() / (T)RAND_MAX;
                if(randNumber<probabilityPerCell) {
                    T randX = (T)rand() / (T)RAND_MAX - (T)1;
                    T randY = (T)rand() / (T)RAND_MAX - (T)1;
                    T randZ = (T)rand() / (T)RAND_MAX - (T)1;
                    Array<T,3> position (
                            particleField.getLocation().x + iX + randX,
                            particleField.getLocation().y + iY + randY,
                            particleField.getLocation().z + iZ + randZ );
                    if (functional(position)) {
                        Particle3D<T,Descriptor>* newparticle = particleTemplate->clone();
                        newparticle->getPosition() = position;
                        particleField.addParticle(domain, newparticle);
                    }
                }
            }
        }
    }
}

template<typename T, template<typename U> class Descriptor, class DomainFunctional>
AnalyticalInjectRandomParticlesFunctional3D<T,Descriptor,DomainFunctional>*
    AnalyticalInjectRandomParticlesFunctional3D<T,Descriptor,DomainFunctional>::clone() const {
    return new AnalyticalInjectRandomParticlesFunctional3D<T,Descriptor,DomainFunctional>(*this);
}

template<typename T, template<typename U> class Descriptor, class DomainFunctional>
void AnalyticalInjectRandomParticlesFunctional3D<T,Descriptor,DomainFunctional>::getTypeOfModification (
        std::vector<modif::ModifT>& modified ) const
{
    modified[0] = modif::dynamicVariables;  // Particle field.
}


/* ******** MaskedInjectRandomParticlesFunctional3D *********************************** */

template<typename T, template<typename U> class Descriptor, class DomainFunctional>
MaskedInjectRandomParticlesFunctional3D<T,Descriptor,DomainFunctional>::MaskedInjectRandomParticlesFunctional3D (
        Particle3D<T,Descriptor>* particleTemplate_, T probabilityPerCell_, DomainFunctional functional_, int flag_ )
    : particleTemplate(particleTemplate_),
      probabilityPerCell(probabilityPerCell_),
      functional(functional_),
      flag(flag_)
{ }

template<typename T, template<typename U> class Descriptor, class DomainFunctional>
MaskedInjectRandomParticlesFunctional3D<T,Descriptor,DomainFunctional>::MaskedInjectRandomParticlesFunctional3D (
        MaskedInjectRandomParticlesFunctional3D<T,Descriptor,DomainFunctional> const& rhs)
    : particleTemplate(rhs.particleTemplate->clone()),
      probabilityPerCell(rhs.probabilityPerCell),
      functional(rhs.functional),
      flag(rhs.flag)
{ }

template<typename T, template<typename U> class Descriptor, class DomainFunctional>
MaskedInjectRandomParticlesFunctional3D<T,Descriptor,DomainFunctional>&
    MaskedInjectRandomParticlesFunctional3D<T,Descriptor,DomainFunctional>::operator= (
        MaskedInjectRandomParticlesFunctional3D<T,Descriptor,DomainFunctional> const& rhs )
{
    MaskedInjectRandomParticlesFunctional3D<T,Descriptor,DomainFunctional>(rhs).swap(*this);
}

template<typename T, template<typename U> class Descriptor, class DomainFunctional>
void MaskedInjectRandomParticlesFunctional3D<T,Descriptor,DomainFunctional>::swap (
        MaskedInjectRandomParticlesFunctional3D<T,Descriptor,DomainFunctional>& rhs )
{
    std::swap(particleTemplate, rhs.particleTemplate);
    std::swap(probabilityPerCell, rhs.probabilityPerCell);
    std::swap(functional, rhs.functional);
    std::swap(flag, rhs.flag);
}

template<typename T, template<typename U> class Descriptor, class DomainFunctional>
MaskedInjectRandomParticlesFunctional3D<T,Descriptor,DomainFunctional>::~MaskedInjectRandomParticlesFunctional3D()
{
    delete particleTemplate;
}

template<typename T, template<typename U> class Descriptor, class DomainFunctional>
void MaskedInjectRandomParticlesFunctional3D<T,Descriptor,DomainFunctional>::processGenericBlocks (
        Box3D domain, std::vector<AtomicBlock3D*> blocks )
{
    PLB_PRECONDITION( blocks.size()==2 );
    ParticleField3D<T,Descriptor>& particleField = *dynamic_cast<ParticleField3D<T,Descriptor>*>(blocks[0]);
    ScalarField3D<int>& flagMatrix = *dynamic_cast<ScalarField3D<int>*>(blocks[1]);
    for (plint iX=domain.x0; iX<=domain.x1; ++iX) {
        for (plint iY=domain.y0; iY<=domain.y1; ++iY) {
            for (plint iZ=domain.z0; iZ<=domain.z1; ++iZ) {
                T randNumber = (T)rand() / (T)RAND_MAX;
                if(randNumber<probabilityPerCell) {
                    T randX = (T)rand() / (T)RAND_MAX - (T)1;
                    T randY = (T)rand() / (T)RAND_MAX - (T)1;
                    T randZ = (T)rand() / (T)RAND_MAX - (T)1;
                    Array<T,3> position (
                            particleField.getLocation().x + iX + randX,
                            particleField.getLocation().y + iY + randY,
                            particleField.getLocation().z + iZ + randZ );
                    if (functional(position) && flagMatrix.get(iX,iY,iZ)==flag) {
                        Particle3D<T,Descriptor>* newparticle = particleTemplate->clone();
                        newparticle->getPosition() = position;
                        particleField.addParticle(domain, newparticle);
                    }
                }
            }
        }
    }
}

template<typename T, template<typename U> class Descriptor, class DomainFunctional>
MaskedInjectRandomParticlesFunctional3D<T,Descriptor,DomainFunctional>*
    MaskedInjectRandomParticlesFunctional3D<T,Descriptor,DomainFunctional>::clone() const
{
    return new MaskedInjectRandomParticlesFunctional3D<T,Descriptor,DomainFunctional>(*this);
}

template<typename T, template<typename U> class Descriptor, class DomainFunctional>
void MaskedInjectRandomParticlesFunctional3D<T,Descriptor,DomainFunctional>::getTypeOfModification (
        std::vector<modif::ModifT>& modified ) const
{
    modified[0] = modif::dynamicVariables;  // Particle field.
    modified[1] = modif::nothing;  // Mask.
}


/* ******** AbsorbParticlesFunctional3D *********************************** */

template<typename T, template<typename U> class Descriptor>
void AbsorbParticlesFunctional3D<T,Descriptor>::processGenericBlocks (
        Box3D domain, std::vector<AtomicBlock3D*> blocks )
{
    PLB_PRECONDITION( blocks.size()==1 );
    ParticleField3D<T,Descriptor>& particleField =
        *dynamic_cast<ParticleField3D<T,Descriptor>*>(blocks[0]);
    particleField.removeParticles(domain);
}

template<typename T, template<typename U> class Descriptor>
AbsorbParticlesFunctional3D<T,Descriptor>* AbsorbParticlesFunctional3D<T,Descriptor>::clone() const {
    return new AbsorbParticlesFunctional3D<T,Descriptor>(*this);
}

template<typename T, template<typename U> class Descriptor>
void AbsorbParticlesFunctional3D<T,Descriptor>::getTypeOfModification (
        std::vector<modif::ModifT>& modified ) const
{
    modified[0] = modif::dynamicVariables;  // Particle field.
}


/* ******** FluidToParticleCoupling3D *********************************** */

template<typename T, template<typename U> class Descriptor>
FluidToParticleCoupling3D<T,Descriptor>::FluidToParticleCoupling3D(T scaling_)
    : scaling(scaling_)
{ }

template<typename T, template<typename U> class Descriptor>
void FluidToParticleCoupling3D<T,Descriptor>::processGenericBlocks (
        Box3D domain, std::vector<AtomicBlock3D*> blocks )
{
    PLB_PRECONDITION( blocks.size()==2 );
    ParticleField3D<T,Descriptor>& particleField =
        *dynamic_cast<ParticleField3D<T,Descriptor>*>(blocks[0]);
    BlockLattice3D<T,Descriptor>& fluid =
        *dynamic_cast<BlockLattice3D<T,Descriptor>*>(blocks[1]);
    particleField.fluidToParticleCoupling(domain, fluid, scaling);
}

template<typename T, template<typename U> class Descriptor>
FluidToParticleCoupling3D<T,Descriptor>* FluidToParticleCoupling3D<T,Descriptor>::clone() const {
    return new FluidToParticleCoupling3D<T,Descriptor>(*this);
}

template<typename T, template<typename U> class Descriptor>
void FluidToParticleCoupling3D<T,Descriptor>::getTypeOfModification (
        std::vector<modif::ModifT>& modified ) const
{
    modified[0] = modif::dynamicVariables; // Particle field.
    modified[1] = modif::nothing;  // Fluid.
}


/* ******** VelocityToParticleCoupling3D *********************************** */

template<typename T, template<typename U> class Descriptor>
VelocityToParticleCoupling3D<T,Descriptor>::VelocityToParticleCoupling3D(T scaling_)
    : scaling(scaling_)
{ }

template<typename T, template<typename U> class Descriptor>
void VelocityToParticleCoupling3D<T,Descriptor>::processGenericBlocks (
        Box3D domain, std::vector<AtomicBlock3D*> blocks )
{
    PLB_PRECONDITION( blocks.size()==2 );
    ParticleField3D<T,Descriptor>* particleFieldPtr =
        dynamic_cast<ParticleField3D<T,Descriptor>*>(blocks[0]);
    PLB_ASSERT( particleFieldPtr );
    ParticleField3D<T,Descriptor>& particleField = *particleFieldPtr;

    TensorField3D<T,3>* velocityPtr =
        dynamic_cast<TensorField3D<T,3>*>(blocks[1]);
    PLB_ASSERT( velocityPtr );
    TensorField3D<T,3>& velocity = *velocityPtr;

    particleField.velocityToParticleCoupling(domain, velocity, scaling);
}

template<typename T, template<typename U> class Descriptor>
VelocityToParticleCoupling3D<T,Descriptor>* VelocityToParticleCoupling3D<T,Descriptor>::clone() const {
    return new VelocityToParticleCoupling3D<T,Descriptor>(*this);
}

template<typename T, template<typename U> class Descriptor>
void VelocityToParticleCoupling3D<T,Descriptor>::getTypeOfModification (
        std::vector<modif::ModifT>& modified ) const
{
    modified[0] = modif::dynamicVariables; // Particle field.
    modified[1] = modif::nothing;  // Fluid.
}


/* ******** AdvanceParticlesFunctional3D *********************************** */

template<typename T, template<typename U> class Descriptor>
AdvanceParticlesFunctional3D<T,Descriptor>::AdvanceParticlesFunctional3D (
        T cutOffValue_ )
  : cutOffValue(cutOffValue_)
{ }

template<typename T, template<typename U> class Descriptor>
void AdvanceParticlesFunctional3D<T,Descriptor>::processGenericBlocks (
        Box3D domain, std::vector<AtomicBlock3D*> blocks )
{
    PLB_PRECONDITION( blocks.size()==1 );
    ParticleField3D<T,Descriptor>& particleField =
        *dynamic_cast<ParticleField3D<T,Descriptor>*>(blocks[0]);
    particleField.advanceParticles(domain, cutOffValue);
}

template<typename T, template<typename U> class Descriptor>
AdvanceParticlesFunctional3D<T,Descriptor>* AdvanceParticlesFunctional3D<T,Descriptor>::clone() const {
    return new AdvanceParticlesFunctional3D<T,Descriptor>(*this);
}

template<typename T, template<typename U> class Descriptor>
BlockDomain::DomainT AdvanceParticlesFunctional3D<T,Descriptor>::appliesTo() const {
    return BlockDomain::bulkAndEnvelope;  // Important: access envelope as well,
                                          // because particles are streamed from the
                                          // envelope into the bulk.
}

template<typename T, template<typename U> class Descriptor>
void AdvanceParticlesFunctional3D<T,Descriptor>::getTypeOfModification (
        std::vector<modif::ModifT>& modified ) const
{
    modified[0] = modif::dynamicVariables;  // Particle field.
}

/* ******** CountAndAccumulateParticles3D *********************************** */

template<typename T, template<typename U> class Descriptor>
void CountAndAccumulateParticles3D<T,Descriptor>::processGenericBlocks (
        Box3D domain, std::vector<AtomicBlock3D*> blocks )
{
    PLB_PRECONDITION( blocks.size()==2 );
    ParticleField3D<T,Descriptor>& particleField =
        *dynamic_cast<ParticleField3D<T,Descriptor>*>(blocks[0]);
    ScalarField3D<plint>& numParticlefield =
        *dynamic_cast<ScalarField3D<plint>*>(blocks[1]);
    Dot3D offset = computeRelativeDisplacement(particleField, numParticlefield);
    std::vector<Particle3D<T,Descriptor>*> particles;
    for (plint iX=domain.x0; iX<=domain.x1; ++iX) {
        for (plint iY=domain.y0; iY<=domain.y1; ++iY) {
            for (plint iZ=domain.z0; iZ<=domain.z1; ++iZ) {
                particleField.findParticles(Box3D(iX,iX,iY,iY,iZ,iZ), particles);
                numParticlefield.get(iX+offset.x,iY+offset.y,iZ+offset.z) += particles.size();
            }
        }
    }
}

template<typename T, template<typename U> class Descriptor>
CountAndAccumulateParticles3D<T,Descriptor>* CountAndAccumulateParticles3D<T,Descriptor>::clone() const {
    return new CountAndAccumulateParticles3D<T,Descriptor>(*this);
}

template<typename T, template<typename U> class Descriptor>
void CountAndAccumulateParticles3D<T,Descriptor>::getTypeOfModification (
        std::vector<modif::ModifT>& modified ) const
{
    modified[0] = modif::nothing;          // Particle field.
    modified[1] = modif::staticVariables;  // Scalar field.
}

/* ******** CountAndAccumulateTaggedParticles3D *********************************** */

template<typename T, template<typename U> class Descriptor>
CountAndAccumulateTaggedParticles3D<T,Descriptor>::CountAndAccumulateTaggedParticles3D(plint tag_)
    : tag(tag_)
{ }

template<typename T, template<typename U> class Descriptor>
void CountAndAccumulateTaggedParticles3D<T,Descriptor>::processGenericBlocks (
        Box3D domain, std::vector<AtomicBlock3D*> blocks )
{
    PLB_PRECONDITION( blocks.size()==2 );
    ParticleField3D<T,Descriptor>& particleField =
        *dynamic_cast<ParticleField3D<T,Descriptor>*>(blocks[0]);
    ScalarField3D<plint>& numParticlefield =
        *dynamic_cast<ScalarField3D<plint>*>(blocks[1]);
    Dot3D offset = computeRelativeDisplacement(particleField, numParticlefield);
    std::vector<Particle3D<T,Descriptor>*> particles;
    for (plint iX=domain.x0; iX<=domain.x1; ++iX) {
        for (plint iY=domain.y0; iY<=domain.y1; ++iY) {
            for (plint iZ=domain.z0; iZ<=domain.z1; ++iZ) {
                particleField.findParticles(Box3D(iX,iX,iY,iY,iZ,iZ), particles);
                for (pluint iParticle=0; iParticle<particles.size(); ++iParticle) {
                    if (particles[iParticle]->getTag() == tag) {
                        ++numParticlefield.get(iX+offset.x,iY+offset.y,iZ+offset.z);
                    }
                }
            }
        }
    }
}

template<typename T, template<typename U> class Descriptor>
CountAndAccumulateTaggedParticles3D<T,Descriptor>* CountAndAccumulateTaggedParticles3D<T,Descriptor>::clone() const {
    return new CountAndAccumulateTaggedParticles3D<T,Descriptor>(*this);
}

template<typename T, template<typename U> class Descriptor>
void CountAndAccumulateTaggedParticles3D<T,Descriptor>::getTypeOfModification (
        std::vector<modif::ModifT>& modified ) const
{
    modified[0] = modif::nothing;          // Particle field.
    modified[1] = modif::staticVariables;  // Scalar field.
}

template< typename T, template<typename U> class Descriptor,
          template<typename T_, template<typename U_> class Descriptor_> class ParticleFieldT >
plint countParticles (
                MultiParticleField3D<ParticleFieldT<T,Descriptor> >& particles, Box3D const& domain )
{
    std::vector<MultiBlock3D*> particleArg;
    particleArg.push_back(&particles);

    CountParticlesFunctional3D<T,Descriptor> functional;
    applyProcessingFunctional(functional, domain, particleArg);
    return functional.getNumParticles();
}

}  // namespace plb

#endif  // PARTICLE_PROCESSING_FUNCTIONAL_3D_HH
