/* This file is part of the Palabos library.
 *
 * Copyright (C) 2011 FlowKit Sarl
 * Avenue de Chailly 23
 * 1012 Lausanne, Switzerland
 * E-mail contact: contact@flowkit.com
 *
 * The most recent release of Palabos can be downloaded at 
 * <http://www.palabos.org/>
 *
 * The library Palabos is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * The library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef VISUAL_PARTICLE_FUNCTIONAL_3D_H
#define VISUAL_PARTICLE_FUNCTIONAL_3D_H

#include "core/globalDefs.h"
#include "atomicBlock/dataProcessingFunctional3D.h"
#include "atomicBlock/atomicContainerBlock3D.h"
#include "offLattice/triangularSurfaceMesh.h"
#include "offLattice/triangleBoundary3D.h"

namespace plb {

template<typename T, template<typename U> class Descriptor, class ParticleT>
class CreateParticleFromVertex3D : public BoxProcessingFunctional3D
{
public:
    CreateParticleFromVertex3D(TriangularSurfaceMesh<T> const& mesh_);
    /// Arguments: [0] Particle-field.
    virtual void processGenericBlocks(Box3D domain, std::vector<AtomicBlock3D*> fields);
    virtual CreateParticleFromVertex3D<T,Descriptor,ParticleT>* clone() const;
    virtual BlockDomain::DomainT appliesTo() const;
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const;
private:
    TriangularSurfaceMesh<T> const& mesh;
};

template<typename T, template<typename U> class Descriptor>
class CloneParticles3D : public BoxProcessingFunctional3D
{
public:
    CloneParticles3D (
            Particle3D<T,Descriptor>* particleTemplate_ );
    CloneParticles3D(CloneParticles3D<T,Descriptor> const& rhs);
    ~CloneParticles3D();
    CloneParticles3D<T,Descriptor>& operator= (
            CloneParticles3D<T,Descriptor> const& rhs );
    void swap(CloneParticles3D<T,Descriptor>& rhs);
    /// Arguments: [0] original particles; [1] cloned particles.
    virtual void processGenericBlocks(Box3D domain, std::vector<AtomicBlock3D*> fields);
    virtual CloneParticles3D<T,Descriptor>* clone() const;
    virtual BlockDomain::DomainT appliesTo() const;
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const;
private:
    Particle3D<T,Descriptor>* particleTemplate;
};

template<typename T, template<typename U> class Descriptor, class ParticleT>
class CopyVertexToParticle3D : public BoxProcessingFunctional3D
{
public:
    CopyVertexToParticle3D(TriangularSurfaceMesh<T> const& mesh_);
    /// Arguments: [0] Particle-field.
    virtual void processGenericBlocks(Box3D domain, std::vector<AtomicBlock3D*> fields);
    virtual CopyVertexToParticle3D<T,Descriptor,ParticleT>* clone() const;
    virtual BlockDomain::DomainT appliesTo() const;
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const;
private:
    TriangularSurfaceMesh<T> const& mesh;
};

template<typename T, template<typename U> class Descriptor>
class CopyParticleToVertex3D : public BoxProcessingFunctional3D
{
public:
    CopyParticleToVertex3D(TriangularSurfaceMesh<T>& mesh_);
    /// Arguments: [0] Particle-field.
    virtual void processGenericBlocks(Box3D domain, std::vector<AtomicBlock3D*> fields);
    virtual CopyParticleToVertex3D<T,Descriptor>* clone() const;
    virtual BlockDomain::DomainT appliesTo() const;
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const;
private:
    TriangularSurfaceMesh<T>& mesh;
};

template< typename T, template<typename U> class Descriptor >
class ComputeParticleForce3D : public BoxProcessingFunctional3D
{
public:
    ComputeParticleForce3D( TriangleBoundary3D<T> const& triangleBoundary_,
                            int flowType_, bool incompressibleModel_ );
    /// Arguments: [0] Particle-field; [1] Fluid; [2] Voxel-Matrix.
    virtual void processGenericBlocks(Box3D domain, std::vector<AtomicBlock3D*> fields);
    virtual ComputeParticleForce3D<T,Descriptor>* clone() const;
    virtual BlockDomain::DomainT appliesTo() const;
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const;
private:
    TriangleBoundary3D<T> const& triangleBoundary;
    int flowType;
    bool incompressibleModel;
};

template<typename T, template<typename U> class Descriptor>
class GetParticleVelocity3D : public BoxProcessingFunctional3D
{
public:
    /// Arguments: [0] Orignal field (point-particles), [1] Cloned field (visual particles).
    virtual void processGenericBlocks(Box3D domain, std::vector<AtomicBlock3D*> fields);
    virtual GetParticleVelocity3D<T,Descriptor>* clone() const;
    virtual BlockDomain::DomainT appliesTo() const;
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const;
};

template<typename T, template<typename U> class Descriptor>
class CreateParticleMap3D : public BoxProcessingFunctional3D
{
public:
    /// Arguments: [0] particle field; [1] x-matrix of map; [2] y-matrix of map; [3] z-matrix of map.
    virtual void processGenericBlocks(Box3D domain, std::vector<AtomicBlock3D*> fields);
    virtual CreateParticleMap3D<T,Descriptor>* clone() const;
    virtual void getTypeOfModification(std::vector<modif::ModifT>& modified) const;
};

}  // namespace plb

#endif  // VISUAL_PARTICLE_FUNCTIONAL_3D_H
