/* This file is part of the Palabos library.
 *
 * Copyright (C) 2011 FlowKit Sarl
 * Avenue de Chailly 23
 * 1012 Lausanne, Switzerland
 * E-mail contact: contact@flowkit.com
 *
 * The most recent release of Palabos can be downloaded at 
 * <http://www.palabos.org/>
 *
 * The library Palabos is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * The library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef VISUAL_PARTICLE_FUNCTIONAL_3D_HH
#define VISUAL_PARTICLE_FUNCTIONAL_3D_HH

#include "particles/visualParticleFunctional3D.h"
#include "particles/particleField3D.h"
#include "dataProcessors/metaStuffFunctional3D.h"
#include "core/plbDebug.h"
#include "core/blockStatistics.h"
#include "atomicBlock/atomicBlock3D.h"
#include "atomicBlock/blockLattice3D.h"
#include "latticeBoltzmann/geometricOperationTemplates.h"
#include "finiteDifference/interpolations2D.h"

namespace plb {

/* ******** CreateParticleFromVertex3D *********************************** */

template<typename T, template<typename U> class Descriptor, class ParticleT>
CreateParticleFromVertex3D<T,Descriptor,ParticleT>::CreateParticleFromVertex3D (
        TriangularSurfaceMesh<T> const& mesh_)
    : mesh(mesh_)
{ }

template<typename T, template<typename U> class Descriptor, class ParticleT>
void CreateParticleFromVertex3D<T,Descriptor,ParticleT>::processGenericBlocks (
        Box3D domain, std::vector<AtomicBlock3D*> blocks )
{
    PLB_PRECONDITION( blocks.size()==1 );
    ParticleField3D<T,Descriptor>& particleField =
        *dynamic_cast<ParticleField3D<T,Descriptor>*>(blocks[0]);

    for (plint iVertex=0; iVertex<mesh.getNumVertices(); ++iVertex) {
        Array<T,3> vertex(mesh.getVertex(iVertex));
        ParticleT* particle = new ParticleT(iVertex, vertex);
        particleField.addParticle(domain, particle);
    }
}

template<typename T, template<typename U> class Descriptor, class ParticleT>
CreateParticleFromVertex3D<T,Descriptor,ParticleT>* CreateParticleFromVertex3D<T,Descriptor,ParticleT>::clone() const {
    return new CreateParticleFromVertex3D<T,Descriptor,ParticleT>(*this);
}

template<typename T, template<typename U> class Descriptor, class ParticleT>
BlockDomain::DomainT CreateParticleFromVertex3D<T,Descriptor,ParticleT>::appliesTo() const {
    return BlockDomain::bulk;
}

template<typename T, template<typename U> class Descriptor, class ParticleT>
void CreateParticleFromVertex3D<T,Descriptor,ParticleT>::getTypeOfModification (
        std::vector<modif::ModifT>& modified ) const
{
    modified[0] = modif::dynamicVariables; // Particle field.
}


/* ******** CopyVertexToParticle3D *********************************** */

template<typename T, template<typename U> class Descriptor, class ParticleT>
CopyVertexToParticle3D<T,Descriptor,ParticleT>::CopyVertexToParticle3D (
        TriangularSurfaceMesh<T> const& mesh_)
    : mesh(mesh_)
{ }

template<typename T, template<typename U> class Descriptor, class ParticleT>
void CopyVertexToParticle3D<T,Descriptor,ParticleT>::processGenericBlocks (
        Box3D domain, std::vector<AtomicBlock3D*> blocks )
{
    PLB_PRECONDITION( blocks.size()==1 );
    ParticleField3D<T,Descriptor>& particleField =
        *dynamic_cast<ParticleField3D<T,Descriptor>*>(blocks[0]);

    std::vector<Particle3D<T,Descriptor>*> found;
    particleField.findParticles(domain, found);
    for(pluint iParticle=0; iParticle<found.size(); ++iParticle) {
        Particle3D<T,Descriptor>& particle = found[iParticle];
        plint tag = particle.getTag();
        Array<T,3> const& vertexPos = mesh.getVertex(tag);
        particle.getPosition() = vertexPos;
    }
}

template<typename T, template<typename U> class Descriptor, class ParticleT>
CopyVertexToParticle3D<T,Descriptor,ParticleT>* CopyVertexToParticle3D<T,Descriptor,ParticleT>::clone() const {
    return new CopyVertexToParticle3D<T,Descriptor,ParticleT>(*this);
}

template<typename T, template<typename U> class Descriptor, class ParticleT>
BlockDomain::DomainT CopyVertexToParticle3D<T,Descriptor,ParticleT>::appliesTo() const {
    return BlockDomain::bulk;
}

template<typename T, template<typename U> class Descriptor, class ParticleT>
void CopyVertexToParticle3D<T,Descriptor,ParticleT>::getTypeOfModification (
        std::vector<modif::ModifT>& modified ) const
{
    modified[0] = modif::dynamicVariables; // Particle field.
}


/* ******** CloneParticles3D *********************************** */

template<typename T, template<typename U> class Descriptor>
CloneParticles3D<T,Descriptor>::CloneParticles3D (
        Particle3D<T,Descriptor>* particleTemplate_ )
    : particleTemplate(particleTemplate_)
{ }

template<typename T, template<typename U> class Descriptor>
CloneParticles3D<T,Descriptor>::CloneParticles3D (
        CloneParticles3D<T,Descriptor> const& rhs )
    : particleTemplate(rhs.particleTemplate->clone())
{ }

template<typename T, template<typename U> class Descriptor>
CloneParticles3D<T,Descriptor>::~CloneParticles3D()
{
    delete particleTemplate;
}

template<typename T, template<typename U> class Descriptor>
CloneParticles3D<T,Descriptor>& CloneParticles3D<T,Descriptor>::operator= (
        CloneParticles3D<T,Descriptor> const& rhs )
{
    CloneParticles3D<T,Descriptor>(rhs).swap(*this);
    return *this;
}

template<typename T, template<typename U> class Descriptor>
void CloneParticles3D<T,Descriptor>::swap(CloneParticles3D<T,Descriptor>& rhs)
{
    std::swap(particleTemplate, rhs.particleTemplate);
}

template<typename T, template<typename U> class Descriptor>
void CloneParticles3D<T,Descriptor>::processGenericBlocks (
        Box3D domain, std::vector<AtomicBlock3D*> blocks )
{
    PLB_PRECONDITION( blocks.size()==2 );
    ParticleField3D<T,Descriptor>& originalField =
        *dynamic_cast<ParticleField3D<T,Descriptor>*>(blocks[0]);
    ParticleField3D<T,Descriptor>& clonedField =
        *dynamic_cast<ParticleField3D<T,Descriptor>*>(blocks[1]);

    std::vector<Particle3D<T,Descriptor>*> found;
    originalField.findParticles(domain, found);

    Dot3D offset = computeRelativeDisplacement(originalField, clonedField);
    Box3D clonedDomain(domain.shift(offset.x, offset.y, offset.z));

    for (pluint iParticle=0; iParticle<found.size(); ++iParticle) {
        Particle3D<T,Descriptor>& originalParticle = *found[iParticle];
        Particle3D<T,Descriptor>* clonedParticle = particleTemplate->clone();
        clonedParticle->reset(originalParticle.getPosition());
        clonedParticle->setTag(originalParticle.getTag());
        clonedField.addParticle(clonedDomain, clonedParticle);
    }
}

template<typename T, template<typename U> class Descriptor>
CloneParticles3D<T,Descriptor>* CloneParticles3D<T,Descriptor>::clone() const {
    return new CloneParticles3D<T,Descriptor>(*this);
}

template<typename T, template<typename U> class Descriptor>
void CloneParticles3D<T,Descriptor>::getTypeOfModification(std::vector<modif::ModifT>& modified) const {
    modified[0] = modif::nothing;  // Original particle field.
    modified[1] = modif::dynamicVariables;   // Cloned particle field.
}

template<typename T, template<typename U> class Descriptor>
BlockDomain::DomainT CloneParticles3D<T,Descriptor>::appliesTo() const {
    // The data processor acts on envelope too, but extension to the envelope
    //   is done manually in processGenericBlocks.
    return BlockDomain::bulk;
}


/* ******** CopyParticleToVertex3D *********************************** */

template<typename T, template<typename U> class Descriptor>
CopyParticleToVertex3D<T,Descriptor>::CopyParticleToVertex3D (
        TriangularSurfaceMesh<T>& mesh_)
    : mesh(mesh_)
{ }

template<typename T, template<typename U> class Descriptor>
void CopyParticleToVertex3D<T,Descriptor>::processGenericBlocks (
        Box3D domain, std::vector<AtomicBlock3D*> blocks )
{
    PLB_PRECONDITION( blocks.size()==1 );
    ParticleField3D<T,Descriptor>& particleField =
        *dynamic_cast<ParticleField3D<T,Descriptor>*>(blocks[0]);
    // Manually extend the domain to the full envelope.
    domain = particleField.getBoundingBox();

    std::vector<Particle3D<T,Descriptor>*> found;
    particleField.findParticles(domain, found);

    for (pluint iParticle=0; iParticle<found.size(); ++iParticle) {
        Particle3D<T,Descriptor>& particle = *found[iParticle];
        Array<T,3> position(particle.getPosition());
        plint vertexId = particle.getTag();
        mesh.replaceVertex(vertexId, position);
    }
}

template<typename T, template<typename U> class Descriptor>
CopyParticleToVertex3D<T,Descriptor>* CopyParticleToVertex3D<T,Descriptor>::clone() const {
    return new CopyParticleToVertex3D<T,Descriptor>(*this);
}

template<typename T, template<typename U> class Descriptor>
BlockDomain::DomainT CopyParticleToVertex3D<T,Descriptor>::appliesTo() const {
    // The data processor acts on envelope too, but extension to the envelope
    //   is done manually in processGenericBlocks.
    return BlockDomain::bulk;
}

template<typename T, template<typename U> class Descriptor>
void CopyParticleToVertex3D<T,Descriptor>::getTypeOfModification (
        std::vector<modif::ModifT>& modified ) const
{
    modified[0] = modif::nothing; // Particle field.
}


/* ******** ComputeParticleForce3D *********************************** */

template<typename T, template<typename U> class Descriptor>
ComputeParticleForce3D<T,Descriptor>::ComputeParticleForce3D (
        TriangleBoundary3D<T> const& triangleBoundary_,
        int flowType_, bool incompressibleModel_ )
    : triangleBoundary(triangleBoundary_),
      flowType(flowType_),
      incompressibleModel(incompressibleModel_)
{ }

template<typename T, template<typename U> class Descriptor>
void ComputeParticleForce3D<T,Descriptor>::processGenericBlocks (
        Box3D domain, std::vector<AtomicBlock3D*> blocks )
{
    PLB_PRECONDITION( blocks.size()==3 );
    ParticleField3D<T,Descriptor>& particleField =
        *dynamic_cast<ParticleField3D<T,Descriptor>*>(blocks[0]);
    BlockLattice3D<T,Descriptor>& fluid =
        *dynamic_cast<BlockLattice3D<T,Descriptor>*>(blocks[1]);
    ScalarField3D<int>& voxelMatrix =
        *dynamic_cast<ScalarField3D<int>*>(blocks[2]);

    // A few comments on the algorithm. To get the values of a macroscopic quantity
    // on a mesh vertex, the quantity is interpolated from the 4 (2D) or 8 (3D)
    // neighbors. The problem: all these neighbors are not necessarily well defined.
    // In the following scheme, which represents a possible situation in an interior
    // flow, the "o" stand for exterior nodes and the "x" for a boundary vertex. The
    // two left-most "o" are undefined, because none of their neighbors are interior
    // nodes.
    /*
               /
        o    o/
             /
            x
             \
        o    o\
               \
    */
    // These two nodes are therefore manually excluded in the code.

    std::vector<Particle3D<T,Descriptor>*> found;
    particleField.findParticles(domain, found);

    std::vector<Dot3D> cellPos(8);
    std::vector<T> weights(8);
    std::vector<Cell<T,Descriptor>*> cells(8);
    for (pluint iParticle=0; iParticle<found.size(); ++iParticle) {
        Particle3D<T,Descriptor>& particle = *found[iParticle];
        Array<T,3> position(particle.getPosition());
        plint vertexId = particle.getTag();
        Array<T,3> n = triangleBoundary.getMesh().computeVertexNormal(vertexId);

        plint numUsableCells=0;
        plint numTrials=0;
        while (numUsableCells==0 && numTrials<5) {
            linearInterpolationCoefficients(fluid, position, cellPos, weights);
            for (plint iCell=0; iCell<8; ++iCell) {
                cells[iCell] = &fluid.get(cellPos[iCell].x,cellPos[iCell].y,cellPos[iCell].z);
                int flag = voxelMatrix.get(cellPos[iCell].x,cellPos[iCell].y,cellPos[iCell].z);
                // In an interior flow for example, we exclude outer non-boundary cells, because
                //   they have undefined flow data.
                if (flag!=voxelFlag::invert(flowType)) {
                    ++numUsableCells;
                }
            }
            position -= 1.e-1*n;
            ++numTrials;
        }
        if (numUsableCells==0) {
            pcout << "Offending vertex: " << position[0] << "," << position[1] << "," << position[2] << std::endl;
        }
        //PLB_ASSERT(numUsableCells>0);
        if (numUsableCells <= 0) { // TODO: This code is temporary!! Must be changed!!!
            std::vector<Array<T,3> > vectors;
            Array<T,3> forceOnWall(0., 0., 0.);
            vectors.push_back(forceOnWall);
            particle.setVectors(vectors);

            T pressure = Descriptor<T>::cs2;
            T wss = 0.0;

            std::vector<T> scalars;
            scalars.push_back(pressure);
            scalars.push_back(wss);
            particle.setScalars(scalars);
            continue;
        }


        Cell<T,Descriptor>* cellOnVertex = 0;
        if (numUsableCells==8) {
            // Use copy constructor in order to initialize dynamics object.
            cellOnVertex = new Cell<T,Descriptor>(*cells[0]);
            for (plint iPop=0; iPop<Descriptor<T>::q; ++iPop) {
                (*cellOnVertex)[iPop] =
                    weights[0]*(*cells[0])[iPop] + weights[1]*(*cells[1])[iPop] + weights[2]*(*cells[2])[iPop] +
                    weights[3]*(*cells[3])[iPop] + weights[4]*(*cells[4])[iPop] + weights[5]*(*cells[5])[iPop] +
                    weights[6]*(*cells[6])[iPop] + weights[7]*(*cells[7])[iPop];
            }
        }
        else {
            bool firstUsableCell = true;
            T sumWeights = T();
            for (plint iCell=0; iCell<8; ++iCell) {
                int flag = voxelMatrix.get(cellPos[iCell].x,cellPos[iCell].y,cellPos[iCell].z);
                // In an interior flow for example, we exclude outer non-boundary cells, because
                //   they have undefined flow data.
                if (flag!=voxelFlag::invert(flowType)) {
                    if (firstUsableCell) {
                        // Use copy constructor in order to initialize dynamics object.
                        cellOnVertex = new Cell<T,Descriptor>(*cells[iCell]);
                        for (plint iPop=0; iPop<Descriptor<T>::q; ++iPop) {
                            (*cellOnVertex)[iPop]=T();
                        }
                        firstUsableCell = false;
                    }
                    sumWeights += weights[iCell];
                    for (plint iPop=0; iPop<Descriptor<T>::q; ++iPop) {
                        (*cellOnVertex)[iPop] += weights[iCell] * (*cells[iCell])[iPop];
                    }
                }
            }
            T invSumWeights = (T)1/sumWeights;
            for (plint iPop=0; iPop<Descriptor<T>::q; ++iPop) {
                (*cellOnVertex)[iPop] *= invSumWeights;
            }
        }

        T rhoBar;
        Array<T,3> j;
        Array<T,SymmetricTensorImpl<T,3>::n> PiNeq;
        cellOnVertex->getDynamics().computeRhoBarJPiNeq (
                *cellOnVertex, rhoBar, j, PiNeq);
        T rho = Descriptor<T>::fullRho(rhoBar);

        Array<T,3> Pi_n;
        SymmetricTensorImpl<T,3>::matVectMult(PiNeq, n, Pi_n);

        T omega = cellOnVertex->getDynamics().getOmega();
        Array<T,3> forceOnFluid;
        if (incompressibleModel) {
            forceOnFluid = -(rho-1.)*Descriptor<T>::cs2*n + rho*(omega/2.-1.)*Pi_n;  // Incompressible vision
        }
        else {
            forceOnFluid = -(rho-1.)*Descriptor<T>::cs2*n + (omega/2.-1.)*Pi_n;  // Compressible vision
        }
        Array<T,3> forceOnWall = -forceOnFluid;  // Remark by J.L: This is Newton!
        delete cellOnVertex;

        std::vector<Array<T,3> > vectors;
        vectors.push_back(forceOnWall);
        particle.setVectors(vectors);

        T pressure = (rho-1.)*Descriptor<T>::cs2;
        T wss = norm(forceOnWall - dot(n,forceOnWall)*n);

        std::vector<T> scalars;
        scalars.push_back(pressure);
        scalars.push_back(wss);
        particle.setScalars(scalars);
    }
}


template<typename T, template<typename U> class Descriptor>
ComputeParticleForce3D<T,Descriptor>*
    ComputeParticleForce3D<T,Descriptor>::clone() const
{
    return new ComputeParticleForce3D<T,Descriptor>(*this);
}

template<typename T, template<typename U> class Descriptor>
BlockDomain::DomainT ComputeParticleForce3D<T,Descriptor>::appliesTo() const {
    return BlockDomain::bulk; // Important: must be bulk-only,
    // because non-local access to fluid is made.
}

template<typename T, template<typename U> class Descriptor>
void ComputeParticleForce3D<T,Descriptor>::getTypeOfModification (
        std::vector<modif::ModifT>& modified ) const
{
    modified[0] = modif::dynamicVariables; // Particle field.
    modified[1] = modif::nothing;  // Fluid.
    modified[2] = modif::nothing;  // Voxel Matrix.
}


/* ******** GetParticleVelocity3D *********************************** */

template<typename T, template<typename U> class Descriptor>
void GetParticleVelocity3D<T,Descriptor>::processGenericBlocks (
        Box3D domain, std::vector<AtomicBlock3D*> blocks )
{
    PLB_PRECONDITION( blocks.size()==2 );
    ParticleField3D<T,Descriptor>& originalField =
        *dynamic_cast<ParticleField3D<T,Descriptor>*>(blocks[0]);
    ParticleField3D<T,Descriptor>& clonedField =
        *dynamic_cast<ParticleField3D<T,Descriptor>*>(blocks[1]);

    std::vector<Particle3D<T,Descriptor>*> found;
    originalField.findParticles(domain, found);

    Dot3D offset = computeRelativeDisplacement(originalField, clonedField);
    Box3D clonedDomain(domain.shift(offset.x, offset.y, offset.z));

    for (pluint iParticle=0; iParticle<found.size(); ++iParticle) {
        Particle3D<T,Descriptor>& originalParticle = *found[iParticle];
        Particle3D<T,Descriptor>* clonedParticle = new VisualParticle3D<T,Descriptor> (
                originalParticle.getTag(), originalParticle.getPosition() );
        std::vector<Array<T,3> > vectors;
        Array<T,3> velocity;
#ifdef PLB_DEBUG
        bool ok =
#endif
            originalParticle.getVector(0, velocity);
        PLB_ASSERT( ok );
        vectors.push_back(velocity);
        clonedParticle->setVectors(vectors),
        clonedField.addParticle(clonedDomain, clonedParticle);
    }
}

template<typename T, template<typename U> class Descriptor>
GetParticleVelocity3D<T,Descriptor>*
    GetParticleVelocity3D<T,Descriptor>::clone() const
{
    return new GetParticleVelocity3D<T,Descriptor>(*this);
}

template<typename T, template<typename U> class Descriptor>
BlockDomain::DomainT GetParticleVelocity3D<T,Descriptor>::appliesTo() const {
    return BlockDomain::bulk; // Important: must be bulk-only,
    // because non-local access to fluid is made.
}

template<typename T, template<typename U> class Descriptor>
void GetParticleVelocity3D<T,Descriptor>::getTypeOfModification (
        std::vector<modif::ModifT>& modified ) const
{
    modified[0] = modif::nothing; // Original field.
    modified[1] = modif::dynamicVariables;  // Cloned field.
}


/* ******** CreateParticleMap3D *********************************** */

template<typename T, template<typename U> class Descriptor>
void CreateParticleMap3D<T,Descriptor>::processGenericBlocks (
        Box3D domain, std::vector<AtomicBlock3D*> blocks )
{
    PLB_PRECONDITION( blocks.size()==4 );
    ParticleField3D<T,Descriptor>& particleField =
        dynamic_cast<ParticleField3D<T,Descriptor>&>(*blocks[0]);
    ScalarField3D<T>& xMatrix =
        dynamic_cast<ScalarField3D<T>&>(*blocks[1]);
    ScalarField3D<T>& yMatrix =
        dynamic_cast<ScalarField3D<T>&>(*blocks[2]);
    ScalarField3D<T>& zMatrix =
        dynamic_cast<ScalarField3D<T>&>(*blocks[3]);

    ScalarField3D<int> counter(xMatrix.getNx(), xMatrix.getNy(), xMatrix.getNz(), 0);

    std::vector<Particle3D<T,Descriptor>*> found;
    particleField.findParticles(domain, found);

    Dot3D location(particleField.getLocation());
    Dot3D offset = computeRelativeDisplacement(particleField, xMatrix);

    for (pluint iParticle=0; iParticle<found.size(); ++iParticle) {
        Particle3D<T,Descriptor>* originalParticle = found[iParticle];
        MappingParticle3D<T,Descriptor>* mappingParticle = dynamic_cast<MappingParticle3D<T,Descriptor>*>(originalParticle);
        if (mappingParticle && mappingParticle->endOfCycle()) {
            Array<T,3> pos1( mappingParticle->getPos1() );
            Array<T,3> pos2( mappingParticle->getPos2() );
            plint xPos = (plint)(pos1[0])-location.x;
            plint yPos = (plint)(pos1[1])-location.y;
            plint zPos = (plint)(pos1[2])-location.z;
            plint xPosOfs = xPos+offset.x;
            plint yPosOfs = yPos+offset.y;
            plint zPosOfs = zPos+offset.z;
            ++counter.get(xPosOfs, yPosOfs, zPosOfs);
            xMatrix.get(xPosOfs,yPosOfs,zPosOfs) += pos2[0];
            yMatrix.get(xPosOfs,yPosOfs,zPosOfs) += pos2[1];
            zMatrix.get(xPosOfs,yPosOfs,zPosOfs) += pos2[2];
        }
    }
    for (plint iX=domain.x0; iX<=domain.x1; ++iX) {
        for (plint iY=domain.y0; iY<=domain.y1; ++iY) {
            for (plint iZ=domain.z0; iZ<=domain.z1; ++iZ) {
                plint xPosOfs = iX+offset.x;
                plint yPosOfs = iY+offset.y;
                plint zPosOfs = iZ+offset.z;
                int count = counter.get(xPosOfs,yPosOfs,zPosOfs);
                if (count==0) {
                    xMatrix.get(xPosOfs,yPosOfs,zPosOfs)=-1.;
                    yMatrix.get(xPosOfs,yPosOfs,zPosOfs)=-1.;
                    zMatrix.get(xPosOfs,yPosOfs,zPosOfs)=-1.;
                }
                else {
                    xMatrix.get(xPosOfs,yPosOfs,zPosOfs) /= (T) count;
                    yMatrix.get(xPosOfs,yPosOfs,zPosOfs) /= (T) count;
                    zMatrix.get(xPosOfs,yPosOfs,zPosOfs) /= (T) count;
                }
            }
        }
    }
}

template<typename T, template<typename U> class Descriptor>
CreateParticleMap3D<T,Descriptor>*
    CreateParticleMap3D<T,Descriptor>::clone() const
{
    return new CreateParticleMap3D<T,Descriptor>(*this);
}

template<typename T, template<typename U> class Descriptor>
void CreateParticleMap3D<T,Descriptor>::getTypeOfModification (
        std::vector<modif::ModifT>& modified ) const
{
    modified[0] = modif::nothing; // Particle field.
    modified[1] = modif::staticVariables;  // x-Matrix.
    modified[2] = modif::staticVariables;  // y-Matrix.
    modified[3] = modif::staticVariables;  // z-Matrix.
}

}  // namespace plb

#endif  // VISUAL_PARTICLE_FUNCTIONAL_3D_HH
